/*
 * motor.c
 *
 *  Created on: 10.07.2010
 *      Author: AndreR
 */

/*
 * Motor 1:
 * Hall IRQ: C0
 * Hall: G6 - G8
 * A1, B1, A2, B2, A3, B3: G9 - G14
 *
 * Motor 2:
 * Hall IRQ: C2
 * Hall: F6 - F8
 * A1, B1, A2, B2, A3, B3: F0 - F5
 *
 * Motor 3:
 * Hall IRQ: C4
 * Hall: E13 - E15
 * A1, B1, A2, B2, A3, B3: F9 - F14
 *
 * Motor 4:
 * Hall IRQ: C3
 * Hall: D13 - D15
 * A1, B1, A2, B2, A3, B3: G0 - G5
 *
 * Motor 5:
 * Hall IRQ: C1
 * Hall: E6 - E8
 * A1, B1, A2, B2, A3, B3: E0 - E5
 */

#include "motor.h"

#include "stm32f10x.h"
#include "misc.h"
#include "stm32f10x_rcc.h"
#include "stm32f10x_gpio.h"
#include "stm32f10x_tim.h"
#include "stm32f10x_exti.h"

#include "util/bit_banding.h"

#include "constants.h"

/*
 * DEFINES
 */
#define M1Mask(a1, b1, a2, b2, a3, b3) (a1 << 9 | b1 << 10 | a2 << 11 | b2 << 12 | a3 << 13 | b3 << 14)
#define M2Mask(a1, b1, a2, b2, a3, b3) (a1 << 0 | b1 << 1  | a2 << 2  | b2 << 3  | a3 << 4  | b3 << 5)
#define M3Mask(a1, b1, a2, b2, a3, b3) (a1 << 9 | b1 << 10 | a2 << 11 | b2 << 12 | a3 << 13 | b3 << 14)
#define M4Mask(a1, b1, a2, b2, a3, b3) (a1 << 0 | b1 << 1  | a2 << 2  | b2 << 3  | a3 << 4  | b3 << 5)
#define M5Mask(a1, b1, a2, b2, a3, b3) (a1 << 0 | b1 << 1  | a2 << 2  | b2 << 3  | a3 << 4  | b3 << 5)

#define MMask(ha, la, hb, lb, hc, lc) \
	( ((ha <<  0 | la <<  1 | hb <<  2 | lb <<  3 | hc <<  4 | lc <<  5) ^ 0x3F) | \
	   (ha << 16 | la << 17 | hb << 18 | lb << 19 | hc << 20 | lc << 21))

/*
 * GLOBAL DATA
 */
MotorGlobal motor;

/*
 * PROTOTYPES
 */

/*
 * IRQ HANDLERS
 */
void TIM5_IRQHandler()
{
	uint32_t sr = TIM5->SR;
	TIM5->SR = 0;	// reset flags

	if(sr & 0x01)	// Counter Update (Overflow)
	{
		GPIOG->BSRR = motor.M[0].A;
		GPIOF->BSRR = motor.M[1].A;
		GPIOF->BSRR = motor.M[2].A;
		GPIOG->BSRR = motor.M[3].A;
	}

	if(sr & 0x02)	// Compare Match 1
	{
		TIM5->CCR1 = motor.M[0].power;
		GPIOG->BSRR = motor.M[0].B;
	}

	if(sr & 0x04)	// Compare Match 2
	{
		TIM5->CCR2 = motor.M[1].power;
		GPIOF->BSRR = motor.M[1].B;
	}

	if(sr & 0x08)	// Compare Match 3, Motor 3
	{
		TIM5->CCR3 = motor.M[2].power;
		GPIOF->BSRR = motor.M[2].B;
	}

	if(sr & 0x10)	// Compare Match 4, Motor 4
	{
		TIM5->CCR4 = motor.M[3].power;
		GPIOG->BSRR =  motor.M[3].B;
	}
}

void TIM8_UP_IRQHandler()
{
	TIM8->SR = 0xFFFE;

	if(motor.M[4].power != 0)
		GPIOE->BSRR = motor.M[4].A;	//set on
}

void TIM8_CC_IRQHandler()
{
	TIM8->SR = 0xFFE1;	// reset CC flags

	TIM8->CCR1 = motor.M[4].power;
	GPIOE->BSRR = motor.M[4].B;	//set off
}

void EXTI0_IRQHandler()	// Motor 1
{
	uint16_t ph = M1Pos;
	if(motor.M[0].ccw)
		ph ^= 0b111;
	motor.M[0].A = motor.M[0].CWLUTA[ph];
	motor.M[0].B = motor.M[0].CWLUTB[ph];
	motor.M[0].coms++;
	BBPeriphSet(EXTI_BASE + 0x14, 0);
}

void EXTI1_IRQHandler()	// Motor 5
{
	uint16_t ph = M5Pos;
	if(motor.M[4].ccw)
		ph ^= 0b111;
	motor.M[4].A = motor.M[4].CWLUTA[ph];
	motor.M[4].B = motor.M[4].CWLUTB[ph];
//	GPIOE->BSRR = motor.M[4].B;	// early cut off to save power, set off
	motor.M[4].coms++;
	BBPeriphSet(EXTI_BASE + 0x14, 1);
}

void EXTI2_IRQHandler()	// Motor 2
{
	uint16_t ph = M2Pos;
	if(motor.M[1].ccw)
		ph ^= 0b111;
	motor.M[1].A = motor.M[1].CWLUTA[ph];
	motor.M[1].B = motor.M[1].CWLUTB[ph];
	motor.M[1].coms++;
	BBPeriphSet(EXTI_BASE + 0x14, 2);
}

void EXTI3_IRQHandler()	// Motor 4
{
	uint16_t ph = M4Pos;
	if(motor.M[3].ccw)
		ph ^= 0b111;
	motor.M[3].A = motor.M[3].CWLUTA[ph];
	motor.M[3].B = motor.M[3].CWLUTB[ph];
	motor.M[3].coms++;
	BBPeriphSet(EXTI_BASE + 0x14, 3);
}

void EXTI4_IRQHandler()	// Motor 3
{
	uint16_t ph = M3Pos;
	if(motor.M[2].ccw)
		ph ^= 0b111;
	motor.M[2].A = motor.M[2].CWLUTA[ph];
	motor.M[2].B = motor.M[2].CWLUTB[ph];
	motor.M[2].coms++;
	BBPeriphSet(EXTI_BASE + 0x14, 4);
}

/*
 * INTERNAL FUNCTIONS
 */
void hallInit()
{
	GPIO_InitTypeDef gpioInit;
	gpioInit.GPIO_Mode = GPIO_Mode_IN_FLOATING;

	gpioInit.GPIO_Pin = GPIO_Pin_0 | GPIO_Pin_1 | GPIO_Pin_2 | GPIO_Pin_3 | GPIO_Pin_4;
	GPIO_Init(GPIOC, &gpioInit);	// Hall interrupts

	gpioInit.GPIO_Pin = GPIO_Pin_6 | GPIO_Pin_7 | GPIO_Pin_8 | GPIO_Pin_13 | GPIO_Pin_14 | GPIO_Pin_15;
	GPIO_Init(GPIOE, &gpioInit);	// H3 + H5

	gpioInit.GPIO_Pin = GPIO_Pin_13 | GPIO_Pin_14 | GPIO_Pin_15;
	GPIO_Init(GPIOD, &gpioInit);	// H4

	gpioInit.GPIO_Pin = GPIO_Pin_6 | GPIO_Pin_7 | GPIO_Pin_8;
	GPIO_Init(GPIOF, &gpioInit);	// H2

	gpioInit.GPIO_Pin = GPIO_Pin_6 | GPIO_Pin_7 | GPIO_Pin_8;
	GPIO_Init(GPIOG, &gpioInit);	// H1

	GPIO_EXTILineConfig(GPIO_PortSourceGPIOC, GPIO_PinSource0);
	GPIO_EXTILineConfig(GPIO_PortSourceGPIOC, GPIO_PinSource1);
	GPIO_EXTILineConfig(GPIO_PortSourceGPIOC, GPIO_PinSource2);
	GPIO_EXTILineConfig(GPIO_PortSourceGPIOC, GPIO_PinSource3);
	GPIO_EXTILineConfig(GPIO_PortSourceGPIOC, GPIO_PinSource4);

	EXTI_InitTypeDef extiInit;
	EXTI_StructInit(&extiInit);
	extiInit.EXTI_Line = EXTI_Line0 | EXTI_Line1 | EXTI_Line2 | EXTI_Line3 | EXTI_Line4;
	extiInit.EXTI_LineCmd = ENABLE;
	extiInit.EXTI_Mode = EXTI_Mode_Interrupt;
	extiInit.EXTI_Trigger = EXTI_Trigger_Rising_Falling;
	EXTI_Init(&extiInit);

	EXTI->IMR &= ~(0x1F);

	NVIC_InitTypeDef nvicInit;
	nvicInit.NVIC_IRQChannelCmd = ENABLE;
	nvicInit.NVIC_IRQChannelPreemptionPriority = IRQL_MOTOR_HALL;
	nvicInit.NVIC_IRQChannelSubPriority = 0;

	nvicInit.NVIC_IRQChannel = EXTI0_IRQn;
	NVIC_Init(&nvicInit);

	nvicInit.NVIC_IRQChannel = EXTI1_IRQn;
	NVIC_Init(&nvicInit);

	nvicInit.NVIC_IRQChannel = EXTI2_IRQn;
	NVIC_Init(&nvicInit);

	nvicInit.NVIC_IRQChannel = EXTI3_IRQn;
	NVIC_Init(&nvicInit);

	nvicInit.NVIC_IRQChannel = EXTI4_IRQn;
	NVIC_Init(&nvicInit);
}

void initMotorDatasetActiveFreewheel(MotorDataSet* pData, uint32_t shift)
{
	pData->A = 0;
	pData->B = MMask(0, 0, 0, 0, 0, 0) << shift;
	pData->ccw = 0;
	pData->coms = 0;
	pData->power = 0;

	pData->CWLUTA[0]	= MMask(0, 0, 0, 0, 0, 0) << shift;
	pData->CWLUTB[0]	= MMask(0, 1, 0, 1, 0, 1) << shift;
	pData->CWLUTA[7]	= MMask(0, 1, 0, 1, 0, 1) << shift;
	pData->CWLUTB[7]	= MMask(0, 1, 0, 1, 0, 1) << shift;

	pData->CWLUTA[5]	= MMask(1, 0, 0, 1, 0, 0) << shift;
	pData->CWLUTA[1]	= MMask(1, 0, 0, 0, 0, 1) << shift;
	pData->CWLUTA[3]	= MMask(0, 0, 1, 0, 0, 1) << shift;
	pData->CWLUTA[2]	= MMask(0, 1, 1, 0, 0, 0) << shift;
	pData->CWLUTA[6]	= MMask(0, 1, 0, 0, 1, 0) << shift;
	pData->CWLUTA[4]	= MMask(0, 0, 0, 1, 1, 0) << shift;

	pData->CWLUTB[5]	= MMask(0, 1, 0, 1, 0, 0) << shift;
	pData->CWLUTB[1]	= MMask(0, 1, 0, 0, 0, 1) << shift;
	pData->CWLUTB[3]	= MMask(0, 0, 0, 1, 0, 1) << shift;
	pData->CWLUTB[2]	= MMask(0, 1, 0, 1, 0, 0) << shift;
	pData->CWLUTB[6]	= MMask(0, 1, 0, 0, 0, 1) << shift;
	pData->CWLUTB[4]	= MMask(0, 0, 0, 1, 0, 1) << shift;
}

/*
 * PUBLIC FUNCTIONS
 */
void MotorInit()
{
	GPIO_InitTypeDef gpioInit;

	motor.maxPWM = 49;

	// Look-Up-Table Definitions
	// WARNING: Modifications can cause sever hardware damage!!!
	motor.PhaseLUT[0] = 0;
	motor.PhaseLUT[1] = 5;
	motor.PhaseLUT[2] = 1;
	motor.PhaseLUT[3] = 3;
	motor.PhaseLUT[4] = 2;
	motor.PhaseLUT[5] = 6;
	motor.PhaseLUT[6] = 4;
	motor.PhaseLUT[7] = 0;

	motor.EXTILUT[0] = 0;
	motor.EXTILUT[1] = 2;
	motor.EXTILUT[2] = 4;
	motor.EXTILUT[3] = 3;
	motor.EXTILUT[4] = 1;

	initMotorDatasetActiveFreewheel(&motor.M[1], 0);
	initMotorDatasetActiveFreewheel(&motor.M[0], 9);
	initMotorDatasetActiveFreewheel(&motor.M[2], 9);
	initMotorDatasetActiveFreewheel(&motor.M[3], 0);
	initMotorDatasetActiveFreewheel(&motor.M[4], 0);

	MotorSetPower(0, 0);
	MotorSetPower(1, 0);
	MotorSetPower(2, 0);
	MotorSetPower(3, 0);
	MotorSetPower(4, 0);

	// GPIO Init
	gpioInit.GPIO_Mode = GPIO_Mode_Out_PP;
	gpioInit.GPIO_Speed = GPIO_Speed_50MHz;

	gpioInit.GPIO_Pin = M1Mask(1, 1, 1, 1, 1, 1) | M4Mask(1, 1, 1, 1, 1, 1);
	GPIO_Init(GPIOG, &gpioInit);

	gpioInit.GPIO_Pin = M2Mask(1, 1, 1, 1, 1, 1) | M3Mask(1, 1, 1, 1, 1, 1);
	GPIO_Init(GPIOF, &gpioInit);

	gpioInit.GPIO_Pin = M5Mask(1, 1, 1, 1, 1,  1);
	GPIO_Init(GPIOE, &gpioInit);

	//Timer Init, TIM5
	RCC_APB1PeriphClockCmd(RCC_APB1Periph_TIM5, ENABLE);

	TIM_TimeBaseInitTypeDef timeBaseInit;
	timeBaseInit.TIM_ClockDivision = TIM_CKD_DIV1;
	timeBaseInit.TIM_CounterMode = TIM_CounterMode_Up;
	timeBaseInit.TIM_Period = motor.maxPWM;	//99
	timeBaseInit.TIM_Prescaler = 89;	// 89 = 1.25us clock, 179 = 2.5us clock
	timeBaseInit.TIM_RepetitionCounter = 0;
	TIM_TimeBaseInit(TIM5, &timeBaseInit);

	TIM_ARRPreloadConfig(TIM5, DISABLE);
	TIM_CCPreloadControl(TIM5, DISABLE);

	TIM_OCInitTypeDef ocInit;
	ocInit.TIM_OCMode = TIM_OCMode_Timing;
	ocInit.TIM_OCIdleState = TIM_OCIdleState_Reset;
	ocInit.TIM_OCNIdleState = TIM_OCNIdleState_Reset;
	ocInit.TIM_OCPolarity = TIM_OCPolarity_Low;
	ocInit.TIM_OCNPolarity = TIM_OCNPolarity_Low;
	ocInit.TIM_OutputState = TIM_OutputState_Disable;
	ocInit.TIM_OutputNState = TIM_OutputNState_Disable;
	ocInit.TIM_Pulse = 0;

	TIM_OC1Init(TIM5, &ocInit);
	TIM_OC2Init(TIM5, &ocInit);
	TIM_OC3Init(TIM5, &ocInit);
	TIM_OC4Init(TIM5, &ocInit);

	TIM_InternalClockConfig(TIM5);

	TIM_ITConfig(TIM5, TIM_IT_Update | TIM_IT_CC1 | TIM_IT_CC2 | TIM_IT_CC3 | TIM_IT_CC4, ENABLE);

	//Timer Init, TIM8
	RCC_APB2PeriphClockCmd(RCC_APB2Periph_TIM8, ENABLE);

	timeBaseInit.TIM_Period = motor.maxPWM;
	timeBaseInit.TIM_Prescaler = 44;	// 44
	TIM_TimeBaseInit(TIM8, &timeBaseInit);

	TIM_ARRPreloadConfig(TIM8, DISABLE);
	TIM_CCPreloadControl(TIM8, DISABLE);

/*	TIM_OC1Init(TIM8, &ocInit);	// FIXME: kills SPI1 MOSI
	TIM_OC2Init(TIM8, &ocInit);
	TIM_OC3Init(TIM8, &ocInit);
	TIM_OC4Init(TIM8, &ocInit);
*/
	TIM_InternalClockConfig(TIM8);

	TIM_ITConfig(TIM8, TIM_IT_Update | TIM_IT_CC1, ENABLE);

	hallInit();

	NVIC_InitTypeDef nvicInit;
	nvicInit.NVIC_IRQChannel = TIM5_IRQn;
	nvicInit.NVIC_IRQChannelCmd = ENABLE;
	nvicInit.NVIC_IRQChannelPreemptionPriority = IRQL_MOTOR_PWM_MOVEMENT;
	nvicInit.NVIC_IRQChannelSubPriority = 0;
	NVIC_Init(&nvicInit);

	nvicInit.NVIC_IRQChannelPreemptionPriority = IRQL_MOTOR_PWM_DRIBBLER;
	nvicInit.NVIC_IRQChannel = TIM8_CC_IRQn;
	NVIC_Init(&nvicInit);

	nvicInit.NVIC_IRQChannel = TIM8_UP_IRQn;
	NVIC_Init(&nvicInit);

	TIM_Cmd(TIM5, ENABLE);
	TIM_Cmd(TIM8, ENABLE);
}

void MotorEmergencyShutdown()
{
	TIM5->CR1 &= ~(TIM_CR1_CEN);
	TIM8->CR1 &= ~(TIM_CR1_CEN);

	GPIOG->BSRR = motor.M[0].CWLUTB[0];
	GPIOF->BSRR = motor.M[1].CWLUTB[0];
	GPIOF->BSRR = motor.M[2].CWLUTB[0];
	GPIOG->BSRR = motor.M[3].CWLUTB[0];
	GPIOE->BSRR = motor.M[4].CWLUTB[0];
}

void MotorEmergencyRecovery()
{
	TIM5->CR1 |= TIM_CR1_CEN;
	TIM8->CR1 |= TIM_CR1_CEN;
}

void MotorSetPower(uint8_t id, uint16_t newPower)
{
	if(newPower >= motor.maxPWM)
		newPower = motor.maxPWM-1;

	if(newPower == 0)
	{
		BBPeriphReset(EXTI_BASE, motor.EXTILUT[id]);	// Disable hall interrupt

		motor.M[id].A = motor.M[id].CWLUTA[0];
		motor.M[id].B = motor.M[id].CWLUTB[0];
		motor.M[id].power = 0;
	}
	else
	{
		motor.M[id].power = newPower;
		BBPeriphSet(EXTI_BASE, motor.EXTILUT[id]);	// Enable interrupt
		BBPeriphSet(EXTI_BASE + 0x10, motor.EXTILUT[id]);	// Trigger Software interrupt
	}
}

void MotorClearDriverFault(uint8_t id)
{
	BBPeriphReset(EXTI_BASE, motor.EXTILUT[id]);	// Disable hall interrupt

	motor.M[id].A = motor.M[id].CWLUTA[0];
	motor.M[id].B = motor.M[id].CWLUTB[0];
	motor.M[id].power = motor.maxPWM-1;
}
