/*
 * init.c
 *
 *  Created on: 17.05.2011
 *      Author: AndreR
 */

#include "init.h"

#include "stm32f10x_iwdg.h"

#include "FreeRTOS.h"
#include "task.h"

#include "atmega/ir_prog.h"
#include "atmega/kicker_prog.h"
#include "atmega/motor_prog.h"
#include "atmega/power_prog.h"

#include "cli.h"
#include "network.h"
#include "ir.h"
#include "spi1_2.h"
#include "movement.h"
#include "power.h"
#include "acceleration.h"
#include "kicker.h"
#include "network/dect_interface.h"
#include "usb/usb.h"
#include "util/console.h"
#include "hal/usart1.h"
#include "usb/usb.h"

void InitTask(void* params)
{
	params = params;

	uint8_t kickerAttached = 1;
	uint8_t irAttached = 1;

	while(1)
	{
		// -- CLI
		ConsoleAddInterface(&usart1.rxFifo, &usart1.txFifo, &USART1StartTx);
		ConsoleAddInterface(&usb.rxFifo, &usb.txFifo, &USBStartTx);

		USBEnable(1);

		ConsolePrint("\f\e[0;36mInitializing...\e[0m\r\n");

		uint32_t uid1 = *((uint32_t*)(0x1FFFF7E8 + 0));	// CPU ID register (96 bit)
		uint32_t uid2 = *((uint32_t*)(0x1FFFF7E8 + 4));
		uint32_t uid3 = *((uint32_t*)(0x1FFFF7E8 + 8));

		ConsolePrint("CPU ID: 0x%X 0x%X 0x%X\r\n", uid1, uid2, uid3);

		xTaskCreate(CLITask, (const signed char* const)"CLI", 256, NULL, 1, NULL);
		ConsolePrint("CLI: \e[0;32mOK\e[0m\r\n");

		// -- Network
		DECTInterfaceSetMode(DECT_MODE_DATA);

		ConsolePrint("Network: \e[0;32mOK\e[0m\r\n");

		// -- SPI 1, 2
		xTaskCreate(SPI1_2Task, (const signed char* const)"SPI", 256, NULL, 1, NULL);

		ConsolePrint("SPI 1 + 2: \e[0;32mOK\e[0m\r\n");

		// -- Power
		uint8_t programUpdated = 0;
		int16_t result = ATMegaUpdateProgram(&power.atmega, power_prog_data, POWER_PROG_SIZE, &programUpdated);

		ConsolePrint("--- Power AVR ---\r\n");
		ConsolePrint("Signature: %X %X %X\r\n", power.atmega.signature[0], power.atmega.signature[1], power.atmega.signature[2]);
		ConsolePrint("Fuses: %X %X %X\r\n", power.atmega.fuses[0], power.atmega.fuses[1], power.atmega.fuses[2]);
		if(programUpdated)
			ConsolePrint("\e[0;36mUpdated power program\e[0m\r\n");
		if(result)
			ConsolePrint("\e[0;31mPower program update failed (0x%X)\e[0m\r\n", result);

		vTaskDelay(100/portTICK_RATE_MS);

		// -- Movement
		vTaskDelay(500/portTICK_RATE_MS);

		movement.atmega.pResetPort->BRR = movement.atmega.resetPin;

		vTaskDelay(500/portTICK_RATE_MS);

		movement.atmega.pResetPort->BSRR = movement.atmega.resetPin;

		vTaskDelay(100/portTICK_RATE_MS);

		programUpdated  = 0;
		result = ATMegaUpdateProgram(&movement.atmega, motor_prog_data, MOTOR_PROG_SIZE, &programUpdated);

		ConsolePrint("--- Motor AVR ---\r\n");
		ConsolePrint("Signature: %X %X %X\r\n", movement.atmega.signature[0], movement.atmega.signature[1], movement.atmega.signature[2]);
		ConsolePrint("Fuses: %X %X %X\r\n", movement.atmega.fuses[0], movement.atmega.fuses[1], movement.atmega.fuses[2]);
		if(programUpdated)
			ConsolePrint("\e[0;36mUpdated motor program\e[0m\r\n");
		if(result)
			ConsolePrint("\e[0;31mMotor program update failed (0x%X)\e[0m\r\n", result);

		vTaskDelay(100/portTICK_RATE_MS);

		// -- IR
		programUpdated = 0;
		result = ATMegaUpdateProgram(&ir.atmega, ir_prog_data, IR_PROG_SIZE, &programUpdated);

		ConsolePrint("--- IR AVR ---\r\n");
		ConsolePrint("Signature: %X %X %X\r\n", ir.atmega.signature[0], ir.atmega.signature[1], ir.atmega.signature[2]);
		ConsolePrint("Fuses: %X %X %X\r\n", ir.atmega.fuses[0], ir.atmega.fuses[1], ir.atmega.fuses[2]);
		if(programUpdated)
			ConsolePrint("\e[0;36mUpdated ir program\e[0m\r\n");
		if(result)
		{
			ConsolePrint("\e[0;31mIR program update failed (0x%X)\r\n", result);
			irAttached = 0;
			ConsolePrint("Not starting IR task\e[0m\r\n");
		}

		vTaskDelay(100/portTICK_RATE_MS);

		// -- Kicker
		programUpdated = 0;
		result = ATMegaUpdateProgram(&kicker.atmega, kicker_prog_data, KICKER_PROG_SIZE, &programUpdated);

		ConsolePrint("--- Kicker AVR ---\r\n");
		ConsolePrint("Signature: %X %X %X\r\n", kicker.atmega.signature[0], kicker.atmega.signature[1], kicker.atmega.signature[2]);
		ConsolePrint("Fuses: %X %X %X\r\n", kicker.atmega.fuses[0], kicker.atmega.fuses[1], kicker.atmega.fuses[2]);
		if(programUpdated)
			ConsolePrint("\e[0;36mUpdated kicker program\e[0m\r\n");
		if(result)
		{
			ConsolePrint("\e[0;31mKicker program update failed (0x%X)\r\n", result);
			kickerAttached = 0;
			ConsolePrint("Not starting kicker task\e[0m\r\n");
		}

		vTaskDelay(100/portTICK_RATE_MS);

		// -- Acceleration
		SPILIS3SetConfiguration(&acceleration.inner, 1);
		SPILIS3SetConfiguration(&acceleration.outer, 1);

		vTaskDelay(10/portTICK_RATE_MS);

		AccelerationCalibrate();

		ConsolePrint("Acceleration: \e[0;32mOK\e[0m\r\n");

		// Create Tasks, MAX PRIORITY: 4
		xTaskCreate(NetworkTask, (const signed char* const)"DECT", 256, NULL, 2, NULL);
		xTaskCreate(PowerTask, (const signed char* const)"Power", 256, NULL, 1, NULL);
		xTaskCreate(MovementTask, (const signed char* const)"Movement", 512, NULL, 3, NULL);
//		xTaskCreate(AccelerationTask, (const signed char* const)"Accel", 256, NULL, 1, NULL);

		if(irAttached)
			xTaskCreate(IRTask, (const signed char* const)"IR", 256, NULL, 1, NULL);

		if(kickerAttached)
			xTaskCreate(KickerTask, (const signed char* const)"Kicker", 256, NULL, 1, NULL);

		ConsolePrint("Tasks started\r\n");
		ConsolePrint("Initialization done.\r\n");

		// Enable independent watchdog
//		IWDG_ReloadCounter();
//		IWDG_Enable();

		ConsolePrint("Watchdog online\r\n");

		vTaskSuspend(NULL);	// Disable this task
	}
}
