#include "movement.h"

#include "FreeRTOS.h"
#include "task.h"

#include "hal/motor.h"
#include "atmega/motor_prog.h"

#include "util/fastmath.h"
#include "commands.h"
#include "network.h"
#include "power.h"

#include "spi1_2.h"

#define MOVE_ATMEGA_CS (1 << 10)	// GPIOB
#define MOVE_ATMEGA_RST (1 << 11)	// GPIOB

#define MOVE_SIN_A		46947L	// [e-6]
#define MOVE_SIN_B		70711L	// [e-6]
#define MOVE_COS_A		88295L	// [e-6]
#define MOVE_COS_B		70711L	// [e-6]
#define MOVE_VM2PID		104540L	// [e-6], 99310 @ r_R=3cm
#define MOVE_PIDMAX		9360L

#define MOVE_DRIBMAX	9000L
#define MOVE_DRIB2PID	360L		// [e-3]

/*
 * VM -> TPF_100 = 0.0763926 [1/mm]
 * MOVE_VM2PID = VM->TPF_100 * 13 = 0.9931038
 *
 * PID Max: 9360
 * Motor RPM max: 10000
 * TPF_100 max: 720
 * Feedback scale: *13
 *
 * Max VM (@10000rpm): 9.425 [m/s]
 *
 * Dribbler max: 25000rpm
 */

// Definition for calculating vx, vy, vw from wheel speeds
#define MOVE_R_VX	42496	// [e-6]
#define MOVE_R_VY1	34501	// [e-6]
#define MOVE_R_VY2	27630	// [e-6]
#define MOVE_R_VW1	30049	// [e-6]
#define MOVE_R_VW2	19950	// [e-6]

// misc fixed values
#define MOVE_BOT_RAD	8200		// [mm*e-2]
#define MOVE_REACTION_THRESHOLD	10	// in PID output units
#define MOVE_STALL_TIMEOUT 1000	// [ms]

MovementGlobal movement;

int32_t DriveFeedback(MovementDataSet* pData)
{
	int32_t dt_real = (xTaskGetTickCount() - pData->lastTickCount)*10*portTICK_RATE_MS;

	int32_t n_enc = (int16_t)pData->pTIM->CNT;
	pData->pTIM->CNT = 0;

	if(n_enc - pData->lastEncCount > 1000)
		n_enc = pData->lastEncCount + 1000;
	else if(pData->lastEncCount - n_enc > 1000)
		n_enc = pData->lastEncCount - 1000;

	pData->lastEncCount = n_enc;

	n_enc = (n_enc * 13 * pData->pid.dt)/dt_real;

	pData->lastTickCount = xTaskGetTickCount();

	return -n_enc;	// invert feedback due to transmission
}

void DriveOutput(uint8_t id, int32_t out)
{
	if(out == 0)
	{
		MotorSetPower(id, 0);
		return;
	}

	// Invert output due to transmission
	out *= -1;

	if(out < 0)
	{
		MotorSetDir(id, 1);
		MotorSetPower(id, -out);
	}

	if(out > 0)
	{
		MotorSetDir(id, 0);
		MotorSetPower(id, out);
	}
}

int32_t PidFeedbackM1()
{
	return DriveFeedback(&movement.M[0]);
}

void PidOutputM1(int32_t out)
{
	DriveOutput(0, (out * motor.maxPWM)/movement.M[0].pid.max);
}

int32_t PidFeedbackM2()
{
	return DriveFeedback(&movement.M[1]);
}

void PidOutputM2(int32_t out)
{
	DriveOutput(1, (out * motor.maxPWM)/movement.M[1].pid.max);
}

int32_t PidFeedbackM3()
{
	return DriveFeedback(&movement.M[2]);
}

void PidOutputM3(int32_t out)
{
	DriveOutput(2, (out * motor.maxPWM)/movement.M[2].pid.max);
}

int32_t PidFeedbackM4()
{
	return DriveFeedback(&movement.M[3]);
}

void PidOutputM4(int32_t out)
{
	DriveOutput(3, (out * motor.maxPWM)/movement.M[3].pid.max);
}

int32_t PidFeedbackM5()
{
	MovementDataSet* pData = &movement.M[4];

	int32_t dt_real = (xTaskGetTickCount() - pData->lastTickCount)*10;

	int32_t n_enc = motor.M[4].ccw == 1 ? motor.M[4].coms : -motor.M[4].coms;
	motor.M[4].coms = 0;

	n_enc = (n_enc * 36 * movement.M[4].pid.dt)/dt_real;

	pData->lastTickCount = xTaskGetTickCount();

	return n_enc;
}

void PidOutputM5(int32_t out)
{
	DriveOutput(4, (out * motor.maxPWM)/movement.M[4].pid.max);
}

void PidOutputRotComp(int32_t out)
{
	movement.rotCompOut = out;
}

int32_t PidFeedbackRotComp()
{
	return movement.rotCompFeedback;
}

void MovementInit()
{
	GPIO_InitTypeDef gpioInit;

	gpioInit.GPIO_Mode = GPIO_Mode_Out_PP;
	gpioInit.GPIO_Pin = MOVE_ATMEGA_CS | MOVE_ATMEGA_RST;
	gpioInit.GPIO_Speed = GPIO_Speed_2MHz;
	GPIO_Init(GPIOB, &gpioInit);

	GPIOB->BSRR = MOVE_ATMEGA_CS | MOVE_ATMEGA_RST;

	ATMegaInit(&movement.atmega, &spi2, GPIOB, MOVE_ATMEGA_CS, GPIOB, MOVE_ATMEGA_RST);

	movement.M[0].pTIM = TIM3;
	movement.M[1].pTIM = TIM2;
	movement.M[2].pTIM = TIM1;
	movement.M[3].pTIM = TIM4;

	movement.mode = MOTOR_MODE_MANUAL;

	for(uint8_t i = 0; i < 5; i++)
	{
		movement.M[i].logEnabled = 0;
		movement.M[i].maxCurrent = 8000;	// 10A
		movement.M[i].overCurrentCounter = 0;
		movement.M[i].lastEncCount = 0;
		movement.M[i].lastReaction = 0;
	}

	movement.M[4].maxCurrent = 3000;	// 3A

	movement.logEnabled = 0;

	for(uint8_t i = 0; i < 4; i++)
	{
		movement.latest[i] = 0;
		movement.target[i] = 0;
	}

	PIDInit(&movement.M[0].pid, &PidOutputM1, &PidFeedbackM1);
	PIDInit(&movement.M[1].pid, &PidOutputM2, &PidFeedbackM2);
	PIDInit(&movement.M[2].pid, &PidOutputM3, &PidFeedbackM3);
	PIDInit(&movement.M[3].pid, &PidOutputM4, &PidFeedbackM4);
	PIDInit(&movement.M[4].pid, &PidOutputM5, &PidFeedbackM5);
	PIDInit(&movement.rotCompPid, &PidOutputRotComp, &PidFeedbackRotComp);

	PIDSetConfiguration(&movement.M[0].pid, 1000, 1000, 0, 100, MOVE_PIDMAX, 0);
	PIDSetConfiguration(&movement.M[1].pid, 1000, 1000, 0, 100, MOVE_PIDMAX, 0);
	PIDSetConfiguration(&movement.M[2].pid, 1000, 1000, 0, 100, MOVE_PIDMAX, 0);
	PIDSetConfiguration(&movement.M[3].pid, 1000, 1000, 0, 100, MOVE_PIDMAX, 0);
	PIDSetConfiguration(&movement.M[4].pid, 10000, 5000, 0, 500, MOVE_DRIBMAX, 0);
	PIDSetConfiguration(&movement.rotCompPid, 20000, 10000, 0, 1000, 10000, 0);
}

void MovementSetFeedbackLog(uint8_t enable)
{
	movement.logEnabled = enable;
}

void MovementSetManual(uint8_t id, int16_t manPower)
{
	if(movement.mode != MOTOR_MODE_MANUAL)
		return;

	if(manPower < 0)
	{
		MotorSetDir(id, 1);
		MotorSetPower(id, -manPower);
	}
	else
	{
		MotorSetDir(id, 0);
		MotorSetPower(id, manPower);
	}
}

void MovementSetPidSetpoint(uint8_t id, int32_t sp)
{
	if(movement.mode != MOTOR_MODE_PID)
		return;

	movement.M[id].lastReaction = xTaskGetTickCount();

	PIDSetSetpoint(&movement.M[id].pid, sp);
}

void setAutomatic(int32_t vx, int32_t vy, int32_t vw)
{
	// vx, vy unit = [e-3] = [mm/s]
	// vw unit = [mrad]

	int32_t vm[4];

	vm[0] = divRound(-vx*MOVE_SIN_A + vy*MOVE_COS_A + vw*MOVE_BOT_RAD, 100000);
	vm[1] = divRound(-vx*MOVE_SIN_A - vy*MOVE_COS_A + vw*MOVE_BOT_RAD, 100000);
	vm[2] = divRound( vx*MOVE_SIN_B - vy*MOVE_COS_B + vw*MOVE_BOT_RAD, 100000);
	vm[3] = divRound( vx*MOVE_SIN_B + vy*MOVE_COS_B + vw*MOVE_BOT_RAD, 100000);

	//vm unit = [e-3] = [mm/s], speed over ground

	for(uint8_t i = 0; i < 4; i++)
	{
		PIDSetSetpoint(&movement.M[i].pid, divRound(vm[i] * MOVE_VM2PID, 100000));
	}
}

void MovementSetAutomatic(int32_t vx, int32_t vy, int32_t vw, int32_t vv)
{
	if(movement.mode == MOTOR_MODE_AUTOMATIC)
	{
		movement.target[MOVEMENT_AXIS_X] = vx;
		movement.target[MOVEMENT_AXIS_Y] = vy;
		movement.target[MOVEMENT_AXIS_W] = vw;
		movement.target[MOVEMENT_AXIS_V] = vv;

		movement.wDiv.vPart = 1000 - ((vw*1000)/(vw + vv));
		movement.wDiv.wPart = 1000 - ((vv*1000)/(vw + vv));

		for(uint8_t i = 0; i < 5; i++)
			movement.M[i].lastReaction = xTaskGetTickCount();
	}
}

void MovementSetMotorPidParams(uint8_t id, int32_t Kp, int32_t Ki, int32_t Kd, int32_t slew)
{
	movement.M[id].pid.Kp = Kp;
	movement.M[id].pid.Ki = Ki;
	movement.M[id].pid.Kd = Kd;
	movement.M[id].pid.slewMax = slew;
}

void MovementSetMode(uint8_t mode)
{
	if(mode > 3)
		return;

	movement.mode = mode;
}

void MovementSetLog(uint8_t id, uint8_t enable)
{
	movement.M[id].logEnabled = enable;
}

void MovementSetDribble(int32_t rpm)
{
	movement.M[4].lastReaction = xTaskGetTickCount();

	PIDSetSetpoint(&movement.M[4].pid, (rpm*MOVE_DRIB2PID)/1000);
}

void MovementTask(void* params)
{
	params = params;

	portTickType xLastWakeTime = xTaskGetTickCount();
	portTickType xLastLogTime = xLastWakeTime;
	portTickType xLastStatusTime = xLastWakeTime;
	portTickType xLastDribblerTime = xLastWakeTime;
	portTickType xLastATmegaTime = xLastWakeTime;

	int32_t alphaReal = 0;
	int32_t alphaUsed = 0;
	int32_t alphaPos  = 0;

	while(1)
	{
		// calculate speed
		int32_t vm[5];

		for(uint8_t i = 0; i < 4; i++)
		{
			vm[i] = 0;

			vm[i] = divRound((movement.M[i].pid.log.current * 100000),  MOVE_VM2PID);

			//vm unit = [mm/s]
		}

		movement.latest[MOVEMENT_AXIS_X] = divRound(MOVE_R_VX * (-vm[0]-vm[1]+vm[2]+vm[3]), 100000);
		movement.latest[MOVEMENT_AXIS_Y] = divRound(MOVE_R_VY1*vm[0] - MOVE_R_VY1*vm[1] - MOVE_R_VY2*vm[2] + MOVE_R_VY2*vm[3], 100000);
		int32_t rotSum = divRound(MOVE_R_VW1*vm[0] + MOVE_R_VW1*vm[1] + MOVE_R_VW2*vm[2] + MOVE_R_VW2*vm[3], MOVE_BOT_RAD);
		// rotSum = [mrad]

		movement.latest[MOVEMENT_AXIS_W] = (rotSum*movement.wDiv.wPart)/1000;
		movement.latest[MOVEMENT_AXIS_V] = (rotSum*movement.wDiv.vPart)/1000;

		// rotate local coordinate system
		int32_t lastAlphaUsed = alphaUsed;
		int32_t lastAlphaPos = alphaPos;
		alphaReal = divRound(movement.latest[MOVEMENT_AXIS_V]*10L, 100);

		alphaPos = lastAlphaPos + alphaReal + lastAlphaUsed;
		movement.rotCompFeedback = alphaPos;
		PIDUpdate(&movement.rotCompPid);
		alphaUsed = movement.rotCompOut;

		int32_t cosA = divRound(fastcos(alphaUsed*100), 100);
		int32_t sinA = divRound(fastsin(alphaUsed*100), 100);
		int32_t xNew = divRound((movement.target[MOVEMENT_AXIS_X] * cosA - movement.target[MOVEMENT_AXIS_Y] * sinA), 10000);
		int32_t yNew = divRound((movement.target[MOVEMENT_AXIS_Y] * cosA + movement.target[MOVEMENT_AXIS_X] * sinA), 10000);

		movement.target[MOVEMENT_AXIS_X] = xNew;
		movement.target[MOVEMENT_AXIS_Y] = yNew;

		if(movement.mode == MOTOR_MODE_AUTOMATIC)
		{
			setAutomatic(xNew, yNew, movement.target[MOVEMENT_AXIS_V] + movement.target[MOVEMENT_AXIS_W]);
		}

		// update motor (PIDs)
		if(movement.mode != MOTOR_MODE_MANUAL)
		{
			for(uint8_t i = 0; i < 4; i++)
			{
				PIDUpdate(&movement.M[i].pid);
			}
		}

		// overcurrent protection
		for(uint8_t i = 0; i < 4; i++)
		{
			if(movement.atData.current[i] > movement.M[i].maxCurrent)
			{
				movement.M[i].overCurrentCounter++;
			}
			else
			{
				movement.M[i].overCurrentCounter = 0;
			}

			if(movement.M[i].overCurrentCounter > 50)
			{
				MotorSetPower(i, 0);
				PIDSetSetpoint(&movement.M[i].pid, 0);
			}
		}

		// Dribbler update (slower PID update - 20Hz)
		if(xLastWakeTime - xLastDribblerTime > 50/portTICK_RATE_MS)
		{
			xLastDribblerTime = xLastWakeTime;

			if(movement.mode != MOTOR_MODE_MANUAL)
			{
				PIDUpdate(&movement.M[4].pid);
			}

			if(movement.atData.current[4] > movement.M[4].maxCurrent)
			{
				movement.M[4].overCurrentCounter++;
			}
			else
			{
				movement.M[4].overCurrentCounter = 0;
			}

			if(movement.M[4].overCurrentCounter > 5)
			{
				MotorSetPower(4, 0);
				PIDSetSetpoint(&movement.M[4].pid, 0);
			}
		}

		// stall protection
		for(uint8_t i = 0; i < 5; i++)
		{
			if(movement.M[i].pid.log.current < -MOVE_REACTION_THRESHOLD || movement.M[i].pid.log.current > MOVE_REACTION_THRESHOLD)
			{
				movement.M[i].lastReaction = xLastWakeTime;
			}

			if(xLastWakeTime - movement.M[i].lastReaction > MOVE_STALL_TIMEOUT/portTICK_RATE_MS)
			{
				// last reaction was more than 5s ago
				if(movement.M[i].pid.lastOutput != 0)
				{
					// commanded to turn, but nothing happens -> Stop wheel
					PIDSetZero(&movement.M[i].pid);

					movement.M[i].lastReaction = xLastWakeTime;
				}
			}
		}

		if(xLastWakeTime - xLastLogTime > 50/portTICK_RATE_MS)
		{
			xLastLogTime = xLastWakeTime;

			for(uint8_t i = 0; i < 5; i++)
			{
				if(!movement.M[i].logEnabled)
					continue;

				MotorPidLog log;
				log.latest = movement.M[i].pid.log.current;
				log.setpoint = movement.M[i].pid.setpoint;
				log.pError = movement.M[i].pid.log.pError;
				log.iError = movement.M[i].pid.log.iError;
				log.dError = movement.M[i].pid.log.dError;
				log.id = i;
				log.output = movement.M[i].pid.log.output;
				log.overload = movement.M[i].pid.log.overload;
				log.eCurrent = movement.atData.current[i];

				PacketHeader header;
				header.section = SECTION_MOTOR;
				header.cmd = CMD_MOTOR_PID_LOG;
				header.len = LEN_MOTOR_PID_LOG;

				NetworkSendPacket(&header, (uint8_t*)&log, 0);
			}
		}

		if(xLastWakeTime - xLastStatusTime > 50/portTICK_RATE_MS && movement.logEnabled)
		{
			xLastStatusTime = xLastWakeTime;

			SystemStatusMovement status;
			status.vx = movement.latest[MOVEMENT_AXIS_X];
			status.vy = movement.latest[MOVEMENT_AXIS_Y];
			status.vw = movement.latest[MOVEMENT_AXIS_V];

			PacketHeader header;
			header.section = SECTION_SYSTEM;
			header.cmd = CMD_SYSTEM_STATUS_MOVEMENT;
			header.len = LEN_SYSTEM_STATUS_MOVEMENT;

			NetworkSendPacket(&header, (uint8_t*)&status, 0);
		}

		if(xLastWakeTime - xLastATmegaTime > 50/portTICK_RATE_MS)
		{
			xLastATmegaTime = xLastWakeTime;

			ATMegaExchange(&movement.atmega, movement.atData.exchange.mosi, movement.atData.exchange.miso, MOTOR_DATA_SIZE);

			for(uint8_t i = 0; i < 5; i++)
			{
				if(movement.atData.fault & (1 << i))
				{
					MotorClearDriverFault(i);
					PIDSetZero(&movement.M[i].pid);
				}
			}
		}

		vTaskDelayUntil(&xLastWakeTime, 10/portTICK_RATE_MS);
	}
}
