/*
 * console.c
 *
 *  Created on: 25.05.2011
 *      Author: AndreR
 */

#include "console.h"

#include "util/printf.h"

#include <string.h>
#include <stdlib.h>
#include <stdarg.h>

#include "util/mem.h"
#include "util/critical.h"

Console console;

uint8_t verifyCmdReady()
{
	if(console.cmdReady)
		return 1;

	TASK_MUTEX2(console.hFifoInUse, console.hCmdDataInUse)
	{
		for(uint8_t i = 0; i < console.usedInterfaces; i++)
		{
			if(FifoPacketSyncedNumPackets(console.rxFifos[i]) > 0)
			{
				uint16_t length;
				FifoPacketSyncedPop(console.rxFifos[i], &length, console.cmdData, CONSOLE_CMD_SIZE-1);
				console.cmdData[length] = '\0';
				console.cmdReady = 1;
				break;
			}
		}
	}

	return console.cmdReady;
}

void ConsoleInit()
{
	console.usedInterfaces = 0;
	console.cmdReady = 0;

	console.hFifoInUse = xSemaphoreCreateMutex();
	console.hPrintDataInUse = xSemaphoreCreateMutex();
	console.hCmdDataInUse = xSemaphoreCreateMutex();
}

int16_t ConsoleAddInterface(FifoPacketSynced* pRxFifo, FifoSynced* pTxFifo, NotifyTxReady readyCb)
{
	if(console.usedInterfaces == CONSOLE_MAX_INTERFACES)
		return ERROR_CONSOLE_NO_FREE_SLOT;

	console.observer[console.usedInterfaces] = readyCb;
	console.rxFifos[console.usedInterfaces] = pRxFifo;
	console.txFifos[console.usedInterfaces] = pTxFifo;

	console.usedInterfaces++;

	return 0;
}

uint16_t ConsoleNumCmdsReady()
{
	uint16_t numCmds = 0;

	TASK_MUTEX(console.hFifoInUse)
	{
		for(uint8_t i = 0; i < console.usedInterfaces; i++)
		{
			numCmds += FifoPacketSyncedNumPackets(console.rxFifos[i]);
		}
	}

	if(console.cmdReady)
		numCmds++;

	return numCmds;
}

uint8_t ConsoleIsCmdReady()
{
	if(console.cmdReady)
		return 1;

	return ConsoleNumCmdsReady() > 0 ?  1: 0;
}

int16_t ConsoleCmdProcessed()
{
	console.cmdReady = 0;

	return 0;
}

int16_t ConsoleCmpCmd(const char* cmd)
{
	if(verifyCmdReady() == 0)
		return 0;

	int16_t result = 0;

	TASK_MUTEX(console.hCmdDataInUse)
	{
		if(strcmp((char*)console.cmdData, cmd) == 0)
			result = 1;
	}

	return result;
}

int16_t ConsoleCmpCmdI16(const char* cmp, uint8_t argc, int16_t* argv)
{
	if(verifyCmdReady() == 0)
		return 0;

	TASK_MUTEX(console.hCmdDataInUse)
	{
		uint8_t* cmd = console.cmdData;

		while(1)
		{
			if(*cmp == '\0')
				break;	//cmp end

			if(*cmd == '\0')
				return 0;	//reached data end

			if(*cmp != *cmd)
				return 0;	//compare mismatch

			cmd++;
			cmp++;
		}

		//UART might have more data, d is not 0, i is current index
		for(uint8_t j = 0; j < argc; j++)
		{
			//seek next space
			while(*cmd != ' ')
			{
				if(*cmd == '\0')
				{
					if(j == argc)	//this was the last expected parameter
						return 1;
					else
						return 0;	//missing parameter
				}

				cmd++;
			}

			argv[j] = atoi((char*)cmd);

			cmd++;	//skip whitespace
		}
	}

	return 1;
}

int16_t ConsoleCmpCmdI32(const char* cmp, uint8_t argc, int32_t* argv)
{
	if(verifyCmdReady() == 0)
		return 0;

	TASK_MUTEX(console.hCmdDataInUse)
	{
		uint8_t* cmd = console.cmdData;

		while(1)
		{
			if(*cmp == '\0')
				break;	//cmp end

			if(*cmd == '\0')
				return 0;	//reached data end

			if(*cmp != *cmd)
				return 0;	//compare mismatch

			cmd++;
			cmp++;
		}

		//UART might have more data, d is not 0, i is current index
		for(uint8_t j = 0; j < argc; j++)
		{
			//seek next space
			while(*cmd != ' ')
			{
				if(*cmd == '\0')
				{
					if(j == argc)	//this was the last expected parameter
						return 1;
					else
						xSemaphoreGive(console.hFifoInUse);
				}

				cmd++;
			}

			argv[j] = atoi((char*)cmd);

			cmd++;	//skip whitespace
		}
	}

	return 1;
}

int16_t ConsolePrint(const char* fmt, ...)
{
	va_list args;

	TASK_MUTEX2(console.hPrintDataInUse, console.hFifoInUse)
	{
		va_start(args, fmt);
		xvsnprintf(console.printData, CONSOLE_PRINT_SIZE, fmt, args);
		va_end(args);

		for(uint8_t i = 0; i < console.usedInterfaces; i++)
		{
			//transfer parse buffer to send buffer
			char* pRead = console.printData;
			while(*pRead != '\0')
			{
				FifoSyncedPushByte(console.txFifos[i], *pRead);
				pRead++;
			}

			(*console.observer[i])();
		}
	}

	return 0;
}

int16_t ConsolePut(uint8_t* data, uint16_t length)
{
	TASK_MUTEX(console.hFifoInUse)
	{
		for(uint8_t i = 0; i < console.usedInterfaces; i++)
		{
			FifoSyncedPush(console.txFifos[i], length, data);

			(*console.observer[i])();
		}
	}

	return 0;
}

int16_t ConsoleGetCmd(uint8_t* pDst, uint16_t bufLen, uint16_t* pBytesRead)
{
	if(verifyCmdReady() == 0)
		return ERROR_CONSOLE_NO_CMD_READY;

	uint16_t len = 0;

	TASK_MUTEX(console.hCmdDataInUse)
	{
		len = strlen((char*)console.cmdData);
		if(len > bufLen)
			len = bufLen;

		MemCpy(console.cmdData, pDst, len);
	}

	*pBytesRead = len;

	return 0;
}
