package edu.dhbw.mannheim.tigers.sim.engine;


import java.io.IOException;
import java.util.concurrent.Callable;
import java.util.concurrent.Future;

import org.apache.log4j.Logger;

import com.jme.image.Image;
import com.jme.input.InputHandler;
import com.jme.input.KeyInput;
import com.jme.input.MouseInput;
import com.jme.input.joystick.JoystickInput;
import com.jme.light.DirectionalLight;
import com.jme.math.Vector3f;
import com.jme.renderer.Camera;
import com.jme.renderer.ColorRGBA;
import com.jme.renderer.RenderContext;
import com.jme.renderer.Renderer;
import com.jme.scene.Node;
import com.jme.scene.Text;
import com.jme.scene.state.CullState;
import com.jme.scene.state.CullState.Face;
import com.jme.scene.state.LightState;
import com.jme.scene.state.ZBufferState;
import com.jme.system.DisplaySystem;
import com.jme.system.GameSettings;
import com.jme.system.JmeException;
import com.jme.util.GameTaskQueue;
import com.jme.util.TextureManager;
import com.jmex.audio.AudioSystem;
import com.jmex.physics.PhysicsSpace;

import edu.dhbw.mannheim.tigers.sim.Sim;
import edu.dhbw.mannheim.tigers.sim.engine.modes.ASimEngineMode;
import edu.dhbw.mannheim.tigers.sim.engine.modes.ESimEngineMode;
import edu.dhbw.mannheim.tigers.sim.engine.modes.display.DisplayMode;
import edu.dhbw.mannheim.tigers.sim.engine.modes.match.MatchMode;
import edu.dhbw.mannheim.tigers.sim.engine.modes.test.TestMode;
import edu.dhbw.mannheim.tigers.sim.engine.observer.ISimEngineObservable;
import edu.dhbw.mannheim.tigers.sim.engine.observer.ISimEngineObserver;
import edu.dhbw.mannheim.tigers.sim.engine.observer.SimEngineObservable;
import edu.dhbw.mannheim.tigers.sim.exceptions.SimResourceException;
import edu.dhbw.mannheim.tigers.sim.resources.ImageProvider;
import edu.dhbw.mannheim.tigers.sim.resources.ImageProvider.EImage;
import edu.dhbw.mannheim.tigers.sim.settings.ISettingsProvider;
import edu.dhbw.mannheim.tigers.sim.settings.data.ISimEngineProperties;
import edu.dhbw.mannheim.tigers.sim.settings.data.ISimProperties;


/**
 * This is the core of the whole program. It handles the whole simulation, is accessible from the gui and manages its
 * inner parts like {@link ASimEngineState}s, {@link ASimEngineMode}s etc.
 * 
 * @author Gero
 */
public class SimEngine extends ACycleControlledEngine implements ISimEngine, ICore {
	
	// --------------------------------------------------------------------------
	// --- variables and constants ----------------------------------------------
	// --------------------------------------------------------------------------
	// Visualization ###################################
	// Window
	public static final String				WINDOW_TITLE	= "CAGE Core - " + Sim.VERSION_TAG;
	
	private int									width, height, depth, freq, minSamples;
//	private int									posX, posY;
	private boolean							fullscreen, displayCreated;
	
	/** time-per-(last)cycle (updated in {@link #update(float)} */
	private float								lastTpc;
//	/** (last) cycles-per-second (updated in {@link #update(float)} */
//	private float								lastCps;
	
	// Camera
	private Camera								cam;
	
	// Scene
	/** The root-node of the scene */
	private Node								root;
	
	/** By convenience the root node for all the static environment */
	private Node								statics;
	
	/** By convenience the root node for all dynamic nodes in the scene */
	private Node								dynamics;
	/** By convenience the root node for all balls in the game */
	private Node								balls;
	/** By convenience the root node for all robots in the game */
	private Node								robots;
	
	/** By convenience the root node for all UI elements */
	private Node								hud;
	
	/** By convenience the root node for all visualization elements */
	private Node								vis;
	

	// Input ###################################
	private InputHandler						input;
	

	// Simulation ###################################
	private PhysicsSpace						space;
	

	// System ###################################
	private final Logger						log				= Logger.getLogger(getClass());
	
	private ISimEngineProperties			engineProperties;
//	private IApplicationProperties		appProperties;
	private SimEngineSettings				newSettings;
	
	private Thread								jmeSimLoop;
	
	private boolean							debug				= false;
	
	// private boolean onCanvas = false;
	
	// Tasks ###################################
	protected final TaskQueue				updateQueue		= new TaskQueue();
	protected final GameTaskQueue			renderQueue		= new GameTaskQueue();
	
	// Mode ###################################
	private ASimEngineMode					mode;
	

	// States ###################################
	private ESimEngineState					state;
	private final ISimEngineObservable	stateObservable;
	
	
	private void changeState(ESimEngineState newState) {
		
		if (state == null || !state.equals(newState)) {
			final ESimEngineState oldState = state;
			state = newState;
			
			switch (newState) {
				case STOPPED:
					currentRenderState = renderStateExceptPaused;
					finish();
					break;
				
				case PAUSED_FULL:
					currentRenderState = renderStatePaused;
					break;
				
				case PAUSED_SIMULATION:
					currentRenderState = renderStateExceptPaused;
					break;
				
				case RUNNING:
					currentRenderState = renderStateExceptPaused;
					finished = false;
					break;
			}
			
			mode.changeStateTo(newState);
			
			notifyObservers(newState);
			
			log.info("Changed state from <" + oldState + "> to <" + newState + ">");
		}
	}
	
	// Render state
	private ASimEngineState			currentRenderState;
	
	/* Does nothing */
	private final ASimEngineState	renderStatePaused			= new ASimEngineState(ESimEngineState.PAUSED_FULL) {
	};
	
	private final ASimEngineState	renderStateExceptPaused	= new ASimEngineState(
			ESimEngineState.getStatesExceptPausedFull()) {
		
		@Override
		public void render(float interpolation) {
			// Render
			display.getRenderer().clearBuffers();
			display.getRenderer().draw(root);
			
			// Debug
			if (debug) {
				mode.debug(interpolation);
			}
			
			// Let mode render whatever it wants!
			mode.render(interpolation);
		};
	};
	
	private final Object				cleanupSync					= new Object();
	private boolean					cleanupDone					= false;
	
	
	@Override
	public void waitForCleanup() {
		
		synchronized (cleanupSync) {
			try {
				if (!cleanupDone) {
					
					log.debug("Waiting for SimEngine to finish cleanup...");
					cleanupSync.wait();
				}
				log.debug("Cleanup confirmed, proceeding...");
				
			} catch (InterruptedException err) {
				log.error("Interrupted while waiting for cleanup!", err);
			}
		}
	}
	

	// START ###################################
	// --------------------------------------------------------------------------
	// --- constructors ---------------------------------------------------------
	// --------------------------------------------------------------------------
	/**
	 * @see SimEngine
	 * @param modeType
	 * @param newSettings May be null, then default settings are taken.
	 */
	public SimEngine(ESimEngineMode modeType, SimEngineSettings newSettings) {
		super();
		
		// Initialization
		this.updateQueue.setExecuteAll(true);
		this.renderQueue.setExecuteAll(true);
		
		// Settings
		this.newSettings = newSettings;
		
		getNewSettings();
		
		loadSettings();
		
		// Mode
		createMode(modeType);
		
		stateObservable = new SimEngineObservable(ESimEngineState.READY);
		changeState(ESimEngineState.READY);
	}
	

	private void loadSettings() {
		// Get window settings
		this.width = settings.getWidth();
		this.height = settings.getHeight();
		this.depth = settings.getDepth();
		this.freq = settings.getFrequency();
		this.fullscreen = settings.isFullscreen();
		
		// Special engine settings
		ISimProperties simProps = ISettingsProvider.INSTANCE.getSimProperties();
		engineProperties = simProps.getSimEngineProperties();
//		appProperties = simProps.getApplicationProperties();
	}
	

	private void createMode(ESimEngineMode type) {
		ASimEngineMode result = null;
		
		switch (type) {
			
			// case SHOOTER:
			// result = new ShooterMode(this);
			// break;
			
			case DISPLAY:
				result = new DisplayMode(this);
				break;
			
			case MATCH:
				result = new MatchMode(this);
				break;
			
			default:
			case TEST:
				result = new TestMode(this);
		}
		
		this.mode = result;
	}
	

	// --------------------------------------------------------------------------
	// --- start ----------------------------------------------------------------
	// --------------------------------------------------------------------------
	@Override
	public void start() {
		changeState(ESimEngineState.RUNNING);
		
		if (jmeSimLoop != null) {
			jmeSimLoop.interrupt();
		}
		
		jmeSimLoop = null;
		jmeSimLoop = new Thread(new Runnable() {
			
			@Override
			public void run() {
				SimEngine.super.start();
			}
		}, "jmeSimLoop");
		jmeSimLoop.setPriority(Thread.MAX_PRIORITY);
		jmeSimLoop.start();
	}
	

	// --------------------------------------------------------------------------
	// --- initialize system ---------------------------------------------------------
	// --------------------------------------------------------------------------
	@Override
	protected void initSystem() {
		// Settings
		int cps = engineProperties.getCPS();
		int fps = engineProperties.getFPS();
//		int posX = appProperties.getRenderWindowPosX();
//		int posY = appProperties.getRenderWindowPosY();
		
		setCyclesPerSecond(cps != 0 ? cps : DEFAULT_CPS);
		setFramesPerSecond(fps != 0 ? fps : DEFAULT_FPS);
		
		// Scene
		root = new Node("root");
		root.attachChild(statics = new Node("statics"));
		root.attachChild(dynamics = new Node("dynamics"));
		dynamics.attachChild(balls = new Node("balls"));
		dynamics.attachChild(robots = new Node("robots"));
		root.attachChild(hud = new Node("hud"));
		root.attachChild(vis = new Node("vis"));
		
		// Display creation
		display = DisplaySystem.getDisplaySystem(getNewSettings().getRenderer());
		
		displayCreated = false;
		display.setMinSamples(minSamples = getNewSettings().getSamples());
		
		// Check whether a display can be created correctly via catching possibly thrown exceptions
		while (!displayCreated) {
			try {
				display.createWindow(width, height, depth, freq, fullscreen);
				displayCreated = true;
			} catch (JmeException e) {
				int oldSamples = minSamples;
				
				if (minSamples > 16) {
					// Reduce to 16!
					minSamples = 16;
				} else {
					// Reduce from 16 to 0
					if (minSamples == 0) {
						// If minSamples are already '0', then any other error caused the JmeException
						log.warn("Unable to create window with your settings! (AntiAliasing: 0)");
						throw e;
					}
					
					minSamples -= minSamples / 2;
					if (minSamples == 1) {
						minSamples--;
					} else if (minSamples % 2 != 0) {
						minSamples++;
					}
				}
				
				log.warn("Some settings-combination is not supported. Trying to reduce AntiAliasing-rate from ("
						+ oldSamples + ") to " + minSamples + "!");
				
				display.setMinSamples(minSamples);
			}
		}
		
//		// Move to last position
//		if (checkScreenPosition(posX, posY)) {
//			display.moveWindowTo(posX, posY);
//		}
		
		getNewSettings().setSamples(minSamples);
		
		// Icon
		try {
			final Image[] imgs = new Image[2];
			imgs[0] = ImageProvider.getInstance().getImageJME(EImage.TIGER_ICON_16x16);
			imgs[1] = ImageProvider.getInstance().getImageJME(EImage.TIGER_ICON_32x32);
			
			display.setIcon(imgs);
		} catch (SimResourceException err) {
			log.warn("Could not set window icon!");
		}
		
		display.setTitle(WINDOW_TITLE);
		// display.setVSyncEnabled(true); // WARNING: VERY expensive!
		
		// Set background color. Very important. Imagine, someone might set it to... pink O.o
		display.getRenderer().setBackgroundColor(ColorRGBA.black);

		// #### MODE
		mode.initSystem();
	}
	
	
//	/**
//	 * @param x
//	 * @param y
//	 * @return <code>True</code> if the given position is visible on at least one screen
//	 */
//	private boolean checkScreenPosition(int x, int y) {
//		GraphicsEnvironment ge = GraphicsEnvironment.getLocalGraphicsEnvironment();
//
//		// Get size of each screen
//		for (GraphicsDevice gd : ge.getScreenDevices()) {
//		    java.awt.DisplayMode dm = gd.getDisplayMode();
//		    if (x + 10 < dm.getWidth() && y + 10 < dm.getHeight()) {
//		   	 return true; // At least a small corner is visible
//		    }
//		}
//
//		return false;
//	}
	

	@Override
	protected void reinit() {
		display.recreateWindow(width, height, depth, freq, fullscreen);
	}
	

	// --------------------------------------------------------------------------
	// --- initialize simulation ------------------------------------------------
	// --------------------------------------------------------------------------
	@Override
	protected void initGame() {
		// Z-Buffer: Nodes are rendered in the right order
		ZBufferState buf = display.getRenderer().createZBufferState();
		buf.setEnabled(true);
		buf.setFunction(ZBufferState.TestFunction.LessThanOrEqualTo);
		root.setRenderState(buf);
		
		// Optimization: Only nodes faces turned to the camera are rendered
		CullState cs = display.getRenderer().createCullState();
		cs.setCullFace(Face.Back);
		root.setRenderState(cs);
		
		// #### MODE
		mode.initSimulation();
		
		// Light
		LightState ls = display.getRenderer().createLightState();
		ls.setEnabled(true);
		addLight(ls, new Vector3f(0.5f, -1, 0.5f));
		addLight(ls, new Vector3f(-0.5f, -1, 0.5f));
		addLight(ls, new Vector3f(0.5f, -1, -0.5f));
		addLight(ls, new Vector3f(-0.5f, -1, -0.5f));
		root.setRenderState(ls);
		
		// update the scene graph for rendering
		root.updateGeometricState(0.0f, true);
		root.updateRenderState();
	}
	

	private void addLight(LightState ls, Vector3f lightDirection) {
		final DirectionalLight light = new DirectionalLight();
		light.setAttenuate(false);
		light.setDiffuse(new ColorRGBA(0.5f, 0.5f, 0.5f, 0.5f));
		light.setAmbient(new ColorRGBA(0.1f, 0.1f, 0.1f, 0));
		light.setEnabled(true);
		light.setDirection(lightDirection);
		ls.attach(light);
	}
	

	// --------------------------------------------------------------------------
	// --- render scene ---------------------------------------------------------
	// --------------------------------------------------------------------------
	@Override
	protected void render(float interpolation) {
		// Consumes every present render task...
		renderQueue.execute();
		
		// ..and the mode!
		currentRenderState.render(interpolation);
	}
	
	@Override
	protected void afterRenderCallback(float interpolation) {
		mode.afterRenderCallback(lastTpc, display.getRenderer());
	}
	

	// --------------------------------------------------------------------------
	// --- update scene ---------------------------------------------------------
	// --------------------------------------------------------------------------
	@Override
	protected void update(float interpolation) {
		final float tpc = timer.getTimePerFrame();
		final float cps = timer.getFrameRate();
		lastTpc = tpc;
		
		// Consumes every present task...
		updateQueue.execute();
		

		// ##### MODE
		mode.update(interpolation, tpc, cps);
		
		// Update the scene due to our changes to it
		root.updateGeometricState(tpc, true);
	}
	

	// --------------------------------------------------------------------------
	// --- cleanup ---------------------------------------------------------
	// --------------------------------------------------------------------------
	@Override
	protected void cleanup() {
		while (!isLastCycleComplete()) {
			sleepWhileCleanup(50);
		}
		
		synchronized (cleanupSync) {

			if (cleanupDone) {
				cleanupSync.notifyAll();
				return;
			}
			
			// ##### MODE
			mode.cleanup();
			
			// TODO Find a way to determine the damn position of this FUCKING RENDERING WINDOW!!!
//			// Save render window position data
//			if (appProperties != null && display != null) {
//				if (DisplaySystem.getSystemProvider().getProviderIdentifier().equals(LWJGLSystemProvider.LWJGL_SYSTEM_IDENTIFIER)) {
//					display.get
////					appProperties.setRenderWindowPosX(loc.x);
////					appProperties.setRenderWindowPosX(loc.y);
//				}
//			}
			
			// Save engine settings
			try {
				getNewSettings().save();
			} catch (IOException err) {
				log.warn("Couldn't persist engine settings!");
			}
			
			TextureManager.doTextureCleanup();
			TextureManager.clearCache();
			
			Text.resetDefaultFontTextureState();
			

			KeyInput.destroyIfInitalized();
			MouseInput.destroyIfInitalized();
			JoystickInput.destroyIfInitalized();
			
			if (AudioSystem.isCreated()) {
				AudioSystem.getSystem().cleanup();
			}
			
			// Visualization
			if (display != null) {
				
				Renderer renderer = display.getRenderer();
				if (renderer != null) {
					try {
						renderer.finish();
						
						// TODO Gero: Is the following necessary? Don't really know... :-/
						renderer.clearBuffers();
						renderer.clearColorBuffer();
						renderer.clearPolygonOffset();
						renderer.clearQueue();
						renderer.clearStencilBuffer();
						renderer.clearStrictBuffers();
						renderer.clearVBOCache();
						renderer.clearZBuffer();
						// ----
						
						renderer.cleanup();
					} catch (Exception exc) {
						log.debug("Exception ocurred while cleaning up renderer.");
					}
				}
				
				RenderContext<?> context = display.getCurrentContext();
				if (context != null) {
					context.invalidateStates();
				}
				
				display.reset();
				display.close();
			}
			

			cam = null;
			
			// Other
			currentRenderState = null;
			
			// Scene
			if (root != null) {
				root.getChildren().clear();
				root = null;
			}
			
			dynamics = null;
			
			statics = null;
			
			vis = null;
			

			// Confirm cleanup
			cleanupDone = true;
			cleanupSync.notifyAll();
		}
	}
	

	private void sleepWhileCleanup(long timeToSleep) {
		try {
			Thread.sleep(timeToSleep);
			log.info("Waiting for last cycle to complete...");
		} catch (InterruptedException exc) {
			log.error("Error while waiting for last cycle to complete!", exc);
		}
	}
	

	// --------------------------------------------------------------------------
	// --- load settings --------------------------------------------------------
	// --------------------------------------------------------------------------
	@Override
	protected GameSettings getNewSettings() {
		
		// if (newSettings == null) {
		// try {
		// newSettings = ISettingsProvider.INSTANCE.getSimProperties().get;
		// } catch (SimSettingsException exc) {
		// log.error(exc);
		// }
		// }
		
		if (newSettings == null) {
			newSettings = SimEngineSettings.getInstance();
		}
		
		return newSettings;
	}
	

	// --------------------------------------------------------------------------
	// --- observer ---------------------------------------------------------
	// --------------------------------------------------------------------------
	@Override
	public boolean addObservers(ISimEngineObserver... observers) {
		return stateObservable.addObservers(observers);
	}
	

	@Override
	public void notifyObservers(ESimEngineState event) {
		stateObservable.notifyObservers(event);
	}
	

	@Override
	public void notifyObservers() {
		stateObservable.notifyObservers();
	}
	

	@Override
	public boolean removeObserver(ISimEngineObserver observer) {
		return stateObservable.removeObserver(observer);
	}
	

	/*
	 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	 * GameTaskQueue
	 */
	@Override
	public <R> Future<R> enqueueUpdateTask(Callable<R> task) {
		return updateQueue.enqueue(task);
	}
	
	
	@Override
	public void shutdownUpdateQueue() {
		updateQueue.cancelQueue();
	}
	

	@Override
	public <R> Future<R> enqueueRenderTask(Callable<R> task) {
		return renderQueue.enqueue(task);
	}
	

	// --------------------------------------------------------------------------
	// --- ISimEngine -----------------------------------------------------------
	// --------------------------------------------------------------------------
	@Override
	protected void quit() {
		changeState(ESimEngineState.STOPPED);
		cleanup();
	}
	

	@Override
	public void shutdown() {
		changeState(ESimEngineState.STOPPED);
		cleanup();
	}
	

	@Override
	public void pause() {
		changeState(ESimEngineState.PAUSED_FULL);
	}
	

	@Override
	public void pauseSimulation() {
		changeState(ESimEngineState.PAUSED_SIMULATION);
	}
	

	@Override
	public void resume() {
		changeState(ESimEngineState.RUNNING);
	}
	

	@Override
	public void stop() {
		changeState(ESimEngineState.STOPPED);
	}
	

	@Override
	protected void handleError(Throwable t) {
		changeState(ESimEngineState.STOPPED);
	}
	

	@Override
	public ASimEngineMode getMode() {
		return mode;
	}
	

	@Override
	public ESimEngineState getState() {
		return state;
	}
	

	// --------------------------------------------------------------------------
	// --- ICore ----------------------------------------------------------------
	// --------------------------------------------------------------------------
	@Override
	public Camera getCamera() {
		return cam;
	}
	

	@Override
	public void setCam(Camera newCam) {
		cam = newCam;
	}
	

	@Override
	public PhysicsSpace getPhysicsSpace() {
		return this.space;
	}
	

	@Override
	public void setPhysicsSpace(PhysicsSpace space) {
		this.space = space;
	}
	

	@Override
	public InputHandler getInputHandler() {
		return input;
	}
	

	@Override
	public void setInputHandler(InputHandler newInput) {
		input = newInput;
	}
	

	@Override
	public long getCurrentCycle() {
		return super.getCurrentCycle();
	}
	

	@Override
	public int getCyclesPerSecond() {
		return super.getCyclesPerSecond();
	}
	

	@Override
	public boolean isDebug() {
		return debug;
	}
	

	@Override
	public void setDebug(boolean debug) {
		this.debug = debug;
	}
	

	/**
	 * @return {@link SimEngine#statics}
	 */
	@Override
	public Node getStatics() {
		return statics;
	}
	

	/**
	 * @return {@link SimEngine#dynamics}
	 */
	@Override
	public Node getDynamics() {
		return dynamics;
	}
	
	
	/**
	 * @return {@link SimEngine#balls}
	 */
	@Override
	public Node getBalls() {
		return balls;
	}
	
	
	/**
	 * @return {@link SimEngine#robots}
	 */
	@Override
	public Node getRobots() {
		return robots;
	}
	

	/**
	 * @return {@link SimEngine#hud}
	 */
	@Override
	public Node getHUD() {
		return hud;
	}
	

	/**
	 * @return {@link SimEngine#vis}
	 */
	@Override
	public Node getVisualizationRoot() {
		return vis;
	}
	

	// Display
	@Override
	public DisplaySystem getDisplay() {
		return display;
	}
	

	@Override
	public int getDisplayWidth() {
		return width; // display.getWidth()
	}
	

	@Override
	public int getDisplayHeight() {
		return height; // display.getHeight();
	}
	

	@Override
	public int getDisplayDepth() {
		return depth; // display.getBitDepth();
	}
	

	@Override
	public int getDisplayFreq() {
		return freq; // display.getFrequency();
	}
	

	@Override
	public boolean isFullscreen() {
		return display.isFullScreen();
	}
	

	@Override
	public void resetTimer() {
		timer.reset();
	}
	
	// public boolean isOnCanvas() {
	// return onCanvas;
	// }
	
	// void setOnCanvas(boolean onCanvas) {
	// this.onCanvas = onCanvas;
	// }
}
