package edu.dhbw.mannheim.tigers.sim.engine;

import java.util.concurrent.Callable;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.Future;
import java.util.concurrent.atomic.AtomicBoolean;

import com.jme.util.GameTask;
import com.jme.util.GameTaskQueue;


/**
 * The concept and code is used from {@link GameTaskQueue}, enhanced with other features
 * 
 * @author Gero
 * 
 * @see Future
 * @see Callable
 */
public class TaskQueue {
	
	public static final String									RENDER		= "render";
	public static final String									UPDATE		= "update";
	
	private final ConcurrentLinkedQueue<GameTask<?>>	queue			= new ConcurrentLinkedQueue<GameTask<?>>();
	private final AtomicBoolean								executeAll	= new AtomicBoolean();
	
	private boolean												cancelled	= false;
	
	
	/**
	 * The state of this <code>GameTaskQueue</code> if it
	 * will execute all enqueued Callables on an execute
	 * invokation.
	 * 
	 * @return
	 *         boolean
	 */
	public boolean isExecuteAll() {
		return executeAll.get();
	}
	

	/**
	 * Sets the executeAll boolean value to determine if
	 * when execute() is invoked if it should simply execute
	 * one Callable, or if it should invoke all. This defaults
	 * to false to keep the game moving more smoothly.
	 * 
	 * @param executeAll
	 */
	public void setExecuteAll(boolean executeAll) {
		this.executeAll.set(executeAll);
	}
	

	/**
	 * Adds the Callable to the internal queue to invoked and
	 * returns a Future that wraps the return. This is useful
	 * for checking the status of the task as well as being able
	 * to retrieve the return object from Callable asynchronously.
	 * 
	 * @param <V>
	 * @param callable
	 * @return
	 */
	public synchronized <V> Future<V> enqueue(Callable<V> callable) {
		GameTask<V> task = new GameTask<V>(callable);
		if (cancelled) {
			task.cancel(true);
			return task;
		} else {
			queue.add(task);
			return task;
		}
	}
	

	/**
	 * Used for shutdown purposes
	 */
	public synchronized void cancelQueue() {
		cancelled = true;
		
		GameTask<?> task = queue.poll();
		do {
			if (task == null) {
				return;
			}
			task.cancel(false);
		} while ((executeAll.get()) && ((task = queue.poll()) != null));
	}
	

	/**
	 * This method should be invoked in the update() or render() method
	 * inside the main game to make sure the tasks are invoked in the
	 * OpenGL thread.
	 */
	public synchronized void execute() {
		GameTask<?> task = queue.poll();
		do {
			if (task == null)
				return;
			while (task.isCancelled()) {
				task = queue.poll();
				if (task == null)
					return;
			}
			task.invoke();
		} while ((executeAll.get()) && ((task = queue.poll()) != null));
	}
}
