package edu.dhbw.mannheim.tigers.sim.engine.com.bot;


import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import org.apache.log4j.Logger;

import edu.dhbw.mannheim.tigers.sim.engine.com.IStartableInterface;


/**
 * The second part of the "botcom-trio", responsible for the transmitting part.
 * 
 * @see ABotCommunicator
 * @see ABotCommandReceiver
 * 
 * @author Gero
 */
public abstract class ABotStatusTransmitter<C> implements IStartableInterface, Runnable, IBotCommandTransmitter<C> {
	
	// --------------------------------------------------------------------------
	// --- variables and constants ----------------------------------------------
	// --------------------------------------------------------------------------
	private final Logger					log							= Logger.getLogger(getClass());
	private Thread							thread;
	private final String					name;
	
	private final ICommunicator<C>	comm;
	
	private volatile int					statusMessageInterval	= 500;
	
	protected final List<C>				tempStatusList				= new ArrayList<C>();
	private volatile boolean			shutdownInProgress		= false;
	
	
	// --------------------------------------------------------------------------
	// --- constructors ---------------------------------------------------------
	// --------------------------------------------------------------------------
	/**
	 * @see ABotStatusTransmitter
	 * @param communicator
	 * @param outputStream
	 */
	public ABotStatusTransmitter(ICommunicator<C> communicator) {
		this.comm = communicator;
		this.name = "BotCmdTransmitter " + comm.getBotId();
	}
	

	// --------------------------------------------------------------------------
	// --- methods --------------------------------------------------------------
	// --------------------------------------------------------------------------
	@Override
	public synchronized void start() {
		thread = new Thread(this, name);
		thread.start();
	}
	

	@Override
	public synchronized void stop() {
		shutdownInProgress = true;
		
		if (thread != null) {
			thread.interrupt();
			thread = null;
		}
		
		try {
			closeOutput();
		} catch (IOException err) {
			log.error("Error while closing output: " + err.getMessage());
		}
		
		comm.notifyShutdown();
	}
	

	@Override
	public void run() {
		log.debug("Transmitter started!");
		
		while (!Thread.currentThread().isInterrupted() && !shutdownInProgress) {
			try {
				boolean success = writeOutput();
				if (!success) {
					if (!shutdownInProgress) {
						log.error("Error while sending status!");
					} else {
						break;
					}
				}
				
				Thread.sleep(statusMessageInterval);
			} catch (IOException e1) {
				break;
			} catch (InterruptedException e) {
				if (!shutdownInProgress) { // Also caught here if sleeping while transmitter is closed so... =)
					log.error(e.getMessage(), e);
					shutdownInProgress = false;
					break;
				}
			}
		}
		
	}
	

	/**
	 * Triggered whenever a new status message should be sent to the server
	 * 
	 * @return success?
	 * @throws IOException
	 */
	protected abstract boolean writeOutput() throws IOException;
	

	/**
	 * Called when the transmitter is shut down for cleanup purposes
	 * 
	 * @throws IOException
	 */
	protected abstract void closeOutput() throws IOException;
	

	@Override
	public void setStatusMessageInterval(int newInterval) {
		this.statusMessageInterval = newInterval;
	}
}