package edu.dhbw.mannheim.tigers.sim.engine.com.bot.tiger;

import java.io.IOException;
import java.net.BindException;
import java.net.InetAddress;
import java.net.NetworkInterface;
import java.net.UnknownHostException;
import java.util.concurrent.TimeUnit;

import org.apache.log4j.Logger;

import edu.dhbw.mannheim.tigers.sim.engine.com.bot.ABotCommunicator;
import edu.dhbw.mannheim.tigers.sim.engine.com.bot.IBotCommandTransmitter;
import edu.dhbw.mannheim.tigers.sim.settings.data.ITeamProperties;
import edu.dhbw.mannheim.tigers.sim.util.network.NetworkUtility;
import edu.dhbw.mannheim.tigers.sim.util.network.UnicastUDPTransceiver;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.botmanager.commands.ACommand;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.botmanager.commands.tiger.TigerSystemSetIdentity;


/**
 * The {@link ABotCommunicator}-implementation for the Tiger-bot.
 * This is a little bit more complex then meant to by {@link ABotCommunicator}, caused by the special mutlicast-behavior
 * of the TigerBot (listening everytime when online), implemented in {@link TigerBotMulticastTransceiver}.
 * 
 * @see TigerBotCommandReceiver
 * @see TigerBotStatusTransmitter
 * 
 * @author Gero
 */
public class TigerBotCommunicator extends ABotCommunicator<ACommand, ACommand> {
	
	// --------------------------------------------------------------------------
	// --- variables and constants ----------------------------------------------
	// --------------------------------------------------------------------------
	private final Logger							log								= Logger.getLogger(getClass());
	
	// Comm
	private TigerBotMulticastTransceiver	mcastTransceiver;
	private TigerBotCommandReceiver			receiver;
	private TigerBotStatusTransmitter		transmitter;
	private final NetworkUtility				networkUtil						= new NetworkUtility();
	
	/**
	 * Switch used to differ between calls to {@link #notifyShutdown()} caused from within this class and others.
	 */
	private volatile boolean					dropShutdownNotification	= false;
	
	/** [ns] */
	private final long							timeout;
	
	
	// --------------------------------------------------------------------------
	// --- constructors ---------------------------------------------------------
	// --------------------------------------------------------------------------
	/**
	 * @param port
	 * @param botController
	 * @param timeout
	 * 
	 * @see TigerBotCommunicator
	 */
	public TigerBotCommunicator(int port, TigerBotController botController, int timeout) {
		super(botController, timeout);
		
		this.timeout = TimeUnit.MILLISECONDS.toNanos(timeout);
	}
	

	// --------------------------------------------------------------------------
	// --- methods --------------------------------------------------------------
	// --------------------------------------------------------------------------
	@Override
	protected void initialize() {
		TigerBotController c = (TigerBotController) botController;
		
		// Detect interfaces to listen on
		ITeamProperties teamProps = c.getProperties();
		
		final String ifaceAddr = teamProps.getMcastInterface();
		final NetworkInterface nif = networkUtil.chooseNetworkInterface(ifaceAddr, 3);
		if (nif == null) {
			log.error("No proper nif for bot-detection in network '" + ifaceAddr + "' found!");
		} else {
			// log.debug("Chose nif for " + c.getID() + ": " + nif.getDisplayName() + ".");
		}
		

		// Set up
		mcastTransceiver = new TigerBotMulticastTransceiver(c, nif);
		mcastTransceiver.start();
		
		mcastTransceiver.sendAnnouncement();
	}
	

	/**
	 * Check for timeout and return <code>false</code> if its has!
	 */
	@Override
	protected boolean checkConnection() {
		
		TigerBotMulticastTransceiver txrx = mcastTransceiver;
		if (txrx == null) {
			return false;
		}
		
		TigerBotCommandReceiver rec = receiver;
		

		long lastReceivedMcast = txrx.getLastPacketReceived();
		long lastReceivedUni = 0;
		if (rec != null) {
			lastReceivedUni = rec.getLastPacketReceived();
		}
		
		// Initialized?
		if (lastReceivedMcast == 0 && lastReceivedUni == 0) {
			return true;
		}
		
		// Timed out?
		final long lastReceived = Math.max(lastReceivedMcast, lastReceivedUni);
		final long now = System.nanoTime();
		if (lastReceived + timeout < now) {
			log.info(botController.getID() + " timed out! (" + (now - lastReceived) + "ns to late!)");
			return false;
		}
		
		return true;
	}
	

	@Override
	protected boolean startTransceiving() throws InterruptedException {
		if (mcastTransceiver != null) {
			mcastTransceiver.resetLastPacketReceived();
		}
		return true;
	}
	

	/**
	 * Stops the unicast-communication to the CS. This got mostly triggered by a disconnect-message from the CS.
	 */
	public synchronized void stopUnicast() {
		dropShutdownNotification = true;
		stopUnicastTxRx();
		dropShutdownNotification = false;
	}
	

	/**
	 * Called whenever a new {@link TigerSystemSetIdentity} command is received.
	 * 
	 * @param info
	 */
	public synchronized void restartUnicast(TigerIdentityInfo info) {
		dropShutdownNotification = true;
		stopUnicastTxRx();
		startUnicastTxRx(info);
		dropShutdownNotification = false;
	}
	

	private void startUnicastTxRx(TigerIdentityInfo info) {
		
		// The following is needed because of routing-problems
		// At this point we need to know the servers address so we know where to send our status-messages to.
		// Normally one'd use the sender of the SetIdentity-cmd (== server == info.fromAddr), but if both applications
		// (Sim and CS) are running local, the following occurs:
		// If we are connected to a network, the sender is "192.168.0.101", e.g., but listening to packets from
		// "127.0.0.1". If the Sim-Bot sends his msgs to "192.168.0.101" now, the server won't see it: What a mess :-(
		// The solution: Below __isLoopbackAddress()__ ! =)
		String serverAddr = info.fromAddr.getHostAddress();
		InetAddress botSeenFromServer = getAddr(info.identity.getIp());
		if (botSeenFromServer != null && botSeenFromServer.isLoopbackAddress()) {
			serverAddr = "127.0.0.1";
		}
		
		try {
			UnicastUDPTransceiver transceiver = new UnicastUDPTransceiver(info.identity.getPort(), info.identity.getIp(),
					serverAddr, info.identity.getServerPort());
			
			receiver = new TigerBotCommandReceiver(this, transceiver, botController.getID().toString());
			receiver.start();
			
			transmitter = new TigerBotStatusTransmitter(this, transceiver, (ITigerBotStatusProvider) botController);
			transmitter.start();
		} catch (BindException err) {
			log.warn("Configuration error: Most likely the IP '" + info.identity.getIp() + "' does not match this PC!");
			return;
		}
	}
	

	private InetAddress getAddr(String addrStr) {
		try {
			return InetAddress.getByName(addrStr);
		} catch (UnknownHostException err) {
			log.error(err.getMessage());
			return null;
		}
	}
	

	private void stopUnicastTxRx() {
		if (receiver != null) {
			receiver.stop();
			receiver = null;
		}
		if (transmitter != null) {
			transmitter.stop();
			transmitter = null;
		}
	}
	

	@Override
	public void notifyShutdown() {
		if (!dropShutdownNotification) {
			super.notifyShutdown();
		}
	}
	

	private void stopMcastTxRx() {
		if (mcastTransceiver != null) {
			mcastTransceiver.stop();
			mcastTransceiver = null;
		}
	}
	

	@Override
	protected void stopTransceiving() {
		// stopMcastTxRx();
		stopUnicastTxRx();
	}
	

	@Override
	protected void deinitialize() {
		stopMcastTxRx();
	}
	

	@Override
	protected IBotCommandTransmitter<ACommand> getTransmitter() {
		return transmitter;
	}
	

	@Override
	public void send(ACommand msg) {
		try {
			transmitter.send(msg);
		} catch (IOException err) {
			log.error("Error while sending msg. Dropped.");
		}
	}
}
