package edu.dhbw.mannheim.tigers.sim.engine.com.bot.tiger;


import java.util.LinkedList;
import java.util.List;
import java.util.Queue;

import org.apache.log4j.Logger;

import edu.dhbw.mannheim.tigers.sim.engine.com.ERoboCupTeam;
import edu.dhbw.mannheim.tigers.sim.engine.com.IRoboCupMatchControlProvider;
import edu.dhbw.mannheim.tigers.sim.engine.com.bot.ABotController;
import edu.dhbw.mannheim.tigers.sim.engine.model.bot.BotID;
import edu.dhbw.mannheim.tigers.sim.engine.model.bot.parts.KickMode;
import edu.dhbw.mannheim.tigers.sim.engine.model.bot.tiger.ITigerBotControl;
import edu.dhbw.mannheim.tigers.sim.engine.model.bot.tiger.TigerBotStatus;
import edu.dhbw.mannheim.tigers.sim.engine.model.bot.tiger.TigerMove;
import edu.dhbw.mannheim.tigers.sim.settings.ISettingsProvider;
import edu.dhbw.mannheim.tigers.sim.settings.data.IMatchModeProperties;
import edu.dhbw.mannheim.tigers.sim.settings.data.ITeamProperties;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.botmanager.commands.ACommand;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.botmanager.commands.CommandConstants;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.botmanager.commands.tiger.TigerDribble;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.botmanager.commands.tiger.TigerKickerKickV2;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.botmanager.commands.tiger.TigerKickerKickV2.Device;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.botmanager.commands.tiger.TigerKickerStatusV2;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.botmanager.commands.tiger.TigerMotorMoveV2;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.botmanager.commands.tiger.TigerMotorSetParams;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.botmanager.commands.tiger.TigerMulticastUpdateAllV2;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.botmanager.commands.tiger.TigerSystemPing;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.botmanager.commands.tiger.TigerSystemPong;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.botmanager.commands.tiger.TigerSystemSetIdentity;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.botmanager.commands.tiger.TigerSystemStatusMovement;


/**
 * {@link ABotController} implementation for the Tiger bot. Handles incoming messages and sends
 * status-reports via a {@link TigerBotCommunicator2}
 * 
 * @author Gero
 */
public class TigerBotController extends ABotController<ITigerBotControl, ACommand, TigerSystemStatusMovement> implements
		ITigerBotController {
	
	// --------------------------------------------------------------------------
	// --- variables and constants ----------------------------------------------
	// --------------------------------------------------------------------------
	public static final int					DISCONNECT_BOT_ID	= 255;
	
	private final Logger						log					= Logger.getLogger(getClass());
	
	// ID
	private final String						cpuId;
	
	// Comm
	private TigerBotCommunicator			comm;
	
	// Processing
	private final Queue<ACommand>			cmdQueue				= new LinkedList<ACommand>();
	
	// Team specific
	private final ITeamProperties 		teamProps;
	
	
	// --------------------------------------------------------------------------
	// --- constructors ---------------------------------------------------------
	// --------------------------------------------------------------------------
	/**
	 * @see TigerBotController
	 * @param bot
	 * @param port
	 * @param matchControlProvider
	 * @param team
	 */
	public TigerBotController(ITigerBotControl bot, int port, IRoboCupMatchControlProvider matchControlProvider,
			ERoboCupTeam team) {
		super(bot, matchControlProvider);
		
		cpuId = bot.getCpuId();

		// Properties
		IMatchModeProperties props = ISettingsProvider.INSTANCE.getSimProperties().getMatchModeProperties();
		
		switch (team) {
			default:
			case YELLOW_TEAM:
				teamProps = props.getTeamYellow();
				break;
			
			case BLUE_TEAM:
				teamProps = props.getTeamBlue();
				break;
		}
		
		
		// Comm
		setBotCommunicator(comm = new TigerBotCommunicator(port, this, teamProps.getTimeout()));
		comm.initialize();
		comm.setStatusInterval(100); // [ms]
	}
	

	// --------------------------------------------------------------------------
	// --- methods --------------------------------------------------------------
	// --------------------------------------------------------------------------
	@Override
	public void start() {
		comm.start();
	}
	

	@Override
	public synchronized void processNewCommand(ACommand newCmd) {
		
		if (isBlocked()) {
			return; // Should command be executed?
		}
		

		try {
			// Handle multi-cmd
			if (newCmd.getCommand() == CommandConstants.CMD_MULTICAST_UPDATE_ALL_V2) {
				TigerMulticastUpdateAllV2 allCmd = (TigerMulticastUpdateAllV2) newCmd;
				
				// Detect slot for this robot
				final int slot = allCmd.getSlot(getID().getIntValue());
				if (slot == -1)
				{
//					log.warn(getID() + " received UPDATE_ALL without information for him!?");
					return;
				}
				
				// Retrieve commands
				cmdQueue.add(allCmd.getKick(slot));
				cmdQueue.add(allCmd.getDribble(slot));
				
				final boolean useMove = allCmd.getUseMove(slot);
				if (useMove)
				{
					cmdQueue.add(allCmd.getMove(slot));
				}
				
			} else {
				cmdQueue.add(newCmd);
			}
			
	
			while (!cmdQueue.isEmpty()) {
				
				ACommand cmd = cmdQueue.poll();
				switch (cmd.getCommand()) {
					
					// Dribble
					case CommandConstants.CMD_MOTOR_DRIBBLE: {
						TigerDribble dribble = (TigerDribble) cmd;
						
						bot.dribble(dribble.getSpeed());
						break;
					}
						
						// Kicker
					case CommandConstants.CMD_KICKER_KICKV2: {
						TigerKickerKickV2 kick = (TigerKickerKickV2) cmd;
						
						KickMode mode = KickMode.createMode(kick.getMode());
						
						if (mode != KickMode.NONE) {
							switch (kick.getDevice()) {
								
								case Device.STRAIGHT:
									// System.out.println("kick: " + kick.getFiringDuration());
									final int duration = (int) (kick.getFiringDuration() / TigerKickerKickV2.MAX_DURATION * 100);
//									System.out.println(getID() + " Received kick! (dur " + duration + ")");
									bot.kickStraight(duration, mode);
									break;
								
								case Device.CHIP:
									bot.lob(100, mode);
									break;
								
								default:
									log.warn("Received command for unknown kicking-device '" + kick.getDevice() + "!");
									break;
							}
						}
						
						break;
					}
						
					
					// Move
					case CommandConstants.CMD_MOTOR_MOVE_V2: {
						TigerMotorMoveV2 newMove = (TigerMotorMoveV2) cmd;
						bot.getMotorControl().setSpeedAutomatic(newMove.getX(), newMove.getY(), newMove.getV() + newMove.getW());	// Ignore W!
						break;
					}
						
		
					case CommandConstants.CMD_MOTOR_SET_PARAMS: {
						TigerMotorSetParams params = (TigerMotorSetParams) cmd;
						bot.getMotorControl().setPIDParams(params);
						break;
					}
						
		
					case CommandConstants.CMD_MOVEMENT_SET_PARAMS: {
						// TODO Set move params
						break;
					}
						
		
					case CommandConstants.CMD_MOTOR_SET_MANUAL: {
						// TODO Set manual
						break;
					}
						
		
					case CommandConstants.CMD_MOTOR_SET_PID_SP: {
						// TODO set PID SP
						break;
					}
					
					
					// Ping pong
					case CommandConstants.CMD_SYSTEM_PING: {
						TigerSystemPing ping = (TigerSystemPing) cmd;
						TigerSystemPong pong = new TigerSystemPong(ping.getId());
						comm.send(pong);
						break;
					}
						
		
					default: {
						break;
					}
				}
			}
		} catch (ClassCastException cce) {
			log.error("A received command missreported its command-id!");
		}
	}
	

	public void passNewIdentity(TigerIdentityInfo someIdentityInfo) {
		TigerSystemSetIdentity someIdentity = someIdentityInfo.identity;
		
		// Is my identity?
		if (!isMyIdentity(someIdentity, cpuId)) {
			return;
		}
		
		log.debug("Received new identity for CPU-ID: " + someIdentity.getCpuId() + "(Bot: " + getID() + ")");
		
		if (someIdentity.getBotId() != getID().getIntValue()) {
			log.error("Configuration error: The SSL-ID sent per command from the CS is different to the one configured in the simulator!");
		}
		
		comm.restartUnicast(someIdentityInfo);
	}
	

	private boolean isMyIdentity(final TigerSystemSetIdentity someIdentity, final String cpuId) {
		return someIdentity.getCpuId().equals(cpuId);
	}
	

	/**
	 * Checks the given {@link TigerSystemSetIdentity}-cmd and disconnects if it is for this bot.
	 * 
	 * @param identity
	 */
	public void processDisconnect(TigerSystemSetIdentity identity) {
		// Really?
		if (identity.getBotId() != DISCONNECT_BOT_ID) {
			log.debug("Invalid disconnect-message, proceeding.");
			return;
		}
		
		// Is this for me?
		if (isMyIdentity(identity, cpuId)) {
			log.debug("Received new disconnect message.");
			comm.stopUnicast();
		}
	}
	

	@Override
	public void stop() {
		comm.stop();
	}
	

	@Override
	public void onConnect() {
		bot.charge(true);
	}
	

	@Override
	public void onDisconnect() {
		bot.getMotorControl().setSpeedAutomatic(0, 0, 0); // Stop bot on disconnect
		
		// Reset kicker
		bot.dribble(0);
		bot.kickStraight(0, KickMode.DISARM);
		bot.lob(0, KickMode.DISARM);
		bot.charge(false);
	}
	

	// --------------------------------------------------------------------------
	// --- getter/setter --------------------------------------------------------
	// --------------------------------------------------------------------------
	public ITeamProperties getProperties() {
		return teamProps;
	}


	@Override
	public List<ACommand> getBotStatus(List<ACommand> store) {
		TigerBotStatus state = bot.getBotStatus();
		
		TigerMove move = state.move;
		store.add(new TigerSystemStatusMovement(move.vx, move.vy, move.vw));
		

		TigerKickerStatusV2 charge = new TigerKickerStatusV2();
		charge.setCapLevel(state.currentCharge);
		charge.setChargeCurrent(state.isCapLoading ? 1000 : 0); // !!!
		store.add(charge);
		
		return store;
	}
	

	@Override
	public BotID getID() {
		return bot.getID();
	}
	

	@Override
	public String getCpuId() {
		return cpuId;
	}
}
