package edu.dhbw.mannheim.tigers.sim.engine.delegates;

import java.text.DecimalFormat;
import java.util.LinkedList;

import com.jme.input.FirstPersonHandler;
import com.jme.input.InputHandler;
import com.jme.input.KeyBindingManager;
import com.jme.input.KeyInput;
import com.jme.input.KeyboardLookHandler;
import com.jme.input.MouseInput;
import com.jme.input.action.InputAction;
import com.jme.input.action.InputActionEvent;
import com.jme.input.keyboard.KeyboardInputHandlerDevice;
import com.jme.renderer.Renderer;
import com.jme.scene.Node;
import com.jme.scene.Spatial.CullHint;
import com.jme.scene.Text;

import edu.dhbw.mannheim.tigers.sim.engine.ESimEngineState;
import edu.dhbw.mannheim.tigers.sim.engine.ICore;
import edu.dhbw.mannheim.tigers.sim.engine.model.CamHelper;
import edu.dhbw.mannheim.tigers.sim.engine.model.UnitConversion;


/**
 * This {@link ASimDelegate} sets up an {@link InputHandler} and cares for all the basic action-stuff like
 * camera-movement and so on
 * 
 * @author Gero
 */
public class BasicInputDelegate extends ASimDelegate {
	
	// --------------------------------------------------------------------------
	// --- variables and constants ----------------------------------------------
	// --------------------------------------------------------------------------
	private InputHandler					input			= null;
	
	// Stats
	private StringBuffer					sbStatics	= null;
	private StringBuffer					sbDynamics	= null;
	

	private boolean						viewStats	= true;
	private final DecimalFormat		df2			= new DecimalFormat("000.00");
	private final DecimalFormat		df4			= new DecimalFormat("0.0000");
	
	private Node							keyHelp;
	private final LinkedList<Text>	keyHelpText	= new LinkedList<Text>();			;
	
	public enum InputType {
		FIRST_PERSON,
		THIRD_PERSON,
		BASIC;
	}
	
	private InputType	type	= null;
	
	
	// --------------------------------------------------------------------------
	// --- constructors ---------------------------------------------------------
	// --------------------------------------------------------------------------
	/**
	 * @see BasicInputDelegate
	 * @param core
	 * @param type
	 */
	public BasicInputDelegate(ICore core, InputType type) {
		super(core);
		
		this.type = type;
	}
	

	// --------------------------------------------------------------------------
	// --- methods --------------------------------------------------------------
	// --------------------------------------------------------------------------
	@Override
	public void initSystem() {
		
		switch (type) {
			case THIRD_PERSON:
				throw new UnsupportedOperationException();
				// break;
				

			case FIRST_PERSON:
				input = new FirstPersonHandler(core.getCamera());
				final FirstPersonHandler firstInput = (FirstPersonHandler) input;
				
				// Fix speed
				KeyboardLookHandler kblh = firstInput.getKeyboardLookHandler();
				kblh.setMoveSpeed(kblh.getMoveSpeed() * UnitConversion.getWorldToModelScale() * 3);
				
				// Fix jME-Keybinding-bug
				InputHandler.addDevice(new KeyboardInputHandlerDevice());
				// for (InputHandlerDevice device : InputHandler.getDevices()) {
				// if (device.getName().equals(InputHandler.DEVICE_KEYBOARD)) {
				// KeyboardInputHandlerDevice keyboard = (KeyboardInputHandlerDevice) device;
				// keyboard.getKeyboardListener()
				// }
				// }
				
				// Keyboard bindings ###################################
				KeyBindingManager keyboard = KeyBindingManager.getKeyBindingManager();
				
				// Exit with "Escape"
				keyboard.set("exit", KeyInput.KEY_ESCAPE);
				
				// Fix elevateDown, as it is bound to Z, lying on QWERTZ-keyboards quite uncomfortable...
				keyboard.set("elevateDown", KeyInput.KEY_Y);
				
				// Remove
				keyboard.remove("lookUp");
				keyboard.remove("lookDown");
				keyboard.remove("turnRight");
				keyboard.remove("turnLeft");
				
				// Release mouse from window
				InputAction mouseSwitch = null;
				input.addAction(mouseSwitch = new InputAction() {
					@Override
					public void performAction(InputActionEvent evt) {
						
						if (!evt.getTriggerPressed()) { // Trigger released
							if (firstInput.getMouseLookHandler().isEnabled()) {
								// turn off mouselook
								firstInput.getMouseLookHandler().setEnabled(false);
								if (!MouseInput.get().isCursorVisible()) {
									// turn on the mouse cursor
									MouseInput.get().setCursorVisible(true);
								}
							} else {
								// turn on mouselook
								firstInput.getMouseLookHandler().setEnabled(true);
								if (MouseInput.get().isCursorVisible()) {
									// turn off the mouse cursor
									MouseInput.get().setCursorVisible(false);
								}
							}
						}
					}
					
				}, InputHandler.DEVICE_KEYBOARD, KeyInput.KEY_SPACE, InputHandler.AXIS_NONE, false);
				
				// Trigger event programmatically to make the window release the mouse at the start!
				InputActionEvent ev = new InputActionEvent();
				ev.setTriggerPressed(false);
				mouseSwitch.performAction(ev);
				break;
			

			case BASIC:
				throw new UnsupportedOperationException();
				// break;
		}
		

		// Debug mode
		input.addAction(new InputAction() {
			@Override
			public void performAction(InputActionEvent evt) {
				if (evt.getTriggerPressed()) {
					core.setDebug(!core.isDebug());
				}
			}
		}, InputHandler.DEVICE_KEYBOARD, KeyInput.KEY_V, InputHandler.AXIS_NONE, false);
		

		// Toggle Pause
		KeyBindingManager.getKeyBindingManager().add("toggle_pause", KeyInput.KEY_P);
		

		// Brings camera back to START_LOCATION
		input.addAction(new InputAction() {
			@Override
			public void performAction(InputActionEvent evt) {
				CamHelper.resetCam(core.getCamera());
			}
		}, InputHandler.DEVICE_KEYBOARD, KeyInput.KEY_HOME, InputHandler.AXIS_NONE, false);
		
		// Stats
		input.addAction(new InputAction() {
			@Override
			public void performAction(InputActionEvent evt) {
				viewStats = !viewStats;
			}
		}, InputHandler.DEVICE_KEYBOARD, KeyInput.KEY_TAB, InputHandler.AXIS_NONE, false);
		
		// Display Key bindings
		input.addAction(new InputAction() {
			
			@Override
			public void performAction(InputActionEvent evt) {
				if (evt.getTriggerPressed()) {
					core.getHUD().attachChild(keyHelp);
				} else {
					keyHelp.removeFromParent();
				}
				
			}
		}, InputHandler.DEVICE_KEYBOARD, KeyInput.KEY_F1, InputHandler.AXIS_NONE, false);
		

		// Show statistic TODO Extract statistics delegate
		final Node simParams = new Node("simParams");
		
		Text statics = Text.createDefaultTextLabel("parameter", "");
		statics.getLocalTranslation().set(0, core.getDisplayHeight() - 15, 0);
		sbStatics = statics.getText();
		simParams.attachChild(statics);
		
		Text dynamics = Text.createDefaultTextLabel("parameter", "");
		dynamics.getLocalTranslation().set(0, core.getDisplayHeight() - 25, 0);
		sbDynamics = dynamics.getText();
		simParams.attachChild(dynamics);
		
		simParams.setCullHint(CullHint.Never);
		simParams.setRenderQueueMode(Renderer.QUEUE_ORTHO);
		simParams.updateGeometricState(0.0f, true);
		simParams.updateRenderState();
		core.getHUD().attachChild(simParams);
		

		// Label for Key bindings
		keyHelp = new Node("Key bindings");
		
		keyHelpText.add(Text.createDefaultTextLabel("", "Key:                  Function:"));
		keyHelpText.add(Text.createDefaultTextLabel("", ""));
		keyHelpText.add(Text.createDefaultTextLabel("", "F1                    Show Key bindings"));
		keyHelpText.add(Text.createDefaultTextLabel("", "F2                    Show Bot Infos"));
		keyHelpText.add(Text.createDefaultTextLabel("", "F5                    Display/hide predicted bot-positions"));
		keyHelpText.add(Text.createDefaultTextLabel("", "F6                    Display/hide predicted ball-positions"));
		keyHelpText.add(Text.createDefaultTextLabel("", "F7                    Display/hide planed bot-aths"));
		keyHelpText.add(Text.createDefaultTextLabel("", "F8                    Start/Stop video capture."));
		keyHelpText.add(Text.createDefaultTextLabel("", "F12                   Do screenshot"));
		keyHelpText.add(Text.createDefaultTextLabel("", "TAB                   Show statistics"));
		keyHelpText.add(Text.createDefaultTextLabel("", ""));
		keyHelpText.add(Text.createDefaultTextLabel("", "SPACE                 Curser released"));
		keyHelpText.add(Text.createDefaultTextLabel("", "V                     Enable/Disable debug mode"));
		keyHelpText.add(Text.createDefaultTextLabel("", "POS1                  Reset camera position"));
		keyHelpText.add(Text.createDefaultTextLabel("", "W,A,S,D,Q,Y           Move camera"));
		keyHelpText.add(Text.createDefaultTextLabel("", ""));
		keyHelpText.add(Text.createDefaultTextLabel("",
				"LEFT_MOUSE_BUTTON     Pick object and drag it anywhere on the field"));
		keyHelpText.add(Text
				.createDefaultTextLabel("", "RIGHT_MOUSE_BUTTON    (While picked) Uprights \"picked object\""));
		keyHelpText.add(Text.createDefaultTextLabel("",
				"MOUSE_WHEEL           (While picked) Lower/raise the picked object"));
		keyHelpText.add(Text.createDefaultTextLabel("", "MIDDLE_MOUSE_BUTTON   (While picked) Restore default height"));
		keyHelpText.add(Text.createDefaultTextLabel("", "T,F,G,H               Move ball"));
		keyHelpText.add(Text.createDefaultTextLabel("", "J                     Let the ball jump"));
		keyHelpText.add(Text.createDefaultTextLabel("", "R                     Reset the ball"));
		keyHelpText.add(Text.createDefaultTextLabel("", "B                     Beams the ball to mosue-position"));
		keyHelpText.add(Text.createDefaultTextLabel("", ""));
		keyHelpText.add(Text.createDefaultTextLabel("", "INSERT,DELETE         Choose/discard a bot to control"));
		keyHelpText
				.add(Text.createDefaultTextLabel("", "UP,DOWN,LEFT,RIGHT    Simple move control for the selected bot"));
		
		int height = 70;
		for (Text text : keyHelpText) {
			text.getLocalTranslation().set(0, core.getDisplayHeight() - height, 0);
			text.setName(height + "");
			keyHelp.attachChild(text);
			height += 15;
		}
		
		keyHelp.setCullHint(CullHint.Never);
		keyHelp.setRenderQueueMode(Renderer.QUEUE_ORTHO);
		keyHelp.updateGeometricState(0.0f, true);
		keyHelp.updateRenderState();
		

		// Populate it
		core.setInputHandler(input);
	}
	

	@Override
	public void initSimulation() {
		
		// Reset the action speed, after every action has been added!
		// input.setActionSpeed(UC.getModelScale());
	}
	

	@Override
	public void update(float interpolation, float tpc, float cps) {
		
		core.getInputHandler().update(tpc);
		
		if (KeyBindingManager.getKeyBindingManager().isValidCommand("toggle_pause", false)) {
			if (core.getState().equals(ESimEngineState.PAUSED_SIMULATION)) {
				core.resume();
			} else {
				core.pauseSimulation();
			}
		}
		
		if (KeyBindingManager.getKeyBindingManager().isValidCommand("exit", false)) {
			core.stop();
		}
		

		if (viewStats && core.isRenderCycle()) {
			sbStatics.replace(0, sbStatics.length(),
					"CPS: " + df2.format(core.getCyclesPerSecond()) + "  FPS: " + df2.format(core.getFramesPerSecond()));
			sbDynamics.replace(
					0,
					sbDynamics.length(),
					"cps: " + df2.format(cps) + "  fps: " + df2.format(core.getFramesPerSecond()) + "  tpc: "
							+ df4.format(tpc));
		}
	}
}
