package edu.dhbw.mannheim.tigers.sim.engine.delegates;


import java.util.List;

import org.odejava.World;

import com.jme.math.Vector3f;
import com.jmex.physics.PhysicsDebugger;
import com.jmex.physics.PhysicsSpace;
import com.jmex.physics.contact.ContactCallback;
import com.jmex.physics.impl.ode.OdePhysicsSpace;
import com.jmex.physics.util.SimplePhysicsGame;

import edu.dhbw.mannheim.tigers.sim.engine.ICore;
import edu.dhbw.mannheim.tigers.sim.engine.model.UnitConversion;
import edu.dhbw.mannheim.tigers.sim.engine.model.material.MaterialContactCallbackXT;


/**
 * @see UnitConversion
 * @see SimplePhysicsGame
 * 
 * @author Gero
 */
public class SimPhysicsDelegate extends ASimDelegate {
	
	// --------------------------------------------------------------------------
	// --- variables and constants ----------------------------------------------
	// --------------------------------------------------------------------------
	private PhysicsSpace				physicsSpace;
	
	private boolean					firstFrame			= true;
	
	public static final Vector3f	GRAVITY				= new Vector3f(0, -9.81f, 0);
	
	// ODE-Constants
	private static final float		ACCURACY				= 0.003f;	// 0.002f
	private static final int		INTERACTION_STEPS	= 15;
	
	private static final float		ODE_CFM				= 0.005f;
	private static final float		ODE_ERP				= 0.2f;
	
	
	// --------------------------------------------------------------------------
	// --- constructors ---------------------------------------------------------
	// --------------------------------------------------------------------------
	/**
	 * @see SimPhysicsDelegate
	 * @param core
	 */
	public SimPhysicsDelegate(ICore core) {
		super(core);
	}
	

	// --------------------------------------------------------------------------
	// --- methods --------------------------------------------------------------
	// --------------------------------------------------------------------------
	@Override
	public void initSystem() {
		
		// Set up PhysicSpace and gravity
		setPhysicsSpace(PhysicsSpace.create());
		
		// Place the Sims own implementation of the ContactCallback-interface. This is especially
		// necessary for the proper simulation of anisotropic friction for the wheels
		List<ContactCallback> callbacks = physicsSpace.getContactCallbacks();
		// Only 'MaterialContactCallback' in here after constructor, so clear should do it without any
		// side-effects
		callbacks.clear();
		callbacks.add(MaterialContactCallbackXT.getInstance(core));
	}
	

	@Override
	public void initSimulation() {
		
		if (physicsSpace instanceof OdePhysicsSpace) {
			OdePhysicsSpace odePS = (OdePhysicsSpace) physicsSpace;
			
			// ### Performance settings
			odePS.setStepFunction(OdePhysicsSpace.SF_STEP_QUICK);
			odePS.setStepInteractions(INTERACTION_STEPS);	// Seems to have no impact on performance... ?
			
			odePS.setStepSize(ACCURACY);
			odePS.setUpdateRate(1 / ACCURACY);
			// ### Performance settings
			
			World odeWorld = odePS.getODEJavaWorld();
			odeWorld.setConstraintForceMix(ODE_CFM);
			odeWorld.setErrorReductionParameter(ODE_ERP);
			
			odePS.setAutoRestThreshold(0.01f);
			
			log.debug("ODE - CFM: " + odeWorld.getConstraintForceMix());
			log.debug("ODE - ERP: " + odeWorld.getErrorReductionParameter());
		}
		
		physicsSpace.setDirectionalGravity(GRAVITY);
		
		// core.getInputHandler().addAction(new InputAction() {
		// private boolean on = true;
		// @Override
		// public void performAction(InputActionEvent evt) {
		// if (!evt.getTriggerPressed()) {
		// if (on) {
		// physicsSpace.setDirectionalGravity(Vector3f.ZERO.clone());
		// on = false;
		// } else {
		// physicsSpace.setDirectionalGravity(GRAVITY);
		// on = true;
		// }
		// }
		// }
		// }, InputHandler.DEVICE_KEYBOARD, KeyInput.KEY_Z, InputHandler.AXIS_NONE, false);
	}
	

	@Override
	public void update(float interpolation, float tpf, float fps) {
		
		if (tpf > 0.2 || Float.isNaN(tpf)) {
			log.warn("Maximum physics update interval is 0.2 seconds - capped.");
			tpf = 0.2f;
		}
		getPhysicsSpace().update(tpf);
		

		if (firstFrame) {
			// drawing and calculating the first frame usually takes longer than the rest
			// to avoid a rushing simulation we reset the timer
			core.resetTimer();
			firstFrame = false;
		}
	}
	

	@Override
	public void debug(float interpolation) {
		
		PhysicsDebugger.drawPhysics(getPhysicsSpace(), core.getDisplay().getRenderer());
	}
	

	@Override
	public void cleanup() {
		if (getPhysicsSpace() != null) {
			getPhysicsSpace().delete();
		}
	}
	

	protected void setPhysicsSpace(PhysicsSpace physicsSpace) {
		if (physicsSpace != this.physicsSpace) {
			if (this.physicsSpace != null) {
				this.physicsSpace.delete();
			}
			this.physicsSpace = physicsSpace;
			core.setPhysicsSpace(physicsSpace);
		}
	}
	

	protected PhysicsSpace getPhysicsSpace() {
		return physicsSpace;
	}
}
