package edu.dhbw.mannheim.tigers.sim.engine.delegates;


import static edu.dhbw.mannheim.tigers.sim.engine.model.UnitConversion.m;

import com.jme.input.InputHandler;
import com.jme.input.KeyInput;
import com.jme.input.action.InputAction;
import com.jme.input.action.InputActionEvent;
import com.jme.math.Vector2f;
import com.jme.math.Vector3f;
import com.jme.system.DisplaySystem;
import com.jmex.physics.DynamicPhysicsNode;

import edu.dhbw.mannheim.tigers.sim.engine.ICore;
import edu.dhbw.mannheim.tigers.sim.engine.model.CamHelper;
import edu.dhbw.mannheim.tigers.sim.engine.model.SimPhysicsPicker;
import edu.dhbw.mannheim.tigers.sim.engine.model.UnitConversion;
import edu.dhbw.mannheim.tigers.sim.engine.model.ball.RoboCupBall;


/**
 * A {@link ASimDelegate} that provides a {@link RoboCupBall} which can be moved by TFGH, lifted by
 * holding J, reset by R and beamed to the mouse-position over the field with B
 * 
 * @author Gero
 */
public class SteerableBall extends ASimDelegate {
	
	// --------------------------------------------------------------------------
	// --- variables and constants ----------------------------------------------
	// --------------------------------------------------------------------------
	/** @see SimPhysicsPicker#HEIGHT_OFFSET_DEFAULT */
	private static final float	HEIGHT_OFFSET_DEFAULT	= UnitConversion.cm(10);
	
	private DynamicPhysicsNode	ball;
	private final Vector3f		basePosition				= new Vector3f(0, m(5), 0);
	
	private final Vector2f		mousePositionOnScreen	= new Vector2f();
	
	
	// --------------------------------------------------------------------------
	// --- constructors ---------------------------------------------------------
	// --------------------------------------------------------------------------
	/**
	 * @param core
	 */
	public SteerableBall(ICore core) {
		super(core);
	}
	

	// --------------------------------------------------------------------------
	// --- methods --------------------------------------------------------------
	// --------------------------------------------------------------------------
	@Override
	public void initSystem() {
		
		InputHandler input = core.getInputHandler();
		
		final float force = 0.5f;
		// TESTCONTROLS #############################################################
		input.addAction(new MoveBallAction(new Vector3f(0, 0, force)), InputHandler.DEVICE_KEYBOARD, KeyInput.KEY_F,
				InputHandler.AXIS_NONE, true);
		input.addAction(new MoveBallAction(new Vector3f(0, 0, -force)), InputHandler.DEVICE_KEYBOARD, KeyInput.KEY_H,
				InputHandler.AXIS_NONE, true);
		input.addAction(new MoveBallAction(new Vector3f(force, 0, 0)), InputHandler.DEVICE_KEYBOARD, KeyInput.KEY_G,
				InputHandler.AXIS_NONE, true);
		input.addAction(new MoveBallAction(new Vector3f(-force, 0, 0)), InputHandler.DEVICE_KEYBOARD, KeyInput.KEY_T,
				InputHandler.AXIS_NONE, true);
		

		// Reset
		input.addAction(new InputAction() {
			
			@Override
			public void performAction(InputActionEvent evt) {
				if (ball != null) {
					ball.clearDynamics();
					ball.getLocalTranslation().set(basePosition);
				}
			}
		}, InputHandler.DEVICE_KEYBOARD, KeyInput.KEY_R, InputHandler.AXIS_NONE, true);
		

		// Beam the ball to the position where the mouse is held
		input.addAction(new MouseTrackingAction(), InputHandler.DEVICE_MOUSE, InputHandler.BUTTON_NONE,
				InputHandler.AXIS_ALL, false);
		input.addAction(new InputAction() {
			
			@Override
			public void performAction(InputActionEvent evt) {
				if (ball != null) {
					Vector3f target = CamHelper.calculateMouseOnGround(mousePositionOnScreen);
					target.y = HEIGHT_OFFSET_DEFAULT; // A bit above the ground
					
					ball.clearDynamics();
					ball.getLocalTranslation().set(target);
				}
			}
		}, InputHandler.DEVICE_KEYBOARD, KeyInput.KEY_B, InputHandler.AXIS_NONE, true);
		

		// Jump!
		final float upForce = 1.25f;
		input.addAction(new InputAction() {
			
			@Override
			public void performAction(InputActionEvent evt) {
				if (ball != null) {
					ball.addForce(new Vector3f(0, upForce, 0));
				}
			}
		}, InputHandler.DEVICE_KEYBOARD, KeyInput.KEY_J, InputHandler.AXIS_NONE, true);
	}
	

	/**
	 * Needed to let this {@link ASimDelegate} know which ball to control
	 * @param newBall
	 */
	public void setBall(DynamicPhysicsNode newBall) {
		ball = newBall;
		
		ball.getLocalTranslation().set(basePosition);
		
		ball.setActive(true);
	}
	
	
	/**
	 * Used to apply the move-forces to the ball!
	 * 
	 * @author Gero
	 */
	private class MoveBallAction extends InputAction {
		
		private final Vector3f	direction;
		private final Vector3f	appliedForce	= new Vector3f();
		
		
		public MoveBallAction(Vector3f direction) {
			this.direction = direction;
		}
		

		@Override
		public void performAction(InputActionEvent evt) {
			if (ball != null) {
				appliedForce.set(direction).multLocal((float) Math.pow(1, evt.getTime()));
				ball.addForce(appliedForce);
			}
		}
	}
	
	
	/**
	 * Updates the mouse-position in {@link SteerableBall#mousePositionOnScreen}!
	 */
	private class MouseTrackingAction extends InputAction {
		
		@Override
		public void performAction(InputActionEvent evt) {
			switch (evt.getTriggerIndex()) {
				case 0:
					mousePositionOnScreen.x = evt.getTriggerPosition() * DisplaySystem.getDisplaySystem().getWidth();
					break;
				case 1:
					mousePositionOnScreen.y = evt.getTriggerPosition() * DisplaySystem.getDisplaySystem().getHeight();
					break;
			}
		}
	}
}
