package edu.dhbw.mannheim.tigers.sim.engine.model;


import java.math.BigDecimal;


/**
 * <p>
 * A short thought about units.<br/>
 * <u>Convention:</u> In the 'model.xml', lengths are defined in <b>mms</b>, and weight in <b>gramms</b>.
 * </p>
 * <p>
 * But what about the simulator? For ODE it'd be perfect if 1 unit = 1 m, then everything would make sense and be easy
 * and everything. Problem is, most of our lengths are between 1cm and 1mm, making 1m-units quite long tailed with
 * around 3 decimal places (0.001...). This is 1. uncomfortable, and 2. - which is a bit more disturbing - makes the
 * simulation veeery unstable.
 * </p>
 * <p>
 * <i>Short: A lot of decimal places generate a lot of rounding errors causing parts of the matrix representing the
 * physical space to collapse, which makes the whole simulation inaccurate and buggy (balls are falling through the
 * ground, e.g.); see <a href="http://www.ode.org/ode-latest-userguide.html">ODE</a> for details.</i>
 * </p>
 * <p>
 * The solution: The model is first created with 1mm (from the 'model.xml') = 1 unit, which makes the model quite big.
 * Then the model is scaled down by {@link #MODEL_SCALE_FACTOR}.<br/>
 * <b>This is the factor '1m = unit' / '1mm = model-data-in-xml' to make the model in the simulation fit to ODE's 1m = 1
 * unit needs.</b>
 * </p>
 * <p>
 * The whole problem makes model programming tricky. If you are using models manufactured by {@link ModelFactory}, they
 * are already sized to fit into simulation by convention. <b>If you try to apply translation to them, however, you have
 * to keep in mind what you are doing and use the conversion functions provided by this class, for example!!!</b>
 * </p>
 * 
 * @author Gero
 */
public final class UnitConversion {
	
	// --------------------------------------------------------------------------
	// --- factors --------------------------------------------------------------
	// --------------------------------------------------------------------------
	private static final BigDecimal	TEN						= new BigDecimal(10);
	private static final BigDecimal	HUNDRET					= new BigDecimal(100);
	private static final BigDecimal	THOUSAND					= new BigDecimal(1000);
	

	// --------------------------------------------------------------------------
	// --- units ----------------------------------------------------------------
	// --------------------------------------------------------------------------
	/**
	 * How 1m is represented in the simulator
	 */
	public static final BigDecimal	WORLD_TO_MODEL_SCALE	= new BigDecimal(1);
	
	/**
	 * Scale by which the data in the 'model.xml' is shifted to 1m<br/>
	 * <b>e.g.: DATA_SCALE = 1000 => 1unit in model.xml = 1mm</b>
	 */
	public static final BigDecimal	DATA_SCALE				= new BigDecimal(1000);
	
	/**
	 * The factor by which model-data must be multiplied with to fit in the (ODE) simulation
	 */
	public static final BigDecimal	MODEL_SCALE_FACTOR	= WORLD_TO_MODEL_SCALE.divide(DATA_SCALE);
	

	/**
	 * The factor to make a <b>dm</b> in reality representing a <b>dm</b> in the model
	 */
	public static final BigDecimal	MODEL_DM_FACTOR		= WORLD_TO_MODEL_SCALE.divide(TEN);
	
	/**
	 * The factor to make a <b>cm</b> in reality representing a <b>cm</b> in the model
	 */
	public static final BigDecimal	MODEL_CM_FACTOR		= WORLD_TO_MODEL_SCALE.divide(HUNDRET);
	
	/**
	 * The factor to make a <b>mm</b> in reality representing a <b>mm</b> in the model
	 */
	public static final BigDecimal	MODEL_MM_FACTOR		= WORLD_TO_MODEL_SCALE.divide(THOUSAND);
	
	
	private static final BigDecimal bd(float fl) {
		return new BigDecimal(fl);
	}
	

	private static final BigDecimal bd(double dl) {
		return new BigDecimal(dl);
	}
	

	// --------------------------------------------------------------------------
	// --- conversion to model --------------------------------------------------
	// --------------------------------------------------------------------------
	/**
	 * @return {@link UnitConversion#MODEL_SCALE_FACTOR}
	 */
	public static final float getModelScale() {
		return MODEL_SCALE_FACTOR.floatValue();
	}
	

	/**
	 * @return 1 / {@link UnitConversion#MODEL_SCALE_FACTOR}
	 */
	public static final float getModelScaleInverse() {
		return BigDecimal.ONE.divide(MODEL_SCALE_FACTOR).floatValue();
	}
	

	/**
	 * @return {@link #WORLD_TO_MODEL_SCALE}
	 */
	public static final float getWorldToModelScale() {
		return WORLD_TO_MODEL_SCALE.floatValue();
	}
	

	/**
	 * @param dataLength A length from the 'model.xml'
	 * @return The equivalent length in the simulation
	 */
	public static final float data(float dataLength) {
		return dataToSim(bd(dataLength)).floatValue();
	}
	

	public static final float data(double dataLength) {
		return dataToSim(bd(dataLength)).floatValue();
	}
	

	private static final BigDecimal dataToSim(BigDecimal dataLength) {
		return dataLength.multiply(MODEL_SCALE_FACTOR);
	}
	

	/**
	 * @param m One m in reality
	 * @return One m in the model
	 */
	public static final float m(float m) {
		return bd(m).multiply(WORLD_TO_MODEL_SCALE).floatValue();
	}
	

	/**
	 * @param dm One dm in reality
	 * @return One dm in the model
	 */
	public static final float dm(float dm) {
		return bd(dm).multiply(MODEL_DM_FACTOR).floatValue();
	}
	

	/**
	 * @param cm One cm in reality
	 * @return One cm in the model
	 */
	public static final float cm(float cm) {
		return bd(cm).multiply(MODEL_CM_FACTOR).floatValue();
	}
	

	/**
	 * @param mm One mm in reality
	 * @return One mm in the model
	 */
	public static final float mm(float mm) {
		return bd(mm).multiply(MODEL_MM_FACTOR).floatValue();
	}
	

	// --------------------------------------------------------------------------
	// --- conversion to reality ------------------------------------------------
	// --------------------------------------------------------------------------
	/**
	 * @param mm mm in the model
	 * @return mm in reality
	 */
	public static final float mmToReal(float mm) {
		return mm / MODEL_MM_FACTOR.floatValue();
	}
}
