package edu.dhbw.mannheim.tigers.sim.engine.model.ball;


import org.apache.log4j.Logger;

import com.jme.bounding.BoundingSphere;
import com.jme.math.Quaternion;
import com.jme.math.Vector3f;
import com.jme.renderer.ColorRGBA;
import com.jme.scene.Node;
import com.jme.scene.shape.Sphere;
import com.jme.scene.state.MaterialState;
import com.jme.system.DisplaySystem;
import com.jmex.physics.impl.ode.OdePhysicsSpace;

import edu.dhbw.mannheim.tigers.sim.engine.model.DynamicPhysicsSimNode;
import edu.dhbw.mannheim.tigers.sim.engine.model.IUpdateableModel;
import edu.dhbw.mannheim.tigers.sim.engine.model.TimedModelLocation;
import edu.dhbw.mannheim.tigers.sim.engine.model.UnitConversion;
import edu.dhbw.mannheim.tigers.sim.engine.model.UserData.Value;
import edu.dhbw.mannheim.tigers.sim.engine.model.data.ball.IBallModelData;
import edu.dhbw.mannheim.tigers.sim.engine.model.material.RoboCupMaterial;
import edu.dhbw.mannheim.tigers.sim.engine.model.observer.AModelStateObserver;
import edu.dhbw.mannheim.tigers.sim.engine.model.observer.IModelStateObservable;
import edu.dhbw.mannheim.tigers.sim.engine.model.observer.ModelStateObservable;


/**
 * This class represents a RoboCup-ball
 * 
 * @author Gero
 */
public class RoboCupBall extends DynamicPhysicsSimNode implements IBallModelControl, IUpdateableModel {
	
	// --------------------------------------------------------------------------
	// --- variables and constants ----------------------------------------------
	// --------------------------------------------------------------------------
	private static final long				serialVersionUID	= 773385204745319497L;
	
	protected final Logger					log					= Logger.getLogger(getClass());
	
	private final IBallModelData			data;
	private final IModelStateObservable	observable;
	
	
	// --------------------------------------------------------------------------
	// --- constructors ---------------------------------------------------------
	// --------------------------------------------------------------------------
	/**
	  * @see RoboCupBall
	  * @param space
	  * @param display
	  * @param data
	  */
	public RoboCupBall(OdePhysicsSpace space, DisplaySystem display, IBallModelData data) {
		super(space);
		
		this.data = data;
		
		this.setName("RoboCupBall");
		
		Sphere graphicBall = new Sphere("RoboCupBallBody", new Vector3f(0, 0, 0), 20, 20, data.getRadius());
		graphicBall.setModelBound(new BoundingSphere());
		graphicBall.updateModelBound();
		
		this.attachChild(graphicBall);
		this.setMaterial(RoboCupMaterial.BALL);
		
		MaterialState matTS = display.getRenderer().createMaterialState();
		matTS.setEnabled(true);
		matTS.setDiffuse(ColorRGBA.orange);
		
		this.setRenderState(matTS);
		
		// Phsics
		this.generatePhysicsGeometry();
		this.setCenterOfMass(Vector3f.ZERO);
		this.setMass(data.getMass());
		
		// IModelStateObservable
		this.observable = new ModelStateObservable(getTimedModelLocation());
		
		makePickable(Value.BALL);
	}
	
	
	// --------------------------------------------------------------------------
	// --- IBallModelControl ----------------------------------------------------
	// --------------------------------------------------------------------------
	@Override
	public void setLocation(Vector3f newPosition) {
		this.setLocation(newPosition, true, true, true);
	}
	

	@Override
	public void setLocation(Vector3f newPosition, boolean clearForce, boolean clearTorque, boolean clearVelocities) {
		
		if (clearForce) {
			this.clearForce();
		}
		
		if (clearTorque) {
			this.clearTorque();
		}
		
		if (clearVelocities) {
			this.setLinearVelocity(Vector3f.ZERO);
		}
		
		this.setLocalTranslation(newPosition);
	}
	

	@Override
	public void setRotation(Quaternion quaternion) {
		setRotation(quaternion, true, true, true);
	}
	

	@Override
	public void setRotation(Quaternion quaternion, boolean clearForce, boolean clearTorque, boolean clearVelocities) {
		
		if (clearForce) {
			this.clearForce();
		}
		
		if (clearTorque) {
			this.clearTorque();
		}
		
		if (clearVelocities) {
			this.setLinearVelocity(Vector3f.ZERO);
		}
		
		this.getLocalRotation().set(quaternion);
	}
	
	
	// --------------------------------------------------------------------------
	// --- IModelStateObservable ------------------------------------------------
	// --------------------------------------------------------------------------
	@Override
	public boolean addObservers(AModelStateObserver... observers) {
		return observable.addObservers(observers);
	}
	

	@Override
	public void notifyObservers(TimedModelLocation event) {
		observable.notifyObservers(event);
	}
	

	@Override
	public void notifyObservers() {
		observable.notifyObservers();
	}
	

	@Override
	public boolean removeObserver(AModelStateObserver observer) {
		return observable.removeObserver(observer);
	}
	

	// --------------------------------------------------------------------------
	// --- IUpdateableModel -----------------------------------------------------
	// --------------------------------------------------------------------------
	@Override
	public void update(long cycleNumber, float tpc) {
		
		updateGeometricState(0f, true); // Prevent world-translation from being (nearly) 0:0!
		observable.notifyObservers(getTimedModelLocation());
		
		// Vector3f vec = this.getWorldTranslation();
		// if (vec.x < 0.0001f && vec.x > -0.0001f &&
		// vec.z < 0.0001f && vec.z > -0.0001f) {
		// log.debug("##### [Ball] 0:0!");
		// }
	}
	

	// --------------------------------------------------------------------------
	// --- getter ---------------------------------------------------------------
	// --------------------------------------------------------------------------
	@Override
	public float getHeight() {
		return data.getRadius() * 2;
	}
	

	/**
	 * @return The radius. (In 'model.xml' dimensions, so you need to explicitly use {@link UnitConversion#data(float)}
	 *         to convert to simulation dimensions!!!)
	 */
	@Override
	public float getRadius() {
		return data.getRadius();
	}
	

	@Override
	public Node getNode() {
		return this;
	}
}
