package edu.dhbw.mannheim.tigers.sim.engine.model.bot;


import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;

import org.apache.log4j.Logger;

import edu.dhbw.mannheim.tigers.sim.engine.com.ERoboCupTeam;


/**
 * This class generates new, unique SSL-{@link BotID}s. Can be kindly asked to create them using the
 * given {@link Integer}s by calling {@link #enqueue(List, ERoboCupTeam))}.
 * 
 * @author Gero
 */
public class BotIDProvider {
	
	// --------------------------------------------------------------------------
	// --- variables and constants ----------------------------------------------
	// --------------------------------------------------------------------------
	protected final Logger				log					= Logger.getLogger(getClass());
	
	private static BotIDProvider		instance				= null;
	
	private final IdStorage				storeYellow			= new IdStorage();
	private final IdStorage				storeBlue			= new IdStorage();
	
	
	// --------------------------------------------------------------------------
	// --- constructors ---------------------------------------------------------
	// --------------------------------------------------------------------------
	/**
	 * @see BotIDProvider
	 */
	private BotIDProvider() {
	}
	

	public static synchronized BotIDProvider getInstance() {
		if (instance == null) {
			instance = new BotIDProvider();
		}
		return instance;
	}
	

	// --------------------------------------------------------------------------
	// --- methods --------------------------------------------------------------
	// --------------------------------------------------------------------------
	/**
	 * @see #getNewUniqueID(ERoboCupTeam)
	 */
	public BotID getNewBotID(ERoboCupTeam team) {
		return getNewUniqueID(team);
	}
	

	/**
	 * @param team	The {@link ERoboCupTeam} to create a new id for
	 * @return A new ID unique for the given team
	 */
	private synchronized BotID getNewUniqueID(ERoboCupTeam team) {
		final IdStorage store = getStore(team);
		BotID result = null;
		
		if (store.queue.isEmpty()) { // Generate new ID
		
			boolean unique = false;
			while (!unique) {
				result = new BotID(++store.lastConsistentID, team);
				
				if (store.isUnique(result)) {
					unique = true;
				}
			}
			
		} else { // Get one from the queue!
		
			// Already checked for uniqueness if they are in the queue! (see below)
			result = store.queue.pop();
		}
		
		store.assignedIDs.add(result);
		
		return result;
	}
	

	/**
	 * The BotIDProvider will try to use the given {@link Integer} to create the next {@link BotID}s
	 * enqueued with {@link #getNewUniqueID(ERoboCupTeam))}
	 * 
	 * @param wishList
	 * @param team
	 */
	public synchronized void enqueue(List<Integer> wishList, ERoboCupTeam team) {
		final IdStorage store = getStore(team);
		
		// Validate
		BotID tempID = null;
		for (Integer wish : wishList) {
			tempID = new BotID(wish, team);
			if (store.isUnique(tempID)) {
				store.queue.add(tempID);
			} else {
				log.warn("BotID with value '" + tempID.getIntValue() + "' is already assigned or queued for team " + team + "!");
			}
		}
	}
	

	/**
	 * Clears all assigned and queued IDs!
	 */
	public synchronized void reset() {
		for (ERoboCupTeam team : ERoboCupTeam.values()) {
			getStore(team).reset();
		}
	}
	

	/**
	 * Calls {@link #reset()}
	 */
	public void cleanup() {
		reset();
	}
	
	
	private IdStorage getStore(ERoboCupTeam team) {
		switch (team) {
			case BLUE_TEAM:
				return storeBlue;
				
			default:
				log.warn("Team '" + team + "' unexpected, taking yellow!");
			case YELLOW_TEAM:
				return storeYellow;
		}
	}
	
	
	/**
	 * Container for everything which is necessary to store and create new unique ids.
	 * 
	 * @author Gero
	 */
	private static class IdStorage {
		private final HashSet<BotID>		assignedIDs			= new HashSet<BotID>();
		private int								lastConsistentID	= 0;
		private final LinkedList<BotID>	queue					= new LinkedList<BotID>();
		
		
		private boolean isUnique(BotID id) {
			return !(assignedIDs.contains(id) || queue.contains(new Integer(id.getIntValue())));
		}
		
		private void reset() {
			assignedIDs.clear();
			queue.clear();
			
			lastConsistentID = 0;
		}
	}
}
