package edu.dhbw.mannheim.tigers.sim.engine.model.bot.parts;


import static edu.dhbw.mannheim.tigers.sim.engine.model.UnitConversion.data;
import static edu.dhbw.mannheim.tigers.sim.engine.model.UnitConversion.mm;

import java.util.ArrayList;
import java.util.List;

import com.jme.bounding.BoundingBox;
import com.jme.math.FastMath;
import com.jme.math.Vector3f;
import com.jme.renderer.ColorRGBA;
import com.jme.scene.shape.Box;
import com.jme.system.DisplaySystem;
import com.jmex.physics.DynamicPhysicsNode;
import com.jmex.physics.Joint;
import com.jmex.physics.RotationalJointAxis;
import com.jmex.physics.geometry.PhysicsBox;

import edu.dhbw.mannheim.tigers.sim.engine.model.ModelHelper;
import edu.dhbw.mannheim.tigers.sim.engine.model.bot.IBotPartsOwner;
import edu.dhbw.mannheim.tigers.sim.engine.model.bot.tiger.TigerBotModel;
import edu.dhbw.mannheim.tigers.sim.engine.model.material.RoboCupMaterial;


/**
 * This class defines a kicking-{@link ABotModelPart} of the {@link TigerBotModel} that is able to lob balls
 * 
 * @author Gero
 */
public class Lobber extends ABotModelPart {
   
	// --------------------------------------------------------------------------
	// --- variables and constants ----------------------------------------------
	// --------------------------------------------------------------------------
	// Physicals
	private final DynamicPhysicsNode		lobber;
	private final RotationalJointAxis	lobMotor;
	
	// Owner
	private final IKickerOwner				kOwner;
	
	// Lobber sizes
	private static final float				PIVOT_FROM_CENTER	= mm(54f);	// Grundklotz, 04.11 (C)
	private static final float				BASE_PLATE_DEPTH	= mm(4);
	private static final float				PIVOT_OVER_BOTTOM	= mm(8.5f) + BASE_PLATE_DEPTH;	// Grundklotz, 04.11 (C)
																															

	private static final float				LOBBER_LENGTH		= 10;
	private static final float				LOBBER_DEPTH		= 2;	// 5
	private static final float				LOBBER_WIDTH		= 60;
	private static final float				LOBBER_ARM_LENGTH	= mm(21) - data(LOBBER_LENGTH / 2);
	private static final float				LOBBER_MASS			= 0.007f;	// kg
																															

	// Lobber movement
	private static final float				ANGLE_MIN			= FastMath.DEG_TO_RAD * 15;
	private static final float				ANGLE_MAX			= FastMath.DEG_TO_RAD * 45;
	private static final float				MAX_ACC				= 100f;
	private static final float				MAX_VEL				= 250f;
	private static final float				RESET_VEL			= 5f;
	
//	private boolean							readyToShoot		= false;
	private static final float				ALLOWED_EPSILON	= FastMath.DEG_TO_RAD * 0.5f;
//	private static final float				TIME_TO_SHOOT_MAX	= 0.5f;	// s
//	private float								timeToShoot			= 0;	// s
																	
	
	// Kick														
	private final SimpleCapModel			cap;
	
	/** [s] */
	private static final float	MAX_TIME_TO_RESET	= 0.5f;
	/** Needed to give the lobber enough time to kick before resetting it */
	private float					timeToReset			= 0;
	private volatile KickMode	currentMode			= KickMode.NONE;
	
	/** [%] */
	private volatile int			kickForce			= 0;
                                                                                         

	// --------------------------------------------------------------------------
	// --- constructors ---------------------------------------------------------
	// --------------------------------------------------------------------------
   /**
     * @see Lobber
     * @param owner
     * @param kOwner
     * @param display
     * @param body
     * @param bodyHeight
     * @param bodyToGroundDist
     * @param com
     */
   public Lobber(IBotPartsOwner owner, IKickerOwner kOwner, DisplaySystem display, DynamicPhysicsNode body,
         float bodyHeight, float bodyToGroundDist, Vector3f com) {
      super(owner, display, body, com);
   	
      this.kOwner = kOwner;
   	this.cap = kOwner.getCapModel();

      // Create lobber
      lobber = space.createDynamicNode();
      
      // ...visual...
      final Box vLobber = new Box("vLobber", new Vector3f(), LOBBER_LENGTH / 2, LOBBER_DEPTH / 2,
            LOBBER_WIDTH / 2);
      vLobber.setModelBound(new BoundingBox());
      vLobber.updateModelBound();
      
      ModelHelper.color(vLobber, display, ColorRGBA.lightGray.clone());
      lobber.attachChild(vLobber);
      

      // ...physical...
      final PhysicsBox phLobber = lobber.createBox("phLobber");
      phLobber.setMaterial(RoboCupMaterial.BOT_KICKER);
      phLobber.setLocalScale(new Vector3f(LOBBER_LENGTH, LOBBER_DEPTH, LOBBER_WIDTH));
      lobber.setMass(LOBBER_MASS);
      

      // Location & Rotation
      lobber.setLocalTranslation(LOBBER_ARM_LENGTH, data(bodyToGroundDist + (bodyHeight / 2)), 0);
      owner.attachNode(lobber);
      

      // Attach to the bot
      final Joint lobPivot = space.createJoint();
      lobMotor = lobPivot.createRotationalAxis();
      lobMotor.setDirection(Vector3f.UNIT_Z.clone());
      lobMotor.setPositionMinimum(ANGLE_MIN);
      lobMotor.setPositionMaximum(ANGLE_MAX);
      lobMotor.setAvailableAcceleration(MAX_ACC);
      lobMotor.setDesiredVelocity(RESET_VEL); // A matter of initialization; we have to move it into
                                       // "null"-position
      
      // This little trick helps us to avoid the creation of a pivot-node and attaching it statically to the body:
      // We move the CoM to the pivot-point for our lobber (as bodys are attached to each other via their CoM), and...
      final Vector3f pivot = new Vector3f(PIVOT_FROM_CENTER, 0, data(bodyHeight / 2)
            - PIVOT_OVER_BOTTOM);
      pivot.negateLocal();
      Vector3f bodyCom = new Vector3f(com.x, com.z, -com.y);	// Convert to body-system
      pivot.addLocal(bodyCom);	// Offset
      
      body.setCenterOfMass(pivot);
      
      // ...simply attach them to each other.
      lobPivot.attach(body, lobber);
      
      // Afterwards we can re-adjust the CoM without any trouble, and the lobber staying where it should stay.
      body.setCenterOfMass(Vector3f.ZERO.clone());
   }
	
	
	// --------------------------------------------------------------------------
	// --- methods --------------------------------------------------------------
	// --------------------------------------------------------------------------
	@Override
	public void doUpdate(long cycleNumber, float tpc) {
		
		switch (currentMode) {
			
			case ARM:
				if (kickForce > 0 && cap.getCurrentLoad() > 0 && kOwner.isBallDetected()) {
					// Ball's there, go for it!
					doKick();
				} else {
					doResetKicker(tpc);
				}
				break;
			
			case FORCE:
				if (kickForce > 0 && cap.getCurrentLoad() > 0) {
					// Kick!
					doKick();
				} else {
					// Reset plunger
					doResetKicker(tpc);
				}
				break;
			
			case DISARM:
			default:
				doResetKicker(tpc);
				break;
		}
	}
	

	private void doKick() {
		lobMotor.setDesiredVelocity(-kickForce * cap.getLoadLevel() * MAX_VEL);
		cap.completeDischarge();
		kickForce = 0;
		timeToReset = MAX_TIME_TO_RESET;
		currentMode = KickMode.NONE;
	}
	

	private void doResetKicker(float tpc) {
		
		if (timeToReset > 0) {
			timeToReset -= tpc;
		}
		if (timeToReset <= 0) {
			if (lobMotor.getPosition() <= lobMotor.getPositionMinimum() + ALLOWED_EPSILON) {
				lobMotor.setDesiredVelocity(RESET_VEL);
			} else if (lobMotor.getPosition() >= lobMotor.getPositionMaximum() - ALLOWED_EPSILON) {
				lobMotor.setDesiredVelocity(0f);
			}
		}
	}


	/**
	 * Fires the plunger HARD
	 * 
	 * @param percent
	 * @param mode
	 */
	public void fire(int percent, KickMode mode) {
		
		// Normalize percent
		if (percent < 0) {
			percent = 0;
		} else if (percent > 100) {
			percent = 100;
		}
		
		kickForce = percent;
		currentMode = mode;
	}


	@Override
	public List<DynamicPhysicsNode> getPhysicNodes() {
		final ArrayList<DynamicPhysicsNode> result = new ArrayList<DynamicPhysicsNode>();
		result.add(lobber);
		return result;
	}
}
