package edu.dhbw.mannheim.tigers.sim.engine.model.bot.parts;

import com.jme.math.FastMath;


/**
 * <pre>
 * Ladezeit: 10s
 * Formel (Umax = 200[V]): U = 200 * (1-e^(-t/2))
 * Formel (angepasst): % = 100 * (1-e^(-t/2))
 * 
 * Entladezeit: 25ms
 * pro Sekunde: 40x
 * Formel: linear
 * </pre>
 * 
 * @author Gero
 */
public class SimpleCapModel implements IBotModelPart {
	
	// --------------------------------------------------------------------------
	// --- variables and constants ----------------------------------------------
	// --------------------------------------------------------------------------
	/** [V] */
	private float			currentU		= 0;
	/** [V] */
	private final float	maxU;
	/** [s] */
	private float			chargeTime	= 0;
	private boolean		charge		= false;
	private final float	tau;
	
	
	// --------------------------------------------------------------------------
	// --- constructors ---------------------------------------------------------
	// --------------------------------------------------------------------------
	/**
	  * @see SimpleCapModel
	  * @param maxU [V]
	  * @param tau (= R * C)
	  */
	public SimpleCapModel(float maxU, float tau) {
		this.maxU = maxU;
		this.tau = tau;
	}
	

	// --------------------------------------------------------------------------
	// --- methods --------------------------------------------------------------
	// --------------------------------------------------------------------------
	/**
	 * Charge and discharge
	 */
	@Override
	public void doUpdate(long cycleNumber, float tpc) {

		if (charge) {
			chargeTime += tpc;
			currentU = maxU * (1 - FastMath.exp(-chargeTime / tau));
		} else {
			chargeTime = 0;
			if (currentU > 0) {
				currentU -= tpc * 40 * maxU;
				
				if (currentU < 0) {
					currentU = 0;
				}
			}
		}
	}

	
	// --------------------------------------------------------------------------
	// --- getter/setter --------------------------------------------------------
	// --------------------------------------------------------------------------
	public synchronized void charge(boolean charge) {
		this.charge = charge;
	}
	
	
	/**
	 * Sets the current charge ({@link #currentU}) to 0 (zero)
	 * 
	 * @return currentU
	 */
	public synchronized float completeDischarge() {
		float result = currentU;
		chargeTime = 0;
		currentU = 0;
		return result;
	}
	

	/**
	 * @return current U, between 0 and {@link #maxU} [V]
	 */
	public synchronized float getCurrentLoad() {
		return currentU;
	}
	
	
	/**
	 * @return {@link #currentU} / {@link #maxU} [%]
	 */
	public synchronized float getLoadLevel() {
		return currentU / maxU;
	}
	
	
	/**
	 * @return Whether getLoadLevel() < 0.98 (98%)
	 */
	public synchronized boolean isLoading() {
		return getLoadLevel() < 0.98f;
	}
}
