package edu.dhbw.mannheim.tigers.sim.engine.model.field;


import com.jme.bounding.BoundingBox;
import com.jme.math.Quaternion;
import com.jme.math.Vector3f;
import com.jme.renderer.ColorRGBA;
import com.jme.renderer.Renderer;
import com.jme.scene.SharedMesh;
import com.jme.scene.shape.Box;
import com.jme.system.DisplaySystem;
import com.jmex.physics.StaticPhysicsNode;
import com.jmex.physics.impl.ode.OdePhysicsSpace;

import edu.dhbw.mannheim.tigers.sim.engine.model.ASimNode;
import edu.dhbw.mannheim.tigers.sim.engine.model.ModelHelper;
import edu.dhbw.mannheim.tigers.sim.engine.model.data.field.IRoboCupFieldData;
import edu.dhbw.mannheim.tigers.sim.engine.model.material.RoboCupMaterial;
import edu.dhbw.mannheim.tigers.sim.engine.resources.TextureProvider.ETexIndex;
import edu.dhbw.mannheim.tigers.sim.exceptions.SimModelException;
import edu.dhbw.mannheim.tigers.sim.exceptions.SimModelResourceException;


/**
 * Represents the wall that surrounds the RoboCup-field
 * 
 * @author Gero
 */
public class SurrWall extends ASimNode {

	// --------------------------------------------------------------------------
	// --- variables and constants ----------------------------------------------
	// --------------------------------------------------------------------------
	private static final long serialVersionUID = -4410927076018894529L;
	
	// Field
	private final float totalFieldLength;
	private final float totalFieldWidth;
	private final float fieldDepth;
	
	// Wall
	private final float wallWidth;
	private final float wallHeight;
	
	
	// --------------------------------------------------------------------------
	// --- constructors ---------------------------------------------------------
	// --------------------------------------------------------------------------
	/**
	  * @see SurrWall
	  * @param space
	  * @param display
	  * @param data
	  * @param fieldDepth
	  * @throws SimModelResourceException
	  * @throws SimModelException
	  */
	public SurrWall(OdePhysicsSpace space, DisplaySystem display, IRoboCupFieldData data, float fieldDepth) throws SimModelResourceException, SimModelException {
		super(space, display, "SurrWall");
		
		this.totalFieldLength = data.getTotalFieldLength();
		this.totalFieldWidth = data.getTotalFieldWidth();
		this.fieldDepth = fieldDepth;
		this.wallWidth = data.getSurrWallWidth();
		this.wallHeight = data.getSurrWallHeight();
		
		createPhysicalNode();
	}


	// --------------------------------------------------------------------------
	// --- methods --------------------------------------------------------------
	// --------------------------------------------------------------------------
	@Override
	protected void createPhysicalNode() throws SimModelResourceException, SimModelException {
		/*
		 * Surroundings
		 * Two simple solid boxes, each shared two times.
		 *   ---
		 *   | |l
		 *   | |
		 *   ---
		 *   w
		 */		
		// #####	The w walls
		final float halfWallLengthW = (totalFieldWidth / 2) + wallWidth;
		final float halfFieldLength = totalFieldLength / 2;
		
		final Box wWall = new Box("wWall", new Vector3f(-halfWallLengthW, -fieldDepth, 0),
				new Vector3f(halfWallLengthW, wallHeight, wallWidth));
		
		// Color and model bounds	
		wWall.setModelBound(new BoundingBox());
		wWall.updateModelBound();
		
		wWall.setDefaultColor(ColorRGBA.white);
		ModelHelper.applyTexture(wWall, display, ETexIndex.DHBW_HEADER, 3, 1);
		
		
		// Share physically...
		final SharedMesh wWallA = new SharedMesh("wWallA", wWall);
		wWallA.setLocalTranslation(0, 0, halfFieldLength);

		final StaticPhysicsNode phWWallA = space.createStaticNode();
		phWWallA.attachChild(wWallA);		
		phWWallA.setMaterial(RoboCupMaterial.WALL);
		phWWallA.generatePhysicsGeometry();

		this.attachChild(phWWallA);

		final SharedMesh wWallB = new SharedMesh("wWallB", wWall);
		final Quaternion wallTurnaround = new Quaternion();	// Turn it by 180, because its not build _around_ 
		wallTurnaround.fromAngleAxis(rad(180), Vector3f.UNIT_Y);	// but _in front of_ the center
		wWallB.setLocalRotation(wallTurnaround);
		wWallB.setLocalTranslation(0, 0, -halfFieldLength);

		final StaticPhysicsNode phWWallB = space.createStaticNode();
		phWWallB.attachChild(wWallB);		
		phWWallB.setMaterial(RoboCupMaterial.WALL);
		phWWallB.generatePhysicsGeometry();
		
		this.attachChild(phWWallB);
		
		
		// #####	Finally: The l walls!!!
		final float halfWallLengthL = (totalFieldLength / 2);
		final float halfFieldWidth = totalFieldWidth / 2;
		
		final Box lWall = new Box("lWall", new Vector3f(0, -fieldDepth, -halfWallLengthL),
				new Vector3f(wallWidth, wallHeight, halfWallLengthL));
		lWall.setModelBound(new BoundingBox());
		lWall.updateModelBound();

		// Color and model bounds
		lWall.setDefaultColor(ColorRGBA.white);
		ModelHelper.applyTexture(lWall, display, ETexIndex.DHBW_HEADER, 5, 1);
		
		
		// Share...
		final SharedMesh lWallP = new SharedMesh("lWallP", lWall);
		lWallP.setLocalTranslation(halfFieldWidth, 0, 0);
		
		final StaticPhysicsNode phLWallP = space.createStaticNode();
		phLWallP.attachChild(lWallP);		
		phLWallP.setMaterial(RoboCupMaterial.WALL);
		phLWallP.generatePhysicsGeometry();
		
		final SharedMesh lWallN = new SharedMesh("lWallN", lWall);
		lWallN.setLocalRotation(wallTurnaround);								// Turn it by 180, because its not build _around_ 
		lWallN.setLocalTranslation(-halfFieldWidth, 0, 0);	// but _in front of_ the center
		
		final StaticPhysicsNode phLWallN = space.createStaticNode();
		phLWallN.attachChild(lWallN);		
		phLWallN.setMaterial(RoboCupMaterial.WALL);
		phLWallN.generatePhysicsGeometry();
		
		this.attachChild(phLWallP);
		this.attachChild(phLWallN);
		
		// Finish Surroundings
		this.setRenderQueueMode(Renderer.QUEUE_OPAQUE);
	}
}
