package edu.dhbw.mannheim.tigers.sim.engine.model.material;


import com.jme.math.Quaternion;
import com.jme.math.Vector2f;
import com.jme.math.Vector3f;
import com.jme.util.export.Savable;
import com.jmex.physics.DynamicPhysicsNode;
import com.jmex.physics.PhysicsCollisionGeometry;
import com.jmex.physics.contact.ContactCallback;
import com.jmex.physics.contact.ContactHandlingDetails;
import com.jmex.physics.contact.PendingContact;
import com.jmex.physics.material.Material;
import com.jmex.physics.material.MaterialContactCallback;

import edu.dhbw.mannheim.tigers.sim.engine.ICore;
import edu.dhbw.mannheim.tigers.sim.engine.delegates.SimPhysicsDelegate;
import edu.dhbw.mannheim.tigers.sim.engine.model.bot.tiger.TigerBotModel.WheelAxis;


/**
 * This class replaces the standard implementation {@link MaterialContactCallback} from jmePhysics2
 * with the purpose of providing anisotropic friction for the wheels of the simulated robots.
 * The actual instances are set in the {@link SimPhysicsDelegate#initSystem()}
 * 
 * @author Gero
 */
public class MaterialContactCallbackXT implements ContactCallback {
	// --------------------------------------------------------------------------
	// --- variables and constants ----------------------------------------------
	// --------------------------------------------------------------------------
	
	private static MaterialContactCallbackXT	instance;
	
	// private final ICore core;
	
	private final Vector2f							motion				= new Vector2f();
	private final Vector3f							materialMotion		= new Vector3f();
	private final Vector3f							normal				= new Vector3f();
	private final Vector3f							dir1					= new Vector3f();
	private final Vector3f							dir2					= new Vector3f();
	private final Quaternion						frictionRotation	= new Quaternion();
	
	
	// private final Vector2f slip = new Vector2f();
	// private final Map<DynamicPhysicsNode, PolylineCurve> map = new HashMap<DynamicPhysicsNode,
	// PolylineCurve>();
	// public DynamicPhysicsNode wheelUnderTest;
	

	// --------------------------------------------------------------------------
	// --- constructors ---------------------------------------------------------
	// --------------------------------------------------------------------------
	/**
	 * @see MaterialContactCallbackXT
	 * @param core
	 */
	private MaterialContactCallbackXT(ICore core) {
		// this.core = core;
	}
	

	public static synchronized MaterialContactCallbackXT getInstance(ICore core) {
		if (instance == null) {
			instance = new MaterialContactCallbackXT(core);
		}
		return instance;
	}
	

	// --------------------------------------------------------------------------
	// --- methods --------------------------------------------------------------
	// --------------------------------------------------------------------------
	@Override
	public boolean adjustContact(PendingContact contact) {
		Material m1 = (contact.getGeometry1() != null) ? contact.getGeometry1().getMaterial() : contact.getNode1()
				.getMaterial();
		Material m2 = (contact.getGeometry2() != null) ? contact.getGeometry2().getMaterial() : contact.getNode2()
				.getMaterial();
		ContactHandlingDetails details = m1.getContactHandlingDetails(m2);
		if (details != null) {
			contact.copy(details);
			
			if (!Float.isNaN(contact.getSpringConstant())) {
				float springPenetrationDepth = m1.getSpringPenetrationDepth() + m2.getSpringPenetrationDepth();
				float springPercent = 1 - contact.getPenetrationDepth() / springPenetrationDepth;
				if (springPercent > 0) {
					contact.setSpringConstant(contact.getSpringConstant() / springPercent);
				} else {
					contact.setSpringConstant(Float.NaN);
				}
			}
		}
		
		// PhysicsCollisionGeometry wheel = null;
		DynamicPhysicsNode pWheel = null;
		if (m1 == RoboCupMaterial.BOT_WHEEL && m2 == RoboCupMaterial.GROUND) {
			// wheel = contact.getGeometry1();
			pWheel = (DynamicPhysicsNode) contact.getNode1();
		} else if (m2 == RoboCupMaterial.BOT_WHEEL && m1 == RoboCupMaterial.GROUND) {
			// wheel = contact.getGeometry2();
			pWheel = (DynamicPhysicsNode) contact.getNode2();
		}
		
		// Apply surface-motion
		applyMotion(m1, contact, contact.getGeometry1());
		applyMotion(m2, contact, contact.getGeometry2());
		

		if (pWheel != null) {
			contact.getContactNormal(normal);
			
			Quaternion rot = pWheel.getWorldRotation();
			rot.getRotationColumn(2, dir2); // Get the wheels Z axis in world coordinate space (as it
														// equals the wheels rotation axis!)
			dir2.y = 0; // Project it onto the ground (X-Z-plane)
			dir2.normalizeLocal();
			dir2.cross(normal, dir1); // This is not really correct. If we take the X-Z plane to
												// project our drift on, the normal is Y...
			dir1.normalizeLocal();
			
			WheelAxis wa = null;
			Savable s = pWheel.getUserData(WheelAxis.WHEEL_AXIS);
			if (s != null && s.getClassTag().equals(WheelAxis.class)) {
				wa = (WheelAxis) s;
				
				if (wa.axis.getVelocity() > 0) {
					dir1.negateLocal();
				}
			}
			
			contact.setFrictionDirection(dir1);
			contact.setSecondaryFrictionDirection(dir2);
			

			// // Debug
			// Vector3f from = new Vector3f();
			// contact.getContactPosition(from);
			// from.y = cm(5);
			//
			// Vector3f to = new Vector3f(dir1);
			// to.multLocal(0.125f);
			// to.addLocal(from);
			// to.y = cm(5);
			//
			// debugLine(from, to, pWheel, ColorRGBA.orange);
			//
			// i++;
			// if (i % 10 == 0 && pWheel == wheelUnderTest) {
			// // System.out.println("#####");
			// // System.out.println("Contact: " + from);
			// // System.out.println("Normal: " + to);
			// // System.out.println("Avel: " + wheelRotation);
			// }
		}
		return details != null;
	}
	

	// private int i = 0;
	//
	// public void debugLine(Vector3f from, Vector3f to, DynamicPhysicsNode dpn, ColorRGBA... color)
	// {
	// PolylineCurve curve = map.get(dpn);
	// if (curve != null) {
	// curve.removeFromParent();
	// }
	//
	// // Draw
	// curve = ModelHelper.draw2DLine(core.getDynamics(), from, to, color);
	//
	// map.put(dpn, curve);
	// }
	

	private void applyMotion(Material material, PendingContact contact, PhysicsCollisionGeometry geometry) {
		material.getSurfaceMotion(materialMotion);
		if (materialMotion.x != 0 || materialMotion.y != 0 || materialMotion.z != 0) {
			contact.getContactNormal(normal);
			geometry.getWorldRotation().multLocal(materialMotion);
			
			contact.getSurfaceMotion(motion);
			if (motion.x != 0 || motion.y != 0) {
				contact.getFrictionDirection(dir1);
				if (Float.isNaN(dir1.x)) {
					contact.getDefaultFrictionDirections(dir1, dir2);
					contact.setFrictionDirection(dir1);
				} else {
					dir1.cross(normal, dir2);
				}
				frictionRotation.fromAxes(dir1, dir2, normal);
				frictionRotation.inverseLocal(); // this is expensive :(
				frictionRotation.multLocal(materialMotion);
				
				motion.x += materialMotion.x;
				motion.y += -materialMotion.y;
			} else {
				dir1.set(materialMotion);
				dir1.scaleAdd(-dir1.dot(normal), normal, dir1);
				float amount = dir1.length();
				contact.setFrictionDirection(dir1.normalizeLocal());
				motion.x = amount;
				motion.y = 0;
			}
			

			contact.setSurfaceMotion(motion);
		}
	}
}