package edu.dhbw.mannheim.tigers.sim.engine.model.material;

import java.util.Collection;
import java.util.HashMap;

import com.jme.math.Vector2f;
import com.jme.math.Vector3f;
import com.jme.util.export.Savable;
import com.jmex.physics.contact.MutableContactInfo;
import com.jmex.physics.material.Material;


/**
 * This class holds the {@link Material}s which are used in the simulation <br/>
 * <strong>Remember:</strong> When adding new materials, add new ContactInformation to - at least! - the ball! <br/>
 * @author Gero
 */
public abstract class RoboCupMaterial extends Material implements Savable {
	
	// --------------------------------------------------------------------------
	// --- variables and constants ----------------------------------------------
	// --------------------------------------------------------------------------
	private static final long						serialVersionUID	= -330822916137553262L;
	

	private static HashMap<String, Material>	materials			= new HashMap<String, Material>();
	
	
	public static Collection<Material> getMaterials() {
		return materials.values();
	}
	

	// --------------------------------------------------------------------------
	// --- constructors ---------------------------------------------------------
	// --------------------------------------------------------------------------
	public RoboCupMaterial(String type) {
		super(type.toString());
		
		doSetProperties();
		
		materials.put(type, this);
	}
	

	protected abstract void doSetProperties();
	
	
	// --------------------------------------------------------------------------
	// --- connections ----------------------------------------------------------
	// --------------------------------------------------------------------------
	private static final MutableContactInfo BALL__WALL = new MutableContactInfo();
	private static final MutableContactInfo BALL__GROUND = new MutableContactInfo();
	private static final MutableContactInfo BALL__NET = new MutableContactInfo();
	private static final MutableContactInfo BALL__BOT_BODY = new MutableContactInfo();
	private static final MutableContactInfo BALL__BOT_DRIBBLER = new MutableContactInfo();
	private static final MutableContactInfo BALL__BOT_KICKER = new MutableContactInfo();
	private static final MutableContactInfo BALL__BOT_BODY_KNOB = new MutableContactInfo();
	
	private static final MutableContactInfo BOT_WHEEL__GROUND = new MutableContactInfo();
	private static final MutableContactInfo BOT_BODY__GROUND = new MutableContactInfo();
	private static final MutableContactInfo BOT_BODY_KNOB__GROUND = new MutableContactInfo();
	private static final MutableContactInfo BOT_KICKER__GROUND = new MutableContactInfo();
	
	private static final MutableContactInfo BOT_WHEEL__BOT_BODY = new MutableContactInfo();
	private static final MutableContactInfo BOT_WHEEL__BOT_DRIBBLER = new MutableContactInfo();
	private static final MutableContactInfo BOT_WHEEL__BOT_KICKER = new MutableContactInfo();
	private static final MutableContactInfo BOT_WHEEL__BOT_WHEEL = new MutableContactInfo();
	
	static {
		// ------	BALL vs WALL	------
		BALL__WALL.setBounce(0.8f);
		BALL__WALL.setMu(0.1f);
		
		
		// ------	BALL vs GROUND	------
		BALL__GROUND.setBounce(0.8f);
		BALL__GROUND.setMu(0.025f);
		
		
		// ------	BALL vs NET	------
		BALL__NET.setBounce(0);
		BALL__NET.setMu(0.1f);
		BALL__NET.setDampingCoefficient(0.7f);
		
		
		// ------	BALL vs BOT_BODY	------
		BALL__BOT_BODY.setMu(0.5f);
		BALL__BOT_BODY.setBounce(0.5f);
		
		
		// ------	BALL vs BOT_DRIBBLER	------
		BALL__BOT_DRIBBLER.setMu(1);
		BALL__BOT_DRIBBLER.setDampingCoefficient(100);
		BALL__BOT_DRIBBLER.setSpringConstant(0.001f);
		BALL__BOT_DRIBBLER.setBounce(0);
		BALL__BOT_DRIBBLER.setSlip(new Vector2f(1, 1));
		
		
		// ------	BALL vs BOT_KICKER	------
		BALL__BOT_KICKER.setMinimumBounceVelocity(1f);
		BALL__BOT_KICKER.setBounce(0.2f);
		BALL__BOT_KICKER.setMu(0.5f);
		
		
		// ------	BALL vs BOT_KNOB	------
		BALL__BOT_BODY_KNOB.setMu(0.5f);
		BALL__BOT_BODY_KNOB.setBounce(0.2f);

		
		// ------	BOT_WHEEL vs GROUND	------
		// NaN caused the contact to ignore spring and damping!
		// If they are set, they are combined with the materials penetration-depth
		// in the contact-callback!!!
		BOT_WHEEL__GROUND.setBounce(1.0f);
		BOT_WHEEL__GROUND.setMinimumBounceVelocity(0.1f);
		
		BOT_WHEEL__GROUND.setDampingCoefficient(0.002f);
		BOT_WHEEL__GROUND.setSpringConstant(Float.NaN);
		
		BOT_WHEEL__GROUND.setMu(Float.POSITIVE_INFINITY);	//100);	//1.8f);
		BOT_WHEEL__GROUND.setMuOrthogonal(0.1f);
		
		BOT_WHEEL__GROUND.setFrictionDirection(new Vector3f(Float.NaN, Float.NaN, Float.NaN));
		BOT_WHEEL__GROUND.setSecondaryFrictionDirection(new Vector3f(0, 0, 0));
		
		BOT_WHEEL__GROUND.setSlip(new Vector2f(0.01f, 0.001f));	//0.05f, 100.00f));
		BOT_WHEEL__GROUND.setSurfaceMotion(new Vector2f(0, 0));

		
		// ------	BOT_BODY vs GROUND	------
		BOT_BODY__GROUND.setMu(0);

		
		// ------	BOT_BODY_KNOB vs GROUND	------
		BOT_BODY_KNOB__GROUND.setMu(0); // Float.POSITIVE_INFINITY);
		BOT_BODY_KNOB__GROUND.setBounce(0.15f);

		
		// ------	BOT_KICKER vs GROUND	------
		BOT_KICKER__GROUND.setBounce(0);
		BOT_KICKER__GROUND.setDampingCoefficient(0.2f);
		BOT_KICKER__GROUND.setMinimumBounceVelocity(0);
		BOT_KICKER__GROUND.setSpringConstant(0);
		BOT_KICKER__GROUND.setMu(0);
		BOT_KICKER__GROUND.setMuOrthogonal(0);
		BOT_KICKER__GROUND.setFrictionDirection(new Vector3f(Float.NaN, Float.NaN, Float.NaN));
		BOT_KICKER__GROUND.setSecondaryFrictionDirection(new Vector3f(0, 0, 0));
		BOT_KICKER__GROUND.setSlip(new Vector2f(0.05f, 0.001f));
		BOT_KICKER__GROUND.setSurfaceMotion(new Vector2f(0, 0));

		
		// ------	BOT_WHEEL vs BOT_BODY	------
		BOT_WHEEL__BOT_BODY.setMu(0);
		BOT_WHEEL__BOT_BODY.setMuOrthogonal(0);
		BOT_WHEEL__BOT_BODY.setBounce(0);
		BOT_WHEEL__BOT_BODY.setDampingCoefficient(0.5f);
		BOT_WHEEL__BOT_BODY.setMinimumBounceVelocity(10);
		
		
		// ------	BOT_WHEEL vs BOT_DRIBBLER	------
		BOT_WHEEL__BOT_DRIBBLER.setMu(0);
		BOT_WHEEL__BOT_DRIBBLER.setMuOrthogonal(0);
		BOT_WHEEL__BOT_DRIBBLER.setBounce(0);
		BOT_WHEEL__BOT_DRIBBLER.setDampingCoefficient(0.5f);
		BOT_WHEEL__BOT_DRIBBLER.setMinimumBounceVelocity(10);

		
		// ------	BOT_WHEEL vs BOT_KICKER	------
		BOT_WHEEL__BOT_KICKER.setMu(0);
		BOT_WHEEL__BOT_KICKER.setMuOrthogonal(0);
		BOT_WHEEL__BOT_KICKER.setBounce(0);
		BOT_WHEEL__BOT_KICKER.setDampingCoefficient(0.5f);
		BOT_WHEEL__BOT_KICKER.setMinimumBounceVelocity(10);

		
		// ------	BOT_WHEEL vs BOT_WHEEL	------
		BOT_WHEEL__BOT_WHEEL.setMu(0);
		BOT_WHEEL__BOT_WHEEL.setMuOrthogonal(0);
		BOT_WHEEL__BOT_WHEEL.setBounce(0);
		BOT_WHEEL__BOT_WHEEL.setDampingCoefficient(0.5f);
		BOT_WHEEL__BOT_WHEEL.setMinimumBounceVelocity(10);
	}
	
	
	// --------------------------------------------------------------------------
	// --- materials ------------------------------------------------------------
	// --------------------------------------------------------------------------
	public static final Material			BALL				= new RoboCupMaterial("BALL") {
																			
		private static final long	serialVersionUID	= -3290595612637289202L;
		
		
		@Override
		public void doSetProperties() {
			
			this.setSpringPenetrationDepth(0);
			
			this.putContactHandlingDetails(WALL, BALL__WALL);
			this.putContactHandlingDetails(GROUND, BALL__GROUND);
			this.putContactHandlingDetails(NET, BALL__NET);
			this.putContactHandlingDetails(BOT_BODY, BALL__BOT_BODY);
			this.putContactHandlingDetails(BOT_DRIBBLER, BALL__BOT_DRIBBLER);
			this.putContactHandlingDetails(BOT_KICKER, BALL__BOT_KICKER);
			this.putContactHandlingDetails(BOT_BODY_KNOB, BALL__BOT_BODY_KNOB);
		}
	};
	

	public static final RoboCupMaterial	BOT_BODY			= new RoboCupMaterial("BOT_BODY") {
																			
		private static final long	serialVersionUID	= -4628269421499504603L;
		
		
		@Override
		public void doSetProperties() {
			this.setDensity(2f);
			this.setSpringPenetrationDepth(0);
			
			this.putContactHandlingDetails(BALL, BALL__BOT_BODY);
			this.putContactHandlingDetails(GROUND, BOT_BODY__GROUND);
			this.putContactHandlingDetails(BOT_WHEEL, BOT_WHEEL__BOT_BODY);
		}
	};
	

	public static final RoboCupMaterial	BOT_BODY_KNOB	= new RoboCupMaterial("BOT_BODY_KNOB") {
																			
		private static final long	serialVersionUID	= -4628269421499504603L;
		
		
		@Override
		public void doSetProperties() {
			this.setDensity(2f);
			this.setSpringPenetrationDepth(0f);
			
			this.putContactHandlingDetails(BALL, BALL__BOT_BODY_KNOB);
			this.putContactHandlingDetails(GROUND, BOT_BODY_KNOB__GROUND);
		}
	};
	

	public static final RoboCupMaterial	BOT_WHEEL		= new RoboCupMaterial("BOT_WHEEL") {
																			
		private static final long	serialVersionUID	= -4628269421499504603L;
		
		
		@Override
		public void doSetProperties() {
			this.setDensity(0.98f);
			this.setSpringPenetrationDepth(0f);
			
			this.putContactHandlingDetails(GROUND, BOT_WHEEL__GROUND);
			
			this.putContactHandlingDetails(BOT_WHEEL, BOT_WHEEL__BOT_WHEEL);
			this.putContactHandlingDetails(BOT_BODY, BOT_WHEEL__BOT_BODY);
			this.putContactHandlingDetails(BOT_DRIBBLER, BOT_WHEEL__BOT_DRIBBLER);
			this.putContactHandlingDetails(BOT_KICKER, BOT_WHEEL__BOT_KICKER);
		}
	};
	

	public static final RoboCupMaterial	BOT_DRIBBLER	= new RoboCupMaterial("BOT_DRIBBLER") {
																			
		private static final long	serialVersionUID	= -4628269421499504603L;
		
		
		@Override
		public void doSetProperties() {
			this.setDensity(1);
			this.setSpringPenetrationDepth(0);	//10);
			
			this.putContactHandlingDetails(BALL, BALL__BOT_DRIBBLER);
			this.putContactHandlingDetails(BOT_WHEEL, BOT_WHEEL__BOT_DRIBBLER);
		}
	};
	

	public static final RoboCupMaterial	BOT_KICKER		= new RoboCupMaterial("BOT_KICKER") {
																			
		private static final long	serialVersionUID	= -4628269421499504603L;
		
		
		@Override
		public void doSetProperties() {
			this.setDensity(2f);
			this.setSpringPenetrationDepth(0f);
			
			this.putContactHandlingDetails(BALL, BALL__BOT_KICKER);
			this.putContactHandlingDetails(GROUND, BOT_KICKER__GROUND);
			this.putContactHandlingDetails(BOT_WHEEL, BOT_WHEEL__BOT_KICKER);
		}
	};
	

	// --------------------------------------------------------------------------
	// --- field ----------------------------------------------------------------
	// --------------------------------------------------------------------------
	public static final RoboCupMaterial	WALL				= new RoboCupMaterial("WALL") {
																			
		private static final long	serialVersionUID	= -2820118963960097120L;
		
		
		@Override
		public void doSetProperties() {
			
			this.setDensity(10);
			this.setSpringPenetrationDepth(0);
			
			this.putContactHandlingDetails(BALL, BALL__WALL);
		}
	};
	

	public static final RoboCupMaterial	GROUND			= new RoboCupMaterial("GROUND") {
																			
		private static final long	serialVersionUID	= 5782829707413040851L;
		
		
		@Override
		public void doSetProperties() {
			
			this.setDensity(1.5f);
			this.setSpringPenetrationDepth(0.1f);
			
			this.putContactHandlingDetails(BALL, BALL__GROUND);
			this.putContactHandlingDetails(BOT_BODY, BOT_BODY__GROUND);
			this.putContactHandlingDetails(BOT_BODY_KNOB, BOT_BODY_KNOB__GROUND);
			this.putContactHandlingDetails(BOT_KICKER, BOT_KICKER__GROUND);
			this.putContactHandlingDetails(BOT_WHEEL, BOT_WHEEL__GROUND);
		}
	};
	

	// --------------------------------------------------------------------------
	// --- goal -----------------------------------------------------------------
	// --------------------------------------------------------------------------
	public static final RoboCupMaterial	GOALPOST			= new RoboCupMaterial("GOALPOST") {
																			
		private static final long	serialVersionUID	= -6936022665038353297L;
		
		
		@Override
		public void doSetProperties() {
			
			this.setDensity(10);
			this.setSpringPenetrationDepth(0);
		}
	};
	

	public static final RoboCupMaterial	NET				= new RoboCupMaterial("NET") {
																			
		private static final long	serialVersionUID	= 6795218755908979060L;
		
		
		@Override
		public void doSetProperties() {
			
			this.setDensity(10);
			this.setSpringPenetrationDepth(0);
			
			this.putContactHandlingDetails(BALL, BALL__NET);
		}
	};
}
