package edu.dhbw.mannheim.tigers.sim.engine.plugins.cam;


import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map.Entry;

import org.apache.commons.configuration.SubnodeConfiguration;
import org.apache.log4j.Logger;

import edu.dhbw.mannheim.tigers.sim.engine.model.TimedModelLocation;
import edu.dhbw.mannheim.tigers.sim.engine.model.data.field.IRoboCupFieldData;
import edu.dhbw.mannheim.tigers.sim.engine.plugins.ASimPlugin;
import edu.dhbw.mannheim.tigers.sim.engine.plugins.EPluginType;
import edu.dhbw.mannheim.tigers.sumatra.model.data.MessagesRobocupSslDetection.SSL_DetectionBall;
import edu.dhbw.mannheim.tigers.sumatra.model.data.MessagesRobocupSslDetection.SSL_DetectionFrame;
import edu.dhbw.mannheim.tigers.sumatra.model.data.MessagesRobocupSslDetection.SSL_DetectionRobot;
import edu.dhbw.mannheim.tigers.sumatra.model.data.MessagesRobocupSslGeometry.SSL_GeometryCameraCalibration;
import edu.dhbw.mannheim.tigers.sumatra.model.data.MessagesRobocupSslGeometry.SSL_GeometryData;
import edu.dhbw.mannheim.tigers.sumatra.model.data.MessagesRobocupSslGeometry.SSL_GeometryFieldSize;
import edu.dhbw.mannheim.tigers.sumatra.model.data.MessagesRobocupSslWrapper.SSL_WrapperPacket;


/**
 * <p>
 * This is the abstract base class which allows the implementation and use of different strategies to build cam-output.
 * </p>
 * <p>
 * Sim : SSL-Vision<br/>
 * X : Y<br/>
 * Y : Z<br/>
 * Z : X<br/>
 * <br/>
 * Orientation:<br/>
 * Sim: mathematically (left = positive, right = negative), relative to Xsim<br/>
 * SSL: mathematically (left = positive, right = negative), relative to Xssl<br/>
 * => Convert O(x) to O(z)!
 * </p>
 * 
 * @author Gero
 */
public abstract class ACamOutputPlugin extends ASimPlugin {
	
	// --------------------------------------------------------------------------
	// --- variables and constants ----------------------------------------------
	// --------------------------------------------------------------------------
	protected final Logger		log			= Logger.getLogger(getClass());
	

	/**
	 * As the time provided by {@link System#nanoTime()} has no point of reference, we have to create
	 * one.
	 */
	protected final double		nanoToMillisOffset;
	
	private final List<ACam>	cameras		= new ArrayList<ACam>();
	
	private IRoboCupFieldData	fieldData	= null;
	private SSL_GeometryData	geomData		= null;
	
	
	// --------------------------------------------------------------------------
	// --- constructors ---------------------------------------------------------
	// --------------------------------------------------------------------------
	/**
	 * Conventional constructor for plugin-instantiation
	 * 
	 * @see ACamOutputPlugin
	 * @param config
	 */
	public ACamOutputPlugin(SubnodeConfiguration config) {
		super(config);
		
		// Compute time offset
		long nowNanos = System.nanoTime();
		
		long currentMillis = System.currentTimeMillis();
		
		long nowMillis = nowNanos / 1000000; // TimeUnit.NANOSECONDS.toMillis
		long nowRestNanos = nowNanos % 1000000;
		double nowRestMillis = nowRestNanos / 1000000;
		nanoToMillisOffset = currentMillis - nowMillis - nowRestMillis;
	}
	

	// --------------------------------------------------------------------------
	// --- methods --------------------------------------------------------------
	// --------------------------------------------------------------------------
	/**
	 * Initializes {@link ASimPlugin}s of this type
	 * 
	 * @param initialPacketCount The number the packet-count should start with
	 */
	public void initialize(long initialPacketCount, IRoboCupFieldData fieldData) {
		this.fieldData = fieldData;

		createCameras(cameras);
		

		for (ACam cam : getCameras()) {
			cam.setPacketCount(initialPacketCount);
		}
		
		/*
		 * namespace FieldConstantsRoboCup2009 {
		 * static const int line_width = 10;
		 * static const int field_length=6050;
		 * static const int field_width=4050;
		 * static const int boundary_width=250;
		 * static const int referee_width=425;
		 * static const int goal_width=700;
		 * static const int goal_depth=180;
		 * static const int goal_wall_width=20;
		 * static const int center_circle_radius=500;
		 * static const int defense_radius=500;
		 * static const int defense_stretch=350;
		 * static const int free_kick_from_defense_dist=200;
		 * static const int penalty_spot_from_field_line_dist=450;
		 * static const int penalty_line_from_spot_dist=400;
		 * }
		 */
		SSL_GeometryFieldSize.Builder fieldBuilder = SSL_GeometryFieldSize.newBuilder();
		fieldBuilder.setBoundaryWidth(250);
		fieldBuilder.setCenterCircleRadius(500);
		fieldBuilder.setDefenseRadius(500);
		fieldBuilder.setDefenseStretch(350);
		fieldBuilder.setFieldLength((int) fieldData.getFieldLength());
		fieldBuilder.setFieldWidth((int) fieldData.getFieldWidth());
		fieldBuilder.setFreeKickFromDefenseDist(200);
		fieldBuilder.setGoalDepth((int) fieldData.getGoalInnerDepth());
		fieldBuilder.setGoalWallWidth((int) fieldData.getGoalSideWallWidth());
		fieldBuilder.setGoalWidth((int) fieldData.getGoalInnerWidth());
		fieldBuilder.setLineWidth((int) fieldData.getFieldLineWidth());
		fieldBuilder.setPenaltyLineFromSpotDist(400);
		fieldBuilder.setPenaltySpotFromFieldLineDist(450);
		fieldBuilder.setRefereeWidth((int) fieldData.getJudgesBorderL());
		final SSL_GeometryFieldSize sslField = fieldBuilder.build();
		

		final List<SSL_GeometryCameraCalibration> cameraCalibrations = new ArrayList<SSL_GeometryCameraCalibration>();
		for (ACam cam : getCameras()) {
			SSL_GeometryCameraCalibration.Builder calibBuilder = SSL_GeometryCameraCalibration.newBuilder();
			calibBuilder.setCameraId(cam.getId());
			if (cam.getId() == 0) {
				calibBuilder.setDerivedCameraWorldTx(1200);
			} else // cam.getId() == 1
			{
				calibBuilder.setDerivedCameraWorldTx(-1200);
			}
			calibBuilder.setDerivedCameraWorldTy(0);
			calibBuilder.setDerivedCameraWorldTz(4000);
			
		/*
		 * Data out of TigersLab
		 * 
			SSLCameraCalibration [
			camera_id=0, 
			focal_length=683.5014, 
			principal_point_x=314.51984, 
			principal_point_y=302.36078, 
			distortion=0.46, q0=0.696996, 
			q1=-0.712688, 
			q2=-0.067983, 
			q3=0.040615, 
			tx=368.4048, 
			ty=-1234.5254, 
			tz=2900.0, 
			derived_camera_world_tx=-1117.1123, 
			derived_camera_world_ty=-57.59629, 
			derived_camera_world_tz=2969.606
			###########Cam 1
			SSLCameraCalibration [
			camera_id=1, 
			focal_length=642.4307, 
			principal_point_x=382.77292, 
			principal_point_y=259.42606, 
			distortion=0.32, q0=-0.693495, 
			q1=-0.72042, 
			q2=-0.007742, 
			q3=-3.95E-4, 
			tx=36.24973, 
			ty=-1078.5725, 
			tz=2900.0, 
			derived_camera_world_tx=1049.6241, 
			derived_camera_world_ty=-29.162462, 
			derived_camera_world_tz=2910.68
			*/
				
		   // @todo Are that the correct data? There are only dummy-datas set!!!

			calibBuilder.setFocalLength(642);
			calibBuilder.setPrincipalPointX(300);
			calibBuilder.setPrincipalPointY(300);
			calibBuilder.setDistortion(0);
			calibBuilder.setQ0(0);
			calibBuilder.setQ1(0);
			calibBuilder.setQ2(0);
			calibBuilder.setQ3(0);
			calibBuilder.setTx(0);
			calibBuilder.setTy(0);
			calibBuilder.setTz(0);
			cameraCalibrations.add(calibBuilder.build());
		}
		

		// Geometry frame?
		SSL_GeometryData.Builder geomBuilder = SSL_GeometryData.newBuilder();
		geomBuilder.addAllCalib(cameraCalibrations);
		geomBuilder.setField(sslField);
		geomData = geomBuilder.build();
	}
	

	/**
	 * Subclasses must add cameras here!
	 * 
	 * @param cameras
	 */
	protected abstract void createCameras(List<ACam> cameras);
	

	/**
	 * This function takes every information available and builds a {@link SSL_WrapperPacket} for every {@link ACam}.
	 * 
	 * @param frameCount
	 * @param ball
	 * @param blueBots
	 * @param yellowBots
	 * @return A list of packets, as there may be several 'cameras'
	 */
	public List<SSL_WrapperPacket> buildCamInputs(final TimedModelLocation ball,
			final HashMap<Integer, TimedModelLocation> blueBots, final HashMap<Integer, TimedModelLocation> yellowBots) {
		
		final List<SSL_WrapperPacket> result = new ArrayList<SSL_WrapperPacket>();
		
		for (ACam cam : getCameras()) {
			final SSL_WrapperPacket.Builder packetBuilder = SSL_WrapperPacket.newBuilder();
			final SSL_DetectionFrame.Builder detectionBuilder = SSL_DetectionFrame.newBuilder();
			
			// Blue robots
			for (Entry<Integer, TimedModelLocation> entry : blueBots.entrySet()) {
				final SSL_DetectionRobot.Builder bot = cam.createRobotDetection(entry.getKey(), entry.getValue());
				if (bot != null) {
					detectionBuilder.addRobotsBlue(bot);
				}
			}
			
			// Yellow robots
			for (Entry<Integer, TimedModelLocation> entry : yellowBots.entrySet()) {
				final SSL_DetectionRobot.Builder bot = cam.createRobotDetection(entry.getKey(), entry.getValue());
				if (bot != null) {
					detectionBuilder.addRobotsYellow(bot);
				}
			}
			
			// Ball
			final SSL_DetectionBall.Builder detectedBall = cam.createBallDetection(ball);
			if (detectedBall != null) {
				detectionBuilder.addBalls(detectedBall);
			}
			

			// Assemble
			prepareDetectionPacket(detectionBuilder, cam, ball.getTimestamp(), nanoToMillisOffset);
			packetBuilder.setDetection(detectionBuilder);
			packetBuilder.setGeometry(geomData);
			result.add(packetBuilder.build());
		}
		

		return result;
	}
	

	// --------------------------------------------------------------------------
	// --- supporting functions -------------------------------------------------
	// --------------------------------------------------------------------------
	/**
	 * Applies the given values to the SSL_DetectionFrame.Builder<br/>
	 * <code>setTSent = System.currentMillis();</code><br/>
	 * <code>setFrameNumber = packetCout++;</code>
	 * 
	 * @param detectionBuilder :
	 * @param camID setCameraId
	 * @param timestampInNanos setTCapture (+ {@link #nanoToMillisOffset})
	 * @param nanoToMillisOffset
	 */
	protected void prepareDetectionPacket(SSL_DetectionFrame.Builder detectionBuilder, ACam cam, long timestampInNanos,
			double nanoToMillisOffset) {
		// Detection
		detectionBuilder.setCameraId(cam.getId()).setFrameNumber((int) cam.incPacketCount())
				.setTCapture(((timestampInNanos / 1000000) + nanoToMillisOffset) / 1000)
				.setTSent((System.nanoTime() / 1000000 + nanoToMillisOffset) / 1000);
	}
	

	@Override
	public final EPluginType getType() {
		return EPluginType.CAM_OUTPUT;
	}
	

	/**
	 * @return The {@link ACam} this implementation uses
	 */
	protected List<ACam> getCameras() {
		return cameras;
	}
	

	/**
	 * @return The dimensions of the RoboCup-field
	 */
	protected IRoboCupFieldData getRobocupFieldData() {
		return fieldData;
	}
}
