package edu.dhbw.mannheim.tigers.sim.gui.swing;


import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;

import javax.swing.JFrame;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;
import javax.swing.UIManager;
import javax.swing.text.DefaultCaret;

import org.apache.log4j.Logger;
import org.apache.log4j.PatternLayout;

import edu.dhbw.mannheim.tigers.sim.Sim;
import edu.dhbw.mannheim.tigers.sim.engine.ESimEngineState;
import edu.dhbw.mannheim.tigers.sim.exceptions.SimResourceException;
import edu.dhbw.mannheim.tigers.sim.gui.swing.actions.ExitAction;
import edu.dhbw.mannheim.tigers.sim.gui.swing.actions.OpenPluginSettingsAction;
import edu.dhbw.mannheim.tigers.sim.gui.swing.actions.OpenSettingsWindowAction;
import edu.dhbw.mannheim.tigers.sim.gui.swing.actions.ReloadModelDataAction;
import edu.dhbw.mannheim.tigers.sim.gui.swing.actions.ReloadPluginsAction;
import edu.dhbw.mannheim.tigers.sim.resources.ImageProvider;
import edu.dhbw.mannheim.tigers.sim.resources.ImageProvider.EImage;
import edu.dhbw.mannheim.tigers.sim.settings.ISettingsProvider;
import edu.dhbw.mannheim.tigers.sim.settings.data.IApplicationProperties;
import edu.dhbw.mannheim.tigers.sim.util.logging.StartupLogAppender;


/**
 * This class represents the main-window of the simualtor-application
 * 
 * @author Gero
 * 
 */
public class SimMainWindow extends JFrame {
	
	// --------------------------------------------------------------------------
	// --- variables and constants ----------------------------------------------
	// --------------------------------------------------------------------------
	private static final long					serialVersionUID	= 1L;
	
	public static final String					WINDOW_TITLE		= "Tigers' CAGE - " + Sim.VERSION_TAG;
	
	private JMenuBar								menubar				= null;
	private SimControlToolbar					toolbar				= null;
	private JScrollPane							loggingScrollPane	= null;
	private JTextArea								loggingArea			= null;
	private LogAppender							logAppender			= null;
	
	private int										mainX, mainY, mainHeigth, mainWidth;
	
	private final SwingGUI						gui;
	private final IApplicationProperties	properties;
	
	
	// --------------------------------------------------------------------------
	// --- constructors ---------------------------------------------------------
	// --------------------------------------------------------------------------
	public SimMainWindow(SwingGUI gui) {
		super(WINDOW_TITLE);
		
		this.gui = gui;
		properties = ISettingsProvider.INSTANCE.getSimProperties().getApplicationProperties();
	}
	

	public void init() {
		initGui();
	}
	

	/**
	 * <p>
	 * Initializes this {@link SimMainWindow} GUI
	 * </p>
	 */
	private void initGui() {
		// Set LookAndFeel to: OS Default, looks and feels best ;-)
		try {
			UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
		} catch (Exception err) {
		}
		

		// Layout
		this.setLayout(new BorderLayout());
		

		// Close
		this.setDefaultCloseOperation(JFrame.DO_NOTHING_ON_CLOSE);
		this.addWindowListener(new WindowAdapter() {
			
			@Override
			public void windowClosing(WindowEvent e) {
				properties.setMainWindowHeight(SimMainWindow.this.getHeight());
				properties.setMainWindowWidth(SimMainWindow.this.getWidth());
				
				ExitAction.getInstance(gui.getInputController()).actionPerformed(null);
			}
		});
		
		// JMenuBar
		menubar = new JMenuBar();
		
		JMenu mainmenu = new JMenu("Main");
		mainmenu.add(ExitAction.getInstance(gui.getInputController()));
		menubar.add(mainmenu);
		
		JMenu settingsMenu = new JMenu("Settings");
		settingsMenu.add(OpenSettingsWindowAction.getInstance(gui));
		menubar.add(settingsMenu);
		
		JMenu pluginsMenu = new JMenu("Plugins");
		pluginsMenu.add(ReloadPluginsAction.getInstance(gui));
		pluginsMenu.add(OpenPluginSettingsAction.getInstance(gui));
		menubar.add(pluginsMenu);
		
		JMenu modelMenu = new JMenu("Model");
		modelMenu.add(ReloadModelDataAction.getInstance(gui.getInputController()));
		menubar.add(modelMenu);
		
		this.setJMenuBar(menubar);
		

		// SimControlToolbar
		toolbar = new SimControlToolbar(gui);
		this.add(toolbar, BorderLayout.NORTH);
		

		// Create the JTextArea for all logging information (autoscroll behaviour)
		loggingArea = new JTextArea();
		loggingArea.setEditable(false);
		DefaultCaret caret = (DefaultCaret) loggingArea.getCaret();
		caret.setUpdatePolicy(DefaultCaret.ALWAYS_UPDATE);
		
		// Assign appender to the RootLogger --> all logger information will be written in the
		// JTextArea
		logAppender = new LogAppender(loggingArea);
		logAppender.setLayout(new PatternLayout(Sim.LOGGER_LAYOUT));
		Logger.getRootLogger().addAppender(logAppender);
		
		// Get all events from StartupLogAppender and add them to the output
		StartupLogAppender startupLogger = gui.getStartupLogAppender();
		startupLogger.pumpEventsTo(logAppender);
		startupLogger.close(); // Close it after its job is done.
		
		// JScrollPane for the JTextArea
		loggingScrollPane = new JScrollPane(loggingArea);
		this.add(loggingScrollPane, BorderLayout.CENTER);
		

		// // Status bar
		// statusbar = new JLabel();
		// statusbar.setBorder(BorderFactory.createLoweredBevelBorder());
		//
		// this.add(statusbar, BorderLayout.SOUTH);
		

		// Set size...
		this.setMinimumSize(new Dimension(400, 200));
		
		mainHeigth = properties.getMainWindowHeight();
		mainWidth = properties.getMainWindowWidth();
		this.setSize(mainWidth, mainHeigth);
		
		mainX = properties.getMainWindowPosX();
		mainY = properties.getMainWindowPosY();
		
		this.setLocation(mainX, mainY);
		
		// Dimension d = Toolkit.getDefaultToolkit().getScreenSize();
		// this.setLocation( (d.width - this.getSize().width) / 2, 0);
		

		// add Tigers-Icon
		try {
			this.setIconImage(ImageProvider.getInstance().getImage(EImage.TIGER_ICON_32x32));
		} catch (SimResourceException e1) {
			e1.printStackTrace();
		}
		this.setVisible(true);
	}
	

	public void close() {
		logAppender.close();
	}
	

	public void refreshEngineState(ESimEngineState newState) {
		toolbar.refresh(newState);
		
		// Options-Action
		OpenSettingsWindowAction oswa = OpenSettingsWindowAction.getInstance(gui);
		ReloadPluginsAction rpa = ReloadPluginsAction.getInstance(gui);
		
		switch (newState) {
			case RUNNING:
				oswa.setEnabled(false);
				rpa.setEnabled(false);
				break;
			
			case READY:
			case STOPPED:
				oswa.setEnabled(true);
				rpa.setEnabled(true);
				break;
			default:
				break;
		}
	}
	

	public JMenuBar getMenubar() {
		return menubar;
	}
	

	public IGUIEngineControl getEngineControl() {
		return toolbar;
	}
}
