package edu.dhbw.mannheim.tigers.sim.gui.swing.settings;


import java.awt.Component;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.Font;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

import javax.swing.BorderFactory;
import javax.swing.JButton;
import javax.swing.JPanel;
import javax.swing.border.TitledBorder;

import net.miginfocom.swing.MigLayout;
import edu.dhbw.mannheim.tigers.sim.controller.IInputController;


/**
 * Abstract base class for settings-panels. Provides the general structure and some helping
 * methods/functions.
 * 
 * @author Gero
 * 
 */
public abstract class ASettingsPanel extends JPanel implements ISettingsPanel {
	
	// --------------------------------------------------------------------------
	// --- variables and constants ----------------------------------------------
	// --------------------------------------------------------------------------
	private static final long			serialVersionUID	= 1496063016084727502L;
	
	protected final String				title;
	protected final IInputController	inputController;
	
	
	// --------------------------------------------------------------------------
	// --- constructors ---------------------------------------------------------
	// --------------------------------------------------------------------------
	/**
	  * @see ASettingsPanel
	  * @param title
	  * @param inputController
	  */
	public ASettingsPanel(String title, IInputController inputController) {
		
		this.title = title;
		this.inputController = inputController;
		
		// General setup
		setLayout(new MigLayout("wrap 1", "[grow, fill]", ""));
	}
	

	// --------------------------------------------------------------------------
	// --- methods --------------------------------------------------------------
	// --------------------------------------------------------------------------
	/**
	 * This method must be called manually!!!
	 */
	@Override
	public void initialize() {
		
		loadSettings();
		
		createGui();
		
		revert(); // Set defaults
	}
	

	/**
	 * Hook-method allowing subclasses to everything before the GUI is created to provide the
	 * properties for proper initialization
	 */
	protected abstract void loadSettings();
	

	/**
	 * Hook-method to allow subclasses to create their own GUI =)
	 */
	protected abstract void createGui();
	

	/**
	 * @param title
	 * @return A panel with the default decorations with the given title
	 */
	protected JPanel createDefaultSettingsPanel(String title) {
		JPanel result = new JPanel();
		result.setLayout(new MigLayout("wrap 2", "[grow, fill][grow, fill]", "[]"));
		
		TitledBorder b = BorderFactory.createTitledBorder(title);
		b.setTitleFont(new Font("Default", Font.BOLD, b.getTitleFont().getSize()));
		result.setBorder(b);
		
		return result;
	}
	

	/**
	 * This function provides a already decorated buttons-panel. It is recommend to use this
	 * function.
	 */
	protected void addButtonsPanel() {
		JPanel buttonPanel = new JPanel();
		buttonPanel.setLayout(new FlowLayout());
		
		JButton b = new JButton("Defaults");
		sizeBtn(b, SettingsFrame.BTN_SIZE);
		b.addActionListener(new ActionListener() {
			
			@Override
			public void actionPerformed(ActionEvent paramActionEvent) {
				defaults();
			}
		});
		buttonPanel.add(b);
		
		b = new JButton("Revert");
		sizeBtn(b, SettingsFrame.BTN_SIZE);
		b.addActionListener(new ActionListener() {
			
			@Override
			public void actionPerformed(ActionEvent paramActionEvent) {
				revert();
			}
		});
		buttonPanel.add(b);
		
		add(buttonPanel);
	}
	

	private void sizeBtn(JButton btn, Dimension size) {
		btn.setSize(size);
		btn.setMinimumSize(size);
		btn.setPreferredSize(size);
		btn.setMaximumSize(size);
	}
	

	@Override
	public void setEnabled(boolean enabled) {
		super.setEnabled(enabled);
		
		enableAllChildren(this, enabled);
	}
	

	private void enableAllChildren(Container container, boolean enable) {
		Component[] components = container.getComponents();
		if (components != null) {
			for (Component comp : components) {
				comp.setEnabled(enable);
				if (comp instanceof Container) {
					enableAllChildren((Container) comp, enable);
				}
			}
		}
	}
	
	
	@Override
	public String getTitle() {
		return title;
	}
}
