package edu.dhbw.mannheim.tigers.sim.gui.swing.settings;


import java.awt.Component;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Vector;

import javax.swing.JComboBox;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JTextField;
import javax.swing.SwingUtilities;

import org.apache.log4j.Logger;
import org.lwjgl.LWJGLException;
import org.lwjgl.opengl.Display;
import org.lwjgl.opengl.DisplayMode;

import com.jme.system.jogl.JOGLSystemProvider;
import com.jme.system.lwjgl.LWJGLSystemProvider;

import edu.dhbw.mannheim.tigers.sim.controller.IInputController;
import edu.dhbw.mannheim.tigers.sim.engine.SimEngineSettings;
import edu.dhbw.mannheim.tigers.sim.settings.ISettingsProvider;
import edu.dhbw.mannheim.tigers.sim.settings.data.IDataFactory;
import edu.dhbw.mannheim.tigers.sim.settings.data.ISimEngineProperties;


/**
 * @author Malte, Gero, inspired by: Matthew D. Hicks (JME)
 */
public class EngineSettingsPanel extends ASettingsPanel {
	
	// --------------------------------------------------------------------------
	// --- variables and constants ----------------------------------------------
	// --------------------------------------------------------------------------
	private static final long					serialVersionUID	= 3490192483002056693L;
	

	protected final Logger						log					= Logger.getLogger(getClass());
	
	public static final int[][]				RESOLUTIONS			= { { 640, 480 }, { 800, 600 },
			{ 1024, 768 }, { 1280, 1024 }, { 1440, 900 }, { 1600, 1024 }, { 1600, 1200 },
			{ 1920, 1200 }												};
	public static final int[]					DEPTHS				= { 16, 24, 32 };
	

	// Sim properties
	private ISimEngineProperties				simProperties;
	
	private JTextField							cpsField;
	private JTextField							fpsField;
	private JTextField							vqField;
	

	// Core engine
	private SimEngineSettings					settings;
	private DisplayMode[]						allModes				= new DisplayMode[0];
	/** Everything for Display.getAvailableDisplayModes... :-/ */
	private final Object							modesSync			= new Object();
	private boolean								modesLock			= false;
	
	private JComboBox								renderer;
	private JComboBox								resolution;
	private JComboBox								depth;
	private JComboBox								frequency;
	private JComboBox								verticalSync;
	private JComboBox								fullscreen;
	private JComboBox								music;
	private JComboBox								sfx;
	private JComboBox								depthBits;
	private JComboBox								alphaBits;
	private JComboBox								stencilBits;
	private JComboBox								samples;
	
	private final Map<String, JComboBox>	map					= new HashMap<String, JComboBox>();
	private final Map<String, Object>		defaults				= new HashMap<String, Object>();
	
	
	// --------------------------------------------------------------------------
	// --- variables and constants ----------------------------------------------
	// --------------------------------------------------------------------------
	/**
	  * @see EngineSettingsPanel
	  * @param inputController
	  */
	public EngineSettingsPanel(IInputController inputController) {
		super("Engine", inputController);
	}
	

	@Override
	public void initialize() {
		synchronized (modesSync) {
			
			super.initialize();
			
			lockModes(); // Wait for getAvailableDisplayModesSometime
		}
	}
	

	@Override
	protected void loadSettings() {
		simProperties = ISettingsProvider.INSTANCE.getSimProperties().getSimEngineProperties();
		
		settings = SimEngineSettings.getInstance();
		
		getAvailableDisplayModesSometime();
	}
	

	@Override
	protected void createGui() {
		
		// Sim settings panel
		JPanel simSettingsPanel = createDefaultSettingsPanel("Sim Properties");
		
		simSettingsPanel.add(new JLabel("Cycles per second:"));
		simSettingsPanel.add(cpsField = new JTextField());
		simSettingsPanel.add(new JLabel("Frames per second:"));
		simSettingsPanel.add(fpsField = new JTextField());
		simSettingsPanel.add(new JLabel("Video capture quality:"));
		simSettingsPanel.add(vqField = new JTextField());
		
		add(simSettingsPanel);
		

		// Core settings panel
		JPanel coreSettingsPanel = createDefaultSettingsPanel("Core Properties");
		
		List<Component> list = getSettingsComponents();
		
		JLabel label = null;
		for (int i = 0; i < list.size(); i++) {
			Component c = list.get(i);
			label = new JLabel(" " + c.getName() + ": ");
//			label.setHorizontalAlignment(SwingConstants.RIGHT);
			
			coreSettingsPanel.add(label);
			coreSettingsPanel.add(c);
		}
		
		add(coreSettingsPanel);
		
		addButtonsPanel();
	}
	

	// --------------------------------------------------------------------------
	// --- methods --------------------------------------------------------------
	// --------------------------------------------------------------------------
	public void addSetting(String name, Object[] choices, Object defaultChoice) {
		defaultChoice = settings.getObject(name, defaultChoice);
		log.info("Default Choice for " + name + " = " + defaultChoice);
		
		JComboBox c = new JComboBox(choices);
		c.setName(name);
		c.setSelectedItem(defaultChoice);
		
		JLabel label = new JLabel(" " + c.getName() + ": ");
		add(label);
		add(c);
		
		map.put(name, c);
		defaults.put(name, defaultChoice);
	}
	

	protected List<Component> getSettingsComponents() {
		List<Component> components = new ArrayList<Component>();
		components.add(createRenderer());
		components.add(createResolution());
		components.add(createDepth());
		components.add(createFrequency());
		components.add(createVerticalSync());
		components.add(createFullscreen());
		components.add(createMusic());
		components.add(createSFX());
		components.add(createDepthBits());
		components.add(createAlphaBits());
		components.add(createStencilBits());
		components.add(createSamples());
		return components;
	}
	

	protected Component createRenderer() {
		renderer = new JComboBox(new Object[] { LWJGLSystemProvider.LWJGL_SYSTEM_IDENTIFIER,
				JOGLSystemProvider.SYSTEM_IDENTIFIER });
		renderer.setName("Renderer");
		renderer.setEnabled(false);
		return renderer;
	}
	

	protected Component createResolution() {
		resolution = new JComboBox(getResolutionArray());
		resolution.setName("Resolution");
		ItemListener resListener = new ItemListener() {
			@Override
			public void itemStateChanged(ItemEvent event) {
				updateResolution();
			}
		};
		resolution.addItemListener(resListener);
		return resolution;
	}
	
	
	private void updateResolution() {
		// The resolution combobox is all we care about
		String[] parser = ((String) resolution.getSelectedItem()).split("x");
		setMenuOptions(Integer.parseInt(parser[0]), Integer.parseInt(parser[1]));
	}
	

	protected Component createDepth() {
		depth = new JComboBox(getDepthArray());
		depth.setName("Depth");
		return depth;
	}
	

	/**
	 * Sets the other menu options based on the given width and height
	 * parameters.
	 * 
	 * @param width
	 * @param height
	 */
	private void setMenuOptions(int width, int height) {
		Vector<DisplayMode> availableModes = getAvailableModesRes(allModes, width, height);
		depth.removeAllItems();
		frequency.removeAllItems();
		HashSet<String> depths = new HashSet<String>();
		HashSet<String> frequencies = new HashSet<String>();
		
		for (DisplayMode aMode : availableModes) {
			depths.add(String.valueOf(aMode.getBitsPerPixel()));
			frequencies.add(String.valueOf(aMode.getFrequency()));
		}
		for (String oneDepth : depths) {
			depth.addItem(oneDepth);
		}
		for (String oneFreq : frequencies) {
			frequency.addItem(oneFreq);
		}
		depth.updateUI();
		frequency.updateUI();
	}
	

	/**
	 * Gets the available modes based on the set of modes for the system and a
	 * resolution.
	 * 
	 * @param theModes
	 * @param width
	 * @param height
	 * @return
	 */
	private static Vector<DisplayMode> getAvailableModesRes(DisplayMode[] theModes, int width,
			int height) {
		Vector<DisplayMode> modes = new Vector<DisplayMode>();
		
		for (int[] res : RESOLUTIONS) {
			if (res[0] == width && res[1] == height) {
				for (DisplayMode aMode : theModes) {
					if (aMode.getHeight() == height && aMode.getWidth() == width) {
						modes.add(aMode);
					}
				}
			}
		}
		return modes;
	}
	

	private static Object[] getResolutionArray() {
		Object[] resolutions = new Object[RESOLUTIONS.length];
		for (int i = 0; i < resolutions.length; i++) {
			resolutions[i] = RESOLUTIONS[i][0] + "x" + RESOLUTIONS[i][1];
		}
		return resolutions;
	}
	

	private static Object[] getDepthArray() {
		Object[] depths = new Object[DEPTHS.length];
		for (int i = 0; i < depths.length; i++) {
			depths[i] = String.valueOf(DEPTHS[i]);
		}
		return depths;
	}
	

	protected Component createFrequency() {
		frequency = new JComboBox(new Object[] { "60", "70", "72", "75", "85", "100", "120", "140" });
		frequency.setName("Frequency");
		return frequency;
	}
	

	protected Component createVerticalSync() {
		verticalSync = new JComboBox(new Object[] { "Yes", "No" });
		verticalSync.setName("Vertical Sync");
		return verticalSync;
	}
	

	protected Component createFullscreen() {
		fullscreen = new JComboBox(new Object[] { "Yes", "No" });
		fullscreen.setName("Fullscreen");
		return fullscreen;
	}
	

	protected Component createMusic() {
		music = new JComboBox(new Object[] { "Yes", "No" });
		music.setName("Music");
		return music;
	}
	

	protected Component createSFX() {
		sfx = new JComboBox(new Object[] { "Yes", "No" });
		sfx.setName("Sound Effects");
		return sfx;
	}
	

	protected Component createDepthBits() {
		depthBits = new JComboBox(new Object[] { "8" });
		depthBits.setName("Depth Bits");
		return depthBits;
	}
	

	protected Component createAlphaBits() {
		alphaBits = new JComboBox(new Object[] { "0" });
		alphaBits.setName("Alpha Bits");
		return alphaBits;
	}
	

	protected Component createStencilBits() {
		stencilBits = new JComboBox(new Object[] { "0" });
		stencilBits.setName("Stencil Bits");
		return stencilBits;
	}
	

	protected Component createSamples() {
		samples = new JComboBox(new Object[] { "0", "2", "4", "8", "16", "32" });
		samples.setName("Samples");
		return samples;
	}
	

	@Override
	public void defaults() {
		synchronized (modesSync) {
			waitForModesLock();
			
			try {
				settings.clear();
				settings.setDefaults();
				
				fillFields(IDataFactory.eINSTANCE.createSimEngineProperties());
			} catch (Exception exc) {
				log.error("Error setting values to default!", exc);
			}
		}
	}
	

	@Override
	public void revert() {
		synchronized (modesSync) {
			waitForModesLock();
			
			fillFields(simProperties);
		}
	}
	

	private void fillFields(ISimEngineProperties props) {
		cpsField.setText(props.getCPS() + "");
		fpsField.setText(props.getFPS() + "");
		vqField.setText(props.getVideoQuality() + "");
		
		renderer.setSelectedItem(settings.getRenderer());
		resolution.setSelectedItem(settings.getWidth() + "x" + settings.getHeight());
		depth.setSelectedItem(String.valueOf(settings.getDepth()));
		frequency.setSelectedItem(String.valueOf(settings.getFrequency()));
		verticalSync.setSelectedItem(settings.isVerticalSync() ? "Yes" : "No");
		fullscreen.setSelectedItem(settings.isFullscreen() ? "Yes" : "No");
		music.setSelectedItem(settings.isMusic() ? "Yes" : "No");
		sfx.setSelectedItem(settings.isSFX() ? "Yes" : "No");
		depthBits.setSelectedItem(String.valueOf(settings.getDepthBits()));
		alphaBits.setSelectedItem(String.valueOf(settings.getAlphaBits()));
		stencilBits.setSelectedItem(String.valueOf(settings.getStencilBits()));
		samples.setSelectedItem(String.valueOf(settings.getSamples()));
		for (String name : map.keySet()) {
			JComboBox combo = map.get(name);
			combo.setSelectedItem(settings.getObject(name, defaults.get(name)));
		}
	}
	

	@Override
	public void apply() {
		synchronized (modesSync) {
			waitForModesLock();
			
			settings.setRenderer((String) renderer.getSelectedItem());
			String[] parser = ((String) resolution.getSelectedItem()).split("x");
			settings.setWidth(Integer.parseInt(parser[0]));
			settings.setHeight(Integer.parseInt(parser[1]));
			settings.setDepth(Integer.parseInt((String) depth.getSelectedItem()));
			settings.setFrequency(Integer.parseInt((String) frequency.getSelectedItem()));
			settings.setVerticalSync(verticalSync.getSelectedItem().equals("Yes"));
			settings.setFullscreen(fullscreen.getSelectedItem().equals("Yes"));
			settings.setMusic(music.getSelectedItem().equals("Yes"));
			settings.setSFX(sfx.getSelectedItem().equals("Yes"));
			settings.setDepthBits(Integer.parseInt((String) depthBits.getSelectedItem()));
			settings.setAlphaBits(Integer.parseInt((String) alphaBits.getSelectedItem()));
			settings.setStencilBits(Integer.parseInt((String) stencilBits.getSelectedItem()));
			settings.setSamples(Integer.parseInt((String) samples.getSelectedItem()));
			
			for (String name : map.keySet()) {
				settings.setObject(name, map.get(name).getSelectedItem());
			}
			
			try {
				settings.save();
			} catch (IOException e) {
				log.error("Error while saving settings!", e);
			}
			

			// Sim Engine Settings
			simProperties.setCPS(Integer.parseInt(cpsField.getText()));
			simProperties.setFPS(Integer.parseInt(fpsField.getText()));
			simProperties.setVideoQuality(Float.parseFloat(vqField.getText()));
		}
	}
	

	@Override
	public boolean checkInput() {
		synchronized (modesSync) {
			waitForModesLock();
			
			return validateDisplay();
		}
	}
	

	private boolean validateDisplay() {
		if (depth.getSelectedItem() == null) {
			return false;
		} else if (frequency.getSelectedItem() == null) {
			return false;
		}
		return true;
	}
	

	private void waitForModesLock() {
		synchronized (modesSync) {
			while (modesLock) {
				try {
					modesSync.wait();
				} catch (InterruptedException err) {
					log.error("Error while waiting for mode-lock!", err);
				}
			}
		}
	}
	
	
	private void unlockModes() {
		synchronized (modesSync) {
			modesLock = false;
			modesSync.notifyAll();

			EngineSettingsPanel.this.setEnabled(true);
		}
	}
	
	
	private void lockModes() {
		synchronized (modesSync) {
			this.setEnabled(false);
			modesLock = true;
		}
	}
	

	/**
	 * This method is necessary because {@link Display#getAvailableDisplayModes()} is far too
	 * time-consuming
	 */
	private void getAvailableDisplayModesSometime() {
		
		new Thread(new Runnable() {
			
			@Override
			public void run() {
				try {
					// First, get modes
					DisplayMode[] modes = Display.getAvailableDisplayModes(); // <-- This operation needs
																									// so much time and
					// thus leads to a bit more effort... :->
					
					// Then, apply them to the gui...
					EngineSettingsPanel.this.allModes = modes;
					
					// ...update (all) text-fields etc. ...
					SwingUtilities.invokeLater(new Runnable() {
						@Override
						public void run() {
							updateResolution();
							
							// ...and finally, unlock it!
							unlockModes();
						}
					});
					
				} catch (LWJGLException err) {
					log.error("Error while updating available display modes!", err);
					unlockModes();	// Release...
				}
			}
		}).start();
	}
}
