package edu.dhbw.mannheim.tigers.sim.resources;

import java.util.HashMap;

import org.apache.log4j.Logger;

import edu.dhbw.mannheim.tigers.sim.exceptions.SimResourceException;


/**
 * This is the base class that loads resources and provides them to the application by mapping it to an {@link Enum}
 * -type.
 * 
 * @author Gero
 * 
 * @param <KType>
 * @param <DType>
 */
public abstract class AResourceProvider<KType, DType> implements IResourceProvider {
	
	// --------------------------------------------------------------------------
	// --- variables and constants ----------------------------------------------
	// --------------------------------------------------------------------------
	protected final Logger				log;
	

	protected HashMap<KType, DType>	storage;
	
	private boolean						initialized	= false;
	protected final Object				storeLock	= new Object();
	
	
	// --------------------------------------------------------------------------
	// --- constructors ---------------------------------------------------------
	// --------------------------------------------------------------------------
	/**
	 * @see AResourceProvider
	 * @param name
	 */
	public AResourceProvider(String name) {
		this.log = Logger.getLogger(name);
	}
	

	/**
	 * @see AResourceProvider
	 * @param clazz
	 */
	public AResourceProvider(Class<?> clazz) {
		this.log = Logger.getLogger(clazz);
	}
	

	// --------------------------------------------------------------------------
	// --- methods --------------------------------------------------------------
	// --------------------------------------------------------------------------
	@Override
	public void initialize() throws SimResourceException {
		synchronized (storeLock) {
			if (initialized) {
				return;
			}
			
			doInitialize();
			
			load();
			
			initialized = true;
		}
	}
	

	/**
	 * This method is called in {@link #initialize()}. It should be used to prepare any own storage solution or so.
	 */
	protected abstract void doInitialize();
	
	
	private void load() {
		synchronized (storeLock) {
			doLoad();
		}
	}
	

	/**
	 * In this method the formerly prepared storage ({@link #doInitialize()}) should be filled with data.
	 */
	protected abstract void doLoad();
	

	@Override
	public void reload() throws SimResourceException {
		synchronized (storeLock) {
			
			if (storage != null) { // For ResourceProvider not using the provided storage (like IModelDataProvider!)
				storage.clear();
			}
			
			load();
		}
	}
	

	/**
	 * Provides access to the loaded data
	 * 
	 * @param key
	 * @return The data that is associated with the given key
	 * @throws SimResourceException If there is no data associated with the given key
	 */
	protected DType getData(KType key) throws SimResourceException {
		synchronized (storeLock) {
			initialize();
			
			DType result = storage.get(key);
			
			if (result == null) {
				throw new SimResourceException("No data mapped to this key: " + key.toString());
			}
			
			return result;
		}
	}
	

	@Override
	public void cleanup() {
		synchronized (storeLock) {
			
			if (storage != null) {
				storage.clear();
				storage = null;
			}
			
			initialized = false;
		}
	}
}
