/*
 * *********************************************************
 * Copyright (c) 2009 - 2010, DHBW Mannheim - Tigers Mannheim
 * Project: TIGERS - Sumatra
 * Date: 22.07.2010
 * Author(s): Gero
 * 
 * *********************************************************
 */
package edu.dhbw.mannheim.tigers.sim.util.network;


import java.io.EOFException;
import java.io.IOException;
import java.net.DatagramPacket;
import java.net.DatagramSocket;
import java.net.InetAddress;
import java.net.SocketException;

import org.apache.log4j.Logger;


/**
 * Receives a datagram-packet on the
 * 
 * @author Gero
 * 
 */
public class UnicastUDPReceiver implements IReceiver {
	
	// --------------------------------------------------------------------------
	// --- variables and constants ----------------------------------------------
	// --------------------------------------------------------------------------
	protected final Logger	log				= Logger.getLogger(getClass());
	
	// Connection
	private DatagramSocket	socket			= null;
	
	/** The internal state-switch of this receiver */
	private boolean			readyToReceive	= false;
	
	
	// --------------------------------------------------------------------------
	// --- constructors ---------------------------------------------------------
	// --------------------------------------------------------------------------
	/**
	 * @see UnicastUDPReceiver
	 * @param localPort
	 */
	public UnicastUDPReceiver(int localPort) {
		
		// Create socket
		boolean socketError = true;
		while (socket == null && socketError) { // Tries to open an socket until it works! =)
			try {
				this.socket = new DatagramSocket(localPort);
			} catch (SocketException err) {
				log.warn("Error while opening a socket on port: '" + localPort + "'! (UDP; maybe port blocked???)", err);
				localPort++;
				log.debug("Trying to open a socket on port '" + localPort + "'...");
			}
		}
		
		synchronized (this) {
			readyToReceive = true;
		}
	}
	

	UnicastUDPReceiver(DatagramSocket socket) {
		this.socket = socket;
		
		synchronized (this) {
			readyToReceive = true;
		}
	}
	

	// --------------------------------------------------------------------------
	// --- method ---------------------------------------------------------------
	// --------------------------------------------------------------------------
	@Override
	public DatagramPacket receive(DatagramPacket store) throws IOException {
		
		// Synchronize access to socket as it belongs to the 'state'
		DatagramSocket socket = null;
		synchronized (this) {
			if (!isReady()) {
				log.error("Receiver is not ready to receive!");
				return null;
			}
			
			socket = this.socket;
		}
		
		// Receive _outside_ the synchronized state, to prevent blocking of the state
		try {
			socket.receive(store); // New data is written into the packets buffer
		} catch (EOFException eof) {
			log.error("EOF error, buffer may be too small!", eof);
		}
		
		return store;
	}
	

	@Override
	public synchronized void cleanup() throws IOException {
		readyToReceive = false;
		
		if (socket != null) {
			socket.close(); // May cause socket to throw IOException; think about it in calling class!
			socket = null;
		}
	}
	

	// --------------------------------------------------------------------------
	// --- getter/setter --------------------------------------------------------
	// --------------------------------------------------------------------------
	@Override
	public synchronized InetAddress getLocalAddress() {
		return readyToReceive ? socket.getLocalAddress() : null;
	}
	

	@Override
	public synchronized int getLocalPort() {
		return readyToReceive ? socket.getLocalPort() : UNDEFINED_PORT;
	}
	

	@Override
	public synchronized boolean isReady() {
		return readyToReceive;
	}
	

	synchronized DatagramSocket getSocket() {
		return socket;
	}
}
